﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace FliProCameraLib
{
    /// <summary>
    /// The FliFrame class encapsulates all data, metadata,
    /// and methods for one image frame.
    /// </summary>
    public partial class FliFrame : FliPro
    {
        //**************************************************************
        #region Constants and Enums

        public const int BITS_PER_PIXEL_12 = 12;
        public const int BITS_PER_PIXEL_16 = 16;
        public const double GAIN_RATIO_FACTOR = 8.7255;
        public const double ADU_SCALE_FACTOR = 3.37;
        public const uint MINIMUM_HIGH_GAIN_COMPARE_VALUE = 1500;  // The rounding error in the low gain pixels is not acceptible below this

        /// <summary>
        /// The types of Image Captures available
        /// </summary>
        public enum FliImageMode
        {
            LDROnly,
            LDROnly_DD,
            HDR,
            HDR_DD,
            RollingHDR,
            RollingHDR_DD,
            GlobalHDR,
            GlobalHDR_DD
        }

        /// <summary>
        /// The types of Images that are stored in the Frame
        /// </summary>
        public enum FliImageType
        {
            HighGain,
            HighGainDummy,
            LowGain,
            LowGainDummy,
            Merged,
            MergedDummy
        }

        protected static int shortByteStride = BitConverter.GetBytes((short)0).Length;

        #endregion Constants and Enums

        //**************************************************************
        #region Constructor/Destructor

        /// <summary>
        /// Initialize instance with Camera metadata and new data buffer
        /// </summary>
        public FliFrame(FliProCamera camera = null, int bufsize = 0) :
               this(camera, (bufsize == 0 ? null : new byte[bufsize]))
        {
            IsInUse = false;
        }

        /// <summary>
        /// Initialize instance with Camera metadata and existing data buffer
        /// </summary>
        public FliFrame(FliProCamera camera, byte[] buf)
        {
            RawData = buf;

            if (RawData != null)
            {
                Array.Clear(RawData, 0, RawData.Length);
            }

            IsQueued = false;
            IsProcessed = false;
            IsCapturedAndProcessed = false;
            IsInUse = false;
            RawDataLength = 0;
        }

        #endregion Constructor/Destructor

        //**************************************************************
        #region Private Properties

        private FliProCameraManager fliProCameraManager = FliProCameraManager.Instance;

        #endregion Private Properties

        //**************************************************************
        #region Camera Metadata

        public FliProCamera Camera { get; protected set; }

        public int CameraMode { get; protected set; }

        public double Temperature { get; protected set; }

        public FliRectangleWH VisibleAreaWH { get; protected set; }

        #endregion Camera Metadata

        //**************************************************************
        #region Frame Metadata

        public FliProFrameMetadata FrameMetaData { get; set; }

        public FliFrameType FrameType { get; protected set; }

        /// <summary>
        /// The area of the captured image without the dummy data
        /// </summary>
        public FliRectangleWH FrameAreaWH { get; set; }

        /// <summary>
        /// The area of the captured image including the dummy data
        /// </summary>
        public FliRectangleWH FrameAreaWHTotal { get; set; }

        public FliImageMode ImageMode { get; set; }

        /// <summary>
        /// The raw frame metadata received from the camera
        /// </summary>
        public byte[] RawMetaData = null;

        public uint SequenceNumber { get; set; }

        #endregion Frame Metadata

        //**************************************************************
        #region Frame Data

        public DateTime CaptureTime { get; set; }

        /// <summary>
        /// True if the Corrected Frame was created from the Subtracted Frame
        /// False if it was create from a base frame
        /// </summary>
        public bool CorrectedAgainstSubtractFrame { get; set; }

        /// <summary>
        /// The entire raw frame data as received from the camera
        /// </summary>
        public byte[] RawData = null;

        public uint RawDataLength { get; set; }

        /// <summary>
        /// The 12 bit packed Frame containing the Dummy Raw Data.
        /// </summary>
        public byte[] DummyRawData = null;

        #endregion Frame Data

        //**************************************************************
        #region ImageStatistics Properties

        #endregion ImageStatistics Properties

        //**************************************************************
        #region FrameQueue Properties

        //-=-= TODO: Add timestamps for actual frame acquisition time
        //-=-= FIXME: Replace timestamps with high resolution Stopwatch timestamps
        public DateTime InsertTime { get; set; }

        public DateTime RemoveTime { get; set; }

        public TimeSpan TimeInQueue { get; set; }

        #endregion FrameQueue Properties

        //**************************************************************
        #region Observable Properties

        // String to uniquely identify the camera. Could be serial number, filename, or test pattern name
        public string CameraID { get; set; }

        public int BitsPerPixel { get; set; }

        public bool IsCapturedAndProcessed { get; set; }

        public bool IsInUse { get; set; }
        
        public bool IsProcessed { get; set; }

        // Flag to signal that the Frame has been inserted into a camera's Frame Queue
        public bool IsQueued { get; set; }

        public bool IsValid { get; set; }

        #endregion Observable Properties

        //**************************************************************
        #region Public Methods

        public void DisposeFrame()
        {
            RawMetaData = null;
            RawData = null;
            DummyRawData = null;
            FrameMetaData = null;
        }

        #endregion Public Methods

        //**************************************************************
        #region Protected Methods
        #endregion Protected Methods

        //**************************************************************
        #region Private Methods

        private string GetValidFilename(string path, string filename, string extension)
        {
            string newFileName = filename;
            DirectoryInfo di = new DirectoryInfo(path);
            FileInfo[] TXTFiles = di.GetFiles(filename + "*" + extension);

            List<int> indexList = new List<int>();
            for (int i = 0; i < TXTFiles.Length; i++)
            {
                string tempName = TXTFiles[i].Name.Substring(filename.Length, TXTFiles[i].Name.Length - (filename.Length + extension.Length)); // subtract off extension ie. ".csv"
                if (!String.IsNullOrEmpty(tempName))
                {
                    try
                    {
                        indexList.Add(Convert.ToInt32(tempName));
                    }
                    catch { }
                }
            }
            int fileIndex = 1;
            while (indexList.Contains(fileIndex))
            {
                fileIndex++;
            }
            return path + "\\" + newFileName + fileIndex.ToString() + extension;
        }

        #endregion Private Methods

        //**************************************************************
        #region Event Handlers
        #endregion Event Handlers

    }
}
