﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using FliUtils.Jobs;
using System.IO;
using System.Drawing.Imaging;
using System.Linq;
using System.Drawing;
using System.Windows.Media.Imaging;
using System.Windows.Media;
using System.Windows;
using System.IO.Compression;
using static FliProCameraLib.FliProCameraManager;

namespace FliProCameraLib
{
    /// <summary>
    /// Each instance of the FliProCamera class represents one Finger Lakes Imaging camera.
    /// This file contains methods and properties common to all classes derived from FliProCamera.
    /// </summary>
    public partial class FliProCamera : FliPro
    {
        //**************************************************************
        #region Public Constants

        public const int FLI_TEMPERATURE_INTERNAL = 0x0000;
        public const int FLI_TEMPERATURE_EXTERNAL = 0x0001;
        public const int FLI_TEMPERATURE_CCD = 0x0000;
        public const int FLI_TEMPERATURE_BASE = 0x0001;
        public const int FLI_TEMPERATURE_AMBIENT = 0x0002;
        public const int FLI_DEFAULT_TEMPERATURE_READ_INTERVAL = 5000;   // Check temperature every 5 seconds

        public const string FILE_EXTENSION_BIN = ".bin";
        public const string FILE_EXTENSION_FIT = ".fit";
        public const string FILE_EXTENSION_JPEG = ".jpg";
        public const string FILE_EXTENSION_PNG = ".png";
        public const string FILE_EXTENSION_RAW = ".raw";
        public const string FILE_EXTENSION_RCD = ".rcd";
        public const string FILE_EXTENSION_TIFF = ".tif";
        public const string FILE_EXTENSION_TXT = ".txt";

        public const string CAMERA_MODEL_KEPLER_0400 = "KEPLER KL400";
        public const string CAMERA_MODEL_KEPLER_0400_2 = "KEPLER 0400";
        public const string CAMERA_MODEL_KEPLER_0400_3 = "KEPLEPKL400";
        public const string CAMERA_MODEL_KEPLER_4040 = "KEPLER KL4040";
        public const string CAMERA_MODEL_KEPLER_2020 = "KEPLER KL2020";
        public const string CAMERA_MODEL_KODAK_29050 = "KODAK 29050";
        public const string CAMERA_MODEL_KODAK_47051 = "KODAK 47051";

        public const string CAMERA_MODEL_COBALT_DC230_42 = "Cobalt DC230-42";
        public const string CAMERA_MODEL_COBALT_DC230_84 = "Cobalt DC230-84";

        private const int I2C_DEVICE_ID_FPGA = 0x31;

        public const string CAMERAL_MODEL_FLAT_FIELD_CORRECTED_RESULTS = "FlatFieldCorrectedResults";
        public const string CAMERAL_MODEL_MERGED_RESULTS = "MergedResults";
        public const string CAMERAL_MODEL_SUBTRACT_RESULTS = "DifferenceResults";

        public const int MAX_THUMBNAIL_SIZE = 0x1000;
        public const int THUMBNAIL_DIMENSION_SIZE = 64; 

        #endregion Public Constants

        //**************************************************************
        #region Constructor and Initialization

        //-------------------------------------------
        /// <summary>
        /// Constructor
        /// </summary>
        public FliProCamera(FPRODEVICEINFO DevInfo)
        {
            ForcePropertyChange = true;

            DeviceLock = new object();
            IOPortLock = new object();
            ImageWriteLock = new object();
            ImageCaptureLock = new object();
            FrameListLock = new object();
            Handle = INVALID_HANDLE;
            IsConnected = false;
            IsCapturing = false;
            FactoryChangesMade = false;
            Type = CameraType.UNKNOWN;

            StatusUpdateCount = 0;
            StatusMonitorJob = new PeriodicJob(UpdateCameraStatus, StatusMonitorInterval, true, 
                                               "StatusMonitor", ThreadPriority.Normal);
            FrameCaptureJob = new StartStopJob(FrameCaptureAction, 
                                               "FrameCapture", ThreadPriority.Highest);

            SetDeviceInfo(DevInfo);
            InitializeBase();
            Initialize();
            SetDefaultGainTables();

            ForcePropertyChange = false;
        }

        //-------------------------------------------
        /// <summary>
        /// We need an empty Constructor to create the property backers for the entire UI
        /// </summary>
        public FliProCamera() : this(new FPRODEVICEINFO())
        {
            DeviceLock = new object();
            IOPortLock = new object();
            ImageCaptureLock = new object();
            FrameListLock = new object();
            Handle = INVALID_HANDLE;
            IsConnected = false;
            IsCapturing = false;
            FactoryChangesMade = false;
            Type = CameraType.UNKNOWN;

            SetDefaultFliFrameCountList();
            SetDefaultFliFrameTypeList();
            SetDefaultSupportedModes();
            //SetDefaultSupportedBinRatios();
            SetDefaultSupportedPixelDepths();
            SetDefaultGainTables();

            LowGainValue = LowGainTable[0];
            HighGainValue = HighGainTable[0];

            StatusUpdateCount = 0;
        }

        //-------------------------------------------
        /// <summary>
        /// Called to set the DevInfo for a Camera that was created as the empty startup camera
        /// </summary>
        /// <param name="DevInfo"></param>
        public virtual void SetDeviceInfo(FPRODEVICEINFO DevInfo)
        {
            DeviceInfo = DevInfo;

            //-=-= TODO: What do we really need for Model?
            if (String.IsNullOrEmpty(DevInfo.cFriendlyName))
            {
                Model = EMPTY_MODEL;
            }
            else
            {
                Model = DeviceInfo.cFriendlyName;
            }

            if (String.IsNullOrEmpty(DevInfo.cSerialNo))
            {
                SerialNumber = EMPTY_SERIAL_NUMBER;
            }
            else
            {
                SerialNumber = DevInfo.cSerialNo;
            }

            switch (DevInfo.USBSpeed)
            {
                case FProUSBSpeed.FULLSPEED:
                    Speed = " (Full Speed)";
                    break;
                case FProUSBSpeed.HIGHSPEED:
                    Speed = " (High Speed)";
                    break;
                case FProUSBSpeed.SUPERSPEED:
                    Speed = " (Super Speed)";
                    break;
                default:
                    Speed = String.Empty;
                    break;
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Initialize all properties
        /// </summary>
        protected virtual void InitializeBase()
        {
            StatusMonitorInterval = TimeSpan.FromMilliseconds(2000);

            HwVersion = 0;
            FwVersion = 0;
            ColorPattern = FliBayerPattern.NONE;
            IsConnected = false;

            //CameraControlSettings = new FliProCameraControlSettings();
            //FrameSettings = new FliProFrameSettings();
            SensorSettings = new FliProSensorSettings();

            SupportedPixelDepths = new ObservableCollection<uint>();
            SupportedBinRatios = new ObservableCollection<BinRatio>();
            SupportedModes = new ObservableCollection<NameValuePair>();
            LowGainTable = new ObservableCollection<SensorGain>();
            HighGainTable = new ObservableCollection<SensorGain>();

            ModeCount = 0;
            BinRatioCount = 0;
            LowGainCount = 0;
            HighGainCount = 0;
            HInverseScanSupported = false;
            VInverseScanSupported = false;

            MaxHBin = 1000;
            MaxVBin = 1000;
            HBin = 1;
            VBin = 1;
            MaxPixelDepth = DEFAULT_MAX_PIXEL_DEPTH;

            MaxImageHeight = MAX_IMAGE_HEIGHT;
            MaxImageWidth = MAX_IMAGE_WIDTH;

            FrameQueue = FliFrameQueue.Instance;
            FrameTimeout = 0;

            FramePool = new ConcurrentQueue<FliFrame>();
            MinimumFramePoolCount = DEFAULT_MIN_FRAME_POOL_COUNT;
            FramePoolBufSize = 0;

            TrackingFramePool = new ConcurrentQueue<FliFrame>();
            MinimumTrackingFramePoolCount = DEFAULT_MIN_TRACKING_FRAME_POOL_COUNT;
            TrackingFramePoolBufSize = 0;

            FrameSequenceNumber = 0;

            // Initialize The Temperature Timer
            TemperatureStatusTimer.Enabled = false;
            TemperatureStatusTimer.Elapsed += TemperatureStatusTimerEvent;
            TemperatureStatusTimer.Interval = FLI_DEFAULT_TEMPERATURE_READ_INTERVAL;
            TemperatureStatusTimer.AutoReset = true;

        }

        #endregion Constructor and Initialization

        //**************************************************************
        #region Support structures and classes

        public struct GainOffsetPair
        {
            public double gainRatio;
            public double temperature1;
            public int lineOffset1;
            public double temperature2;
            public int lineOffset2;
            public double temperature3;
            public int lineOffset3;
            public double stdDev;
        }

        #endregion Support structures and classes

        //**************************************************************
        #region Public Types

        public enum Exposure_Enum
        {
            exposure_0 = 0,
            exposure_5 = 5,
            exposure_10 = 10,
            exposure_30 =30,
            exposure_60 = 60,
            exposure_100 = 100,
            exposure_120 = 120,
            exposure_600 = 600
        }

        public enum Exposure_Triggers
        {
            Normal,
            External_Low,
            External_High
        }

        public enum ExposureUnits_Enum
        {
            units_uSecs,
            units_mSecs,
            units_secs,
            units_mins,
            units_hours
        }

        public enum File_Type
        {
            RawData,
            Low,
            High,
            Merged
        }

        public enum FrameInterval_Enum
        {
            interval_100 = 100,
            interval_600 = 600,
            interval_Other = 10000
        }

        public enum Resolution_Enum
        {
            row_period,
            clock_period
        }

        #endregion Public Types

        //**************************************************************
        #region Public Constants

        public const int MILLISECS_PER_HOUR = 3600 * 1000;
        public const int MILLISECS_PER_MIN = 60 * 1000;
        public const int MILLISECS_PER_SEC = 1000;

        #endregion Public Constants

        //**************************************************************
        #region Protected Properties

        protected object DeviceLock;
        protected object FrameListLock;
        protected object ImageCaptureLock;
        protected object ImageWriteLock;
        protected object IOPortLock;

        protected int Handle
        {
            get { return _Handle; }
            set { _Handle = value; }
        }
        protected int _Handle;

        protected PeriodicJob StatusMonitorJob { get; set; }
        protected TimeSpan StatusMonitorInterval { get; set; }

        protected FliFrameQueue FrameQueue { get; set; }

        protected double DelayCounter { get; set; }
        protected double FrameDelay { get; set; }

        protected StartStopJob FrameCaptureJob { get; set; }

        protected ConcurrentQueue<FliFrame> FramePool { get; set; }

        protected bool ReadErrorEncountered { get; set; }

        protected ConcurrentQueue<FliFrame> TrackingFramePool { get; set; }

        protected System.Timers.Timer TemperatureStatusTimer = new System.Timers.Timer();

        #endregion

        //**************************************************************
        #region Public Properties
        //public string Name { get { return DeviceId.Name; } }

        public int BitsPerPixel { get; set; }
        public string SerialNumber { get; set; }
        public string Speed { get; protected set; }
        public int HwVersion { get; protected set; }
        public int FwVersion { get; protected set; }
        public string Model { get; protected set; }
        public CameraType Type { get; protected set; }
        public uint FrameSequenceNumber { get; protected set; }

        public int LEDOnTime { get; set; }

        public int DeviceType { get; set; }
        public int MaxImageHeight { get; set; }
        public int MaxImageWidth { get; set; }
        public int MetaDataSize = MAX_FRAME_METADATA_BYTES;

        public int MinimumFramePoolCount { get; set; }
        public int MinimumTrackingFramePoolCount { get; set; }

        public int CaptureThreadId { get { return (FrameCaptureJob == null) ? 0 : FrameCaptureJob.ThreadId; } }

        public GainOffsetPair[,] GainOffsetPairArray = null;
        public GainOffsetPair[,] ManufacturersDefaultGainOffsetPairArray = null;

        #endregion Public Properties

        //**************************************************************
        #region Observable Collections
        #endregion Observable Collections

        //**************************************************************
        #region Observable Properties

        public int AbsoluteBrightestThreshold
        {
            get { return _AbsoluteBrightestThreshold; }
            set { _AbsoluteBrightestThreshold = value; }
        }
        private int _AbsoluteBrightestThreshold = 0;

        public int AbsoluteDimmestThreshold
        {
            get { return _AbsoluteDimmestThreshold; }
            set { _AbsoluteDimmestThreshold = value; }
        }
        private int _AbsoluteDimmestThreshold = 0;

        public virtual bool ApplyChannelAveragingVisible { get { return false; } }

        public double BrightestThreshold
        {
            get { return _BrightestThreshold; }
            set
            {
                if (SetValue(ref _BrightestThreshold, value))
                {
                    SetHistogramAtDefault();
                }
            }
        }
        private double _BrightestThreshold = FliProToolSettings.DEFAULT_HISTOGRAM_THRESHOLD;

        //-------------------------------------------
        public FPRODEVICEINFO DeviceInfo
        {
            get { return _DeviceInfo; }
            set { SetValue<FPRODEVICEINFO>(ref _DeviceInfo, value); }
        }
        protected FPRODEVICEINFO _DeviceInfo;

        public Visibility DeviceTypeNot400Visibility
        {
            get { return (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400.GetHashCode() ? Visibility.Visible : Visibility.Collapsed); }
        }

        public double DimmestThreshold
        {
            get { return _DimmestThreshold; }
            set
            {
                if (SetValue(ref _DimmestThreshold, value))
                {
                    SetHistogramAtDefault();
                }
            }
        }
        private double _DimmestThreshold = FliProToolSettings.DEFAULT_HISTOGRAM_THRESHOLD;

        //-------------------------------------------
        /// <summary>
        /// "User friendly" display string
        /// </summary>
        public string DisplayString
        {
            get { return _DisplayString; }
            protected set { SetValue<string>(ref _DisplayString, value); }
        }
        protected string _DisplayString;

        public bool EngineeringBuild { get; set; }
        public bool FirstTraining = true;         // So we don't train the first time the temperature timer fires

        public bool ImmediateExposureVisible
        {
            get { return _ImmediateExposureVisible; }
            set
            {
                if (SetValue(ref _ImmediateExposureVisible, value))
                {
                    // If we are hiding the checkbox, set the Immediate register to off
                    if (!value && Immediate)
                    {
                        Immediate = value;
                    }
                }
            }
        }
        private bool _ImmediateExposureVisible = false;

        //-------------------------------------------
        /// <summary>
        /// Flag to signify that the device Grab is currently being Aborted
        /// </summary>
        public bool IsAborted
        {
            get { return _IsAborted; }
            set { SetValue<bool>(ref _IsAborted, value); }
        }
        protected bool _IsAborted = false;

        //-------------------------------------------
        /// <summary>
        /// True if device is connected
        /// </summary>
        public bool IsConnected
        {
            get { return _IsConnected; }
            protected set { SetValue<bool>(ref _IsConnected, value); }
        }
        protected bool _IsConnected;

        //-------------------------------------------
        /// <summary>
        /// True if device is currently capturing frames
        /// </summary>
        public bool IsCapturing
        {
            get { return _IsCapturing; }
            protected set
            {
                if (SetValue<bool>(ref _IsCapturing, value))
                {
                    if (value == false)
                    {
                        OnCaptureCompleteAction(new EventArgs());
                    }
                }
            }
        }
        protected bool _IsCapturing = true;

        public bool IsDiscardPixelsChecked
        {
            get { return _IsDiscardPixelsChecked; }
            set
            {
                if (SetValue(ref _IsDiscardPixelsChecked, value))
                {
                    SetHistogramAtDefault();
                }
            }
        }
        private bool _IsDiscardPixelsChecked = false;

        //-------------------------------------------
        /// <summary>
        /// False if device is currently capturing frames
        /// </summary>
        public bool IsEnabled
        {
            get { return _IsEnabled; }
            protected set { SetValue<bool>(ref _IsEnabled, value); }
        }
        protected bool _IsEnabled = true;

        //-------------------------------------------
        /// <summary>
        /// Used to prevent multiple initializations during power up
        /// </summary>
        public bool IsInitializedFromSettings
        {
            get { return _IsInitializedFromSettings; }
            set { SetValue<bool>(ref _IsInitializedFromSettings, value); }
        }
        protected bool _IsInitializedFromSettings = false;

        public virtual bool IsLegacyCamera { get { return false; } }

        public bool IsSaveLatestFrameToCameraChecked
        {
            get { return _IsSaveLatestFrameToCameraChecked; }
            set { SetValue(ref _IsSaveLatestFrameToCameraChecked, value); }
        }
        private bool _IsSaveLatestFrameToCameraChecked = false;

        public bool IsUseAbsoluteThresholdsChecked
        {
            get { return _IsUseAbsoluteThresholdsChecked; }
            set
            {
                if (SetValue(ref _IsUseAbsoluteThresholdsChecked, value))
                {
                    SetHistogramAtDefault();
                }
            }
        }
        private bool _IsUseAbsoluteThresholdsChecked = false;

        public bool IsUseLogScaleChecked
        {
            get { return _IsUseLogScaleChecked; }
            set
            {
                if (SetValue(ref _IsUseLogScaleChecked, value))
                {
                    SetHistogramAtDefault();
                }
            }
        }
        private bool _IsUseLogScaleChecked = false;

        public bool SaveSectorsAsFiles
        {
            get { return _SaveSectorsAsFiles; }
            set { SetValue<bool>(ref _SaveSectorsAsFiles, value); }
        }
        protected bool _SaveSectorsAsFiles = false;

        //-------------------------------------------
        /// <summary>
        /// Incremented on each status update
        /// </summary>
        public int StatusUpdateCount
        {
            get { return _StatusUpdateCount; }
            protected set { SetValue<int>(ref _StatusUpdateCount, value); }
        }
        protected int _StatusUpdateCount;

        //-------------------------------------------
        /// <summary>
        /// Property to display camera status in the bottom of the Main Window
        /// </summary>
        public string CameraStatusText
        {
            get { return _StatusText; }
            set { SetValue<string>(ref _StatusText, value); }
        }
        protected string _StatusText;

        //-------------------------------------------
        /// <summary>
        /// Flag to signify that reference data pixels should be included in the statistics
        /// </summary>
        public bool TotalStats
        {
            get { return _TotalStats; }
            set { SetValue<bool>(ref _TotalStats, value); }
        }
        protected bool _TotalStats = true;

        public bool UnscrambleImage
        {
            get { return _UnscrambleImage; }
            set { SetValue<bool>(ref _UnscrambleImage, value); }
        }
        protected bool _UnscrambleImage = false;

        public int UnscrambleImageBandWidth
        {
            get { return _UnscrambleImageBandWidth; }
            set { SetValue(ref _UnscrambleImageBandWidth, value); }
        }
        private int _UnscrambleImageBandWidth = 256;

        //-------------------------------------------
        /// <summary>
        /// Video Mode Capable
        /// </summary>
        public bool VideoModeCapable
        {
            get { return _VideoModeCapable; }
            set { SetValue<bool>(ref _VideoModeCapable, value); }
        }
        protected bool _VideoModeCapable = false;

        //-------------------------------------------
        /// <summary>
        /// Video Mode 
        /// </summary>
        public bool VideoModeEnabled
        {
            get { return _VideoModeEnabled; }
            set
            {
                if (SetValue<bool>(ref _VideoModeEnabled, value))
                {
                    SetVideoModeEnabled(value);
                }
            }
        }
        protected bool _VideoModeEnabled;

        #endregion Observable Properties

        //**************************************************************
        #region Observable Settings Properties

        //-------------------------------------------
        /// <summary>
        /// The current settings of the Camera
        /// </summary>
        //public FliProCameraControlSettings CameraControlSettings
        //{
        //    get { return _CameraControlSettings; }
        //    set { SetValue(ref _CameraControlSettings, value); }
        //}
        //protected FliProCameraControlSettings _CameraControlSettings;

        //-------------------------------------------
        /// <summary>
        /// The current settings for the Frame capture
        /// </summary>
        //public FliProFrameSettings FrameSettings
        //{
        //    get { return _FrameSettings; }
        //    set { SetValue(ref _FrameSettings, value); }
        //}
        //protected FliProFrameSettings _FrameSettings;

        //-------------------------------------------
        /// <summary>
        /// The current settings of the sensor
        /// </summary>
        public FliProSensorSettings SensorSettings
        {
            get { return _SensorSettings; }
            set { SetValue(ref _SensorSettings, value); }
        }
        protected FliProSensorSettings _SensorSettings;

        #endregion Observable Settings Properties

        //**************************************************************
        #region Histogram Settings Properties

        public double ContrastCurvePower { get; set; }
        public int EdgeDiscardPixels { get; set; }
        public int SubtractOffset { get; set; }

        #endregion Histogram Settings Properties

        //**************************************************************
        #region Public Methods

        //-------------------------------------------
        /// <summary>
        /// Open the connection to this device if not already open, 
        /// and obtain basic information if available
        /// (FliOpen, FliGetSerialString, FliGetHWRevision, FliGetFWRevision)
        /// </summary>
        public virtual int Connect()
        {
            int status = -1;
            lock (DeviceLock)
            {
                if (!IsConnected)
                {
                    Handle = INVALID_HANDLE;
                    status = OpenDevice();

                    if (status < 0)
                    {
                        return status;
                    }

                    if (Handle != INVALID_HANDLE)
                    {
                        IsConnected = true;

                        //-=-=ForcePropertyChange = true;

                        OnConnect();
                        OnConnectBase();

                        //-=-=ForcePropertyChange = false;

                        //Console.WriteLine("... Sleeping...");   //-=-= DEBUG
                        //Thread.Sleep(5000); //-=-= DEBUG

                        UpdateCameraStatus();
                        StatusMonitorJob.Start();
                    }
                }
                else
                {
                    // We're already connected
                    status = 0;
                }
            }

            return status;
        }

        public void CopyCameraProperties(FliProCamera copyCamera)
        {
            IsSaveExposureToBINFileChecked = copyCamera.IsSaveExposureToBINFileChecked;
            IsSaveExposureToFITFileChecked = copyCamera.IsSaveExposureToFITFileChecked;
            IsSaveExposureToJPEGFileChecked = copyCamera.IsSaveExposureToJPEGFileChecked;
            IsSaveExposureToPNGFileChecked = copyCamera.IsSaveExposureToPNGFileChecked;
            IsSaveExposureToRAWFileChecked = copyCamera.IsSaveExposureToRAWFileChecked;
            IsSaveExposureToRCDFileChecked = copyCamera.IsSaveExposureToRCDFileChecked;
            IsSaveExposureToTIFFFileChecked = copyCamera.IsSaveExposureToTIFFFileChecked;
            IsSaveMetaDataToFileChecked = copyCamera.IsSaveMetaDataToFileChecked;

            UseHighGainImageChecked = copyCamera.UseHighGainImageChecked;
            UseLowGainImageChecked = copyCamera.UseLowGainImageChecked;
            UseMergedImageChecked = copyCamera.UseMergedImageChecked;

        }

        public byte[] Compress(byte[] data)
        {
            byte[] compressArray = null;
            using (MemoryStream memoryStream = new MemoryStream())
            {
                using (DeflateStream deflateStream = new DeflateStream(memoryStream, CompressionMode.Compress))
                {
                    deflateStream.Write(data, 0, data.Length);
                }
                compressArray = memoryStream.ToArray();
            }
            return compressArray;
        }

        public byte[] Decompress(byte[] data)
        {
            byte[] decompressedArray = null;
            using (MemoryStream decompressedStream = new MemoryStream())
            {
                using (MemoryStream compressStream = new MemoryStream(data))
                {
                    using (DeflateStream deflateStream = new DeflateStream(compressStream, CompressionMode.Decompress))
                    {
                        deflateStream.CopyTo(decompressedStream);
                    }
                }
                decompressedArray = decompressedStream.ToArray();
            }

            return decompressedArray;
        }

        //-------------------------------------------
        /// <summary>
        /// Close the connection to this device, if open
        /// </summary>
        public void Disconnect()
        {
            lock (DeviceLock)
            {
                StatusMonitorJob.Stop();

                if (IsConnected)
                {
                    try
                    {
                        // We don't care about any errors from Close. The camera could be already disconnected
                        CloseDevice();
                    }
                    catch { }
                    Handle = INVALID_HANDLE;
                    IsConnected = false;

                    if (CurrentGPSState == FliGPSState.FPRO_GPS_DETECTED_AND_SAT_LOCK)
                    {
                        CurrentGPSState = FliGPSState.FPRO_GPS_DETECTED_NO_SAT_LOCK;
                        OnPropertyChanged("GPSUnlockedVisibility");
                        OnPropertyChanged("GPSLockedVisibility");
                    }
                }
            }
        }

        public virtual void EnableDriverDebugLogActive(bool enabled)
        {
            if (enabled)
            {
                DebugLevel = DBGLEVEL.TRACE;
            }
            else
            {
                DebugLevel = DBGLEVEL.INFO;
            }
            DebugEnable = enabled;
        }

        public void EnableTemperatureTimer(bool enabled)
        {
            TemperatureStatusTimer.Enabled = enabled;
        }

        public virtual FliRectangle GetBaseExposureArea()
        {
            return new FliRectangle();
        }

        public static string GetClockFromTime(long value, FliProCamera.ExposureUnits_Enum unit)
        {
            string result = String.Empty;
            long factor = 1;
            long hours = 0;
            long mins = 0;
            long secs = 0;
            string valueFormat = "000";
            double newValue = value;
            int divisorFactor = 1;

            switch (unit)
            {
                case ExposureUnits_Enum.units_uSecs:
                    factor = 1;
                    valueFormat = "000000";
                    divisorFactor = 1000;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = MILLISECS_PER_SEC;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = MILLISECS_PER_MIN;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = MILLISECS_PER_HOUR;
                    break;
            }
            newValue *= factor;

            // Value is now in millisecs
            hours = (long)(newValue / (MILLISECS_PER_HOUR * divisorFactor));
            newValue -= hours * (MILLISECS_PER_HOUR * divisorFactor);
            mins = (long)(newValue / (MILLISECS_PER_MIN * divisorFactor));
            newValue -= mins * (MILLISECS_PER_MIN * divisorFactor);
            secs = (long)(newValue / (MILLISECS_PER_SEC * divisorFactor));
            newValue -= secs * (MILLISECS_PER_SEC * divisorFactor);
            result = hours.ToString("00") + ":" + mins.ToString("00") + ":" + secs.ToString("00") + "." + newValue.ToString(valueFormat);
            return result;
        }

        /// <summary>
        /// Read the Cooler Duty Cycle
        /// </summary>
        /// <param name="DutyCycle">The percentage at which the Cooler is running</param>
        /// <returns></returns>
        public virtual int GetCoolerDutyCycle(ref int DutyCycle)
        {
            int status = 0;
            return status;
        }

        public bool GetEnableTemperatureTimer()
        {
            return TemperatureStatusTimer.Enabled;
        }

        public virtual int GetHDREnable(ref bool bIsHDR)
        {
            int status = 0;
            bIsHDR = true;
            return status;
        }

        public virtual FliRectangle GetExposureArea()
        {
            return new FliRectangle();
        }

        
        public virtual int GetTemperatureSetPoint(ref double SetPoint)
        {
            SetPoint = 25;
            return 0;
        }

        public static long GetTimeFromUnits(long value, FliProCamera.ExposureUnits_Enum unit)
        {
            long factor = 1;
            long newValue = value;

            switch (unit)
            {
                case ExposureUnits_Enum.units_uSecs:
                    factor = 1;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1000;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = MILLISECS_PER_SEC * 1000;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = MILLISECS_PER_MIN * 1000;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = (long)MILLISECS_PER_HOUR * 1000;
                    break;
            }
            newValue *= factor;

            // Value is now in millisecs
            return newValue;
        }

        public virtual uint GetFWRevision()
        {
            if ((fliProCameraManager.ActiveCamera.Type == CameraType.GEN1) ||
                (fliProCameraManager.ActiveCamera.Type == CameraType.GEN2))
            {
                return fliProCameraManager.ActiveCamera.GetFWRevision();
            }
            return 0;
        }

        public virtual string GetLibVersion()
        {
            return String.Empty;
        }

        public virtual void GetDeviceVersion(ref string firmwareVersion, ref string fpgaVersion, ref string controllerVersion)
        {
            firmwareVersion = "0.0.0";
            fpgaVersion = "0.0.0";
        }

        /// <summary>
        /// Gets all of the Temperatures on the camera
        /// </summary>
        /// <param name="AmbientTemp">The current Ambient temperature on the sensor</param>
        /// <param name="BaseTemp">The current Base temperature on the sensor</param>
        /// <param name="CoolerTemp">The temperature of the Cooler unit on the camera</param>
        /// <returns></returns>
        public virtual int GetTemperatures(ref double AmbientTemp, ref double BaseTemp, ref double CoolerTemp)
        {
            int status = 0;
            return status;
        }

        public Bitmap BitmapFromSource(BitmapSource bitmapsource)
        {
            Bitmap bitmap;
            using (var outStream = new MemoryStream())
            {
                BitmapEncoder enc = new BmpBitmapEncoder();
                enc.Frames.Add(BitmapFrame.Create(bitmapsource));
                enc.Save(outStream);
                bitmap = new Bitmap(outStream);
            }
            return bitmap;
        }

        public virtual bool CheckGPS()
        {
            return false;
        }

        public FliFrame GetThumbnailFrame(byte[] thumbnailArray)
        {
            byte[] rawDataArray = thumbnailArray;
            //byte[] rawDataArray = Decompress(thumbnailArray);
            FliFrame thumbnailFliFrame = new FliFrame(this);
            thumbnailFliFrame.CameraID = SerialNumber;
            thumbnailFliFrame.UnscrambleImage = false;
            thumbnailFliFrame.IsProcessed = false;
            thumbnailFliFrame.IsQueued = false;
            thumbnailFliFrame.IsTestFrame = false;
            thumbnailFliFrame.TestName = String.Empty;
            thumbnailFliFrame.IsValid = true;

            FliProFrameMetadata fliProFrameMetadata = new FliProFrameMetadata();
            int lineSize = THUMBNAIL_DIMENSION_SIZE * 2;   // There are 2 bytes per pixel 
            int byteSize = lineSize * THUMBNAIL_DIMENSION_SIZE;

            fliProFrameMetadata.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16.GetHashCode();
            fliProFrameMetadata.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16.GetHashCode();
            fliProFrameMetadata.DummyPixelsPerRow = 0;
            fliProFrameMetadata.HDRMode = false;
            fliProFrameMetadata.SetPaddedModel(Model);
            fliProFrameMetadata.SetPaddedSerialNumber(SerialNumber);
            fliProFrameMetadata.HorizontalPixels = THUMBNAIL_DIMENSION_SIZE;
            fliProFrameMetadata.VerticalPixels = THUMBNAIL_DIMENSION_SIZE;
            fliProFrameMetadata.ImageStartRow = 0;
            fliProFrameMetadata.ImageStopRow = THUMBNAIL_DIMENSION_SIZE - 1;
            fliProFrameMetadata.ImageStartColumn = 0;
            fliProFrameMetadata.ImageStopColumn = THUMBNAIL_DIMENSION_SIZE - 1;
            fliProFrameMetadata.PreDummyRows = 0;
            fliProFrameMetadata.PostDummyRows = 0;
            fliProFrameMetadata.ExposureTime = 0;
            fliProFrameMetadata.SensorColdTemp = Convert.ToInt32(BaseTemperature);
            fliProFrameMetadata.SensorTemp = Convert.ToInt32(Temperature);
            fliProFrameMetadata.HDRMode = false;
            fliProFrameMetadata.HorizontalBinning = 1;
            fliProFrameMetadata.VerticalBinning = 1;
            fliProFrameMetadata.PixelOrderedImage = true;

            fliProFrameMetadata.FrameNumber = 1;
            fliProFrameMetadata.FirmwareVersion = GetFWRevision();

            thumbnailFliFrame.FrameMetaData = fliProFrameMetadata;

            byte[] rawMetaData = fliProFrameMetadata.GetByteArray();

            thumbnailFliFrame.RawDataLength = (uint)(byteSize + fliProFrameMetadata.MetadataLength);
            thumbnailFliFrame.RawData = new byte[thumbnailFliFrame.RawDataLength];

            // First insert the MetaData into the RawData array
            Buffer.BlockCopy(rawMetaData, 0, thumbnailFliFrame.RawData, 0, fliProFrameMetadata.MetadataLength);

            //
            // NOTE: Creating the Thumbnail converts our Gray16 image to a Gray8 image. This leaves us missing half the data.
            //       We will need to expand each pixel by multiplying by 16 because we know we started with a 12 bit value.
            //
            ushort[] expandedArray = new ushort[rawDataArray.Length];

            for (int i = 0; i < rawDataArray.Length; i++)
            {
                expandedArray[i] = (ushort)(rawDataArray[i] * 16);
            }

            // Now copy the expanded data into the RawData array
            Buffer.BlockCopy(expandedArray, 0, thumbnailFliFrame.RawData, fliProFrameMetadata.MetadataLength, byteSize);

            LastFrameQueued = thumbnailFliFrame;

            return thumbnailFliFrame;
        }

        public byte[] GetThumbnailBytes(FliFrame fliFrame)
        {
            byte[] bitmapArray = fliFrame.GetTransformedPixelArrayForDisplay(FliFrame.FliImageType.LowGain);
            ImageWidth = fliFrame.GetPixelArrayWidth(FliFrame.FliImageType.LowGain);
            ImageHeight = fliFrame.GetPixelArrayHeight(FliFrame.FliImageType.LowGain);
            int stride = fliFrame.FrameAreaWHTotal.W;
            var pixelFormats = PixelFormats.Gray8;

            BitmapSource source = (((bitmapArray == null) || (bitmapArray.Length == 0)) ? null : BitmapSource.Create(ImageWidth,
                                                              ImageHeight,
                                                              96,
                                                              96,
                                                              pixelFormats,
                                                              null,
                                                              bitmapArray,
                                                              stride));
            TransformedBitmap thumbnailBmp = new TransformedBitmap(source, new ScaleTransform(((double)THUMBNAIL_DIMENSION_SIZE) / ImageWidth, ((double)THUMBNAIL_DIMENSION_SIZE) / ImageHeight));
            Bitmap bitmap;
            using (var outStream = new MemoryStream())
            {
                BitmapEncoder enc = new BmpBitmapEncoder();
                enc.Frames.Add(BitmapFrame.Create(thumbnailBmp));
                enc.Save(outStream);
                bitmap = new Bitmap(outStream);
            }

            byte[] thumbnailArray = new byte[MAX_THUMBNAIL_SIZE];
            BitmapData bitmap_data = bitmap.LockBits(new Rectangle(0, 0, THUMBNAIL_DIMENSION_SIZE, THUMBNAIL_DIMENSION_SIZE),
                ImageLockMode.ReadOnly, System.Drawing.Imaging.PixelFormat.Format8bppIndexed);
            Marshal.Copy(bitmap_data.Scan0, thumbnailArray, 0, MAX_THUMBNAIL_SIZE);
            bitmap.UnlockBits(bitmap_data);

            return thumbnailArray;
        }

        public string GetValidFilename(string folder, string filename, string extension)
        {
            int fileIndex = 1;
            string newFileName = filename;
            DirectoryInfo di = new DirectoryInfo(folder);
            FileInfo[] TXTFiles = null;

            try
            {
                TXTFiles = di.GetFiles(filename + "*" + extension);

                List<int> indexList = new List<int>();
                for (int i = 0; i < TXTFiles.Length; i++)
                {
                    string tempName = TXTFiles[i].Name.Substring(filename.Length, TXTFiles[i].Name.Length - (filename.Length + 4)); // subtract off ".bin"
                    try
                    {
                        indexList.Add(Convert.ToInt32(tempName));
                    }
                    catch { }
                }
                while (indexList.Contains(fileIndex))
                {
                    fileIndex++;
                }
            }
            catch { }

            return folder + "\\" + newFileName + ((TXTFiles != null) && (TXTFiles.Length > 0) ? fileIndex.ToString() : String.Empty) + extension;
        }

        private int originalStatusLength = 0;
        public void PublishStatus(string text, bool isAppend)
        {
            if (isAppend && !String.IsNullOrEmpty(CameraStatusText))
            {
                CameraStatusText = CameraStatusText.Substring(0, originalStatusLength);
                CameraStatusText += text;
            }
            else
            {
                CameraStatusText = text;
                originalStatusLength = text.Length;
            }
        }

        /// <summary>
        /// Reads the Temperature of the given channel
        /// </summary>
        /// <param name="channel">The channel from which the temperature is to be read</param>
        /// <param name="temp">The returned temperature</param>
        /// <returns></returns>
        public virtual int ReadTemperature(int channel, ref double temp)
        {
            int status = 0;
            return status;
        }

        public void ResetToFactoryDefaults()
        {
            OnConnect();
        }

        public void SetDefaultFliFrameCountList()
        {
            FliFrameCountList = new System.Collections.ObjectModel.ObservableCollection<FliFrameCountNameValueClass>();
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Single", FliFrameCount.SINGLE));
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Multiple", FliFrameCount.MULTIPLE));
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Continuous", FliFrameCount.CONTINUOUS));

            FrameCountType = FliFrameCount.SINGLE;
        }

        public void SetDefaultFliFrameTypeList()
        {
            FliFrameTypeList = new System.Collections.ObjectModel.ObservableCollection<FliFrameTypeNameValueClass>();
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Normal", FliFrameType.NORMAL));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Flat", FliFrameType.FLAT));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Dark", FliFrameType.DARK));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Flood", FliFrameType.FLOOD));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("RBI_Flush", FliFrameType.RBI_FLUSH));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Bias", FliFrameType.BIAS));

            FrameType = FliFrameType.NORMAL;
        }

        public void SetDefaultGainTables()
        {
            DNETableLG = new ObservableCollection<SensorGain>();
            DNETableLG.Add(new SensorGain(0, 0.0201, 0));
            DNETableLG.Add(new SensorGain(4, 0.0401, 1));
            DNETableLG.Add(new SensorGain(1, 0.0576, 2));
            DNETableLG.Add(new SensorGain(2, 0.0788, 3));
            DNETableLG.Add(new SensorGain(3, 0.1162, 4));
            DNETableLG.Add(new SensorGain(5, 0.1128, 5));
            DNETableLG.Add(new SensorGain(6, 0.152, 6));
            DNETableLG.Add(new SensorGain(7, 0.206, 7));

            DNETableLGFactory = new ObservableCollection<SensorGain>();
            DNETableLGFactory.Add(new SensorGain(0, 0.0201, 0));
            DNETableLGFactory.Add(new SensorGain(4, 0.0401, 1));
            DNETableLGFactory.Add(new SensorGain(1, 0.0576, 2));
            DNETableLGFactory.Add(new SensorGain(2, 0.0788, 3));
            DNETableLGFactory.Add(new SensorGain(3, 0.1162, 4));
            DNETableLGFactory.Add(new SensorGain(5, 0.1128, 5));
            DNETableLGFactory.Add(new SensorGain(6, 0.152, 6));
            DNETableLGFactory.Add(new SensorGain(7, 0.206, 7));

            DNETableHG = new ObservableCollection<SensorGain>();
            DNETableHG.Add(new SensorGain(0, 0.1316, 0));
            DNETableHG.Add(new SensorGain(4, 0.2633, 1));
            DNETableHG.Add(new SensorGain(1, 0.3897, 2));
            DNETableHG.Add(new SensorGain(2, 0.5717, 3));
            DNETableHG.Add(new SensorGain(3, 0.9178, 4));
            DNETableHG.Add(new SensorGain(5, 0.9088, 5));
            DNETableHG.Add(new SensorGain(6, 1.2332, 6));
            DNETableHG.Add(new SensorGain(7, 1.8022, 7));

            DNETableHGFactory = new ObservableCollection<SensorGain>();
            DNETableHGFactory.Add(new SensorGain(0, 0.1316, 0));
            DNETableHGFactory.Add(new SensorGain(4, 0.2633, 1));
            DNETableHGFactory.Add(new SensorGain(1, 0.3897, 2));
            DNETableHGFactory.Add(new SensorGain(2, 0.5717, 3));
            DNETableHGFactory.Add(new SensorGain(3, 0.9178, 4));
            DNETableHGFactory.Add(new SensorGain(5, 0.9088, 5));
            DNETableHGFactory.Add(new SensorGain(6, 1.2332, 6));
            DNETableHGFactory.Add(new SensorGain(7, 1.8022, 7));

            LowGainTable = new ObservableCollection<SensorGain>();
            LowGainTable.Add(new SensorGain(0, 660, 0));
            LowGainTable.Add(new SensorGain(4, 1290, 1));
            LowGainTable.Add(new SensorGain(1, 1850, 2));
            LowGainTable.Add(new SensorGain(2, 2490, 3));
            LowGainTable.Add(new SensorGain(3, 3680, 4));
            LowGainTable.Add(new SensorGain(5, 3700, 5));
            LowGainTable.Add(new SensorGain(6, 4950, 6));
            LowGainTable.Add(new SensorGain(7, 7250, 7));

            HighGainTable = new ObservableCollection<SensorGain>();
            HighGainTable.Add(new SensorGain(0, 660, 0));
            HighGainTable.Add(new SensorGain(4, 1290, 1));
            HighGainTable.Add(new SensorGain(1, 1850, 2));
            HighGainTable.Add(new SensorGain(2, 2490, 3));
            HighGainTable.Add(new SensorGain(3, 3680, 4));
            HighGainTable.Add(new SensorGain(5, 3700, 5));
            HighGainTable.Add(new SensorGain(6, 4950, 6));
            HighGainTable.Add(new SensorGain(7, 7250, 7));

            RestoreMergeSettingsFromFactorDefaults();

            GainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];
            ManufacturersDefaultGainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];

            for (int l = 0; l < DNETableLG.Count; l++)
            {
                int epcLowIndex = GetFactoryElectronsPerCountLowIndex(DNETableLG[l].Index);
                if (epcLowIndex >= 0)
                {
                    DNETableLG[l].Gain = FactoryElectronsPerCountLowGainList[epcLowIndex].ElectronsPerCount;
                }

                for (int h = 0; h < DNETableHG.Count; h++)
                {
                    if (l == 0)
                    {
                        int epcHighIndex = GetFactoryElectronsPerCountHighIndex(DNETableHG[h].Index);
                        if (epcHighIndex >= 0)
                        {
                            DNETableHG[h].Gain = FactoryElectronsPerCountHighGainList[epcHighIndex].ElectronsPerCount;
                        }
                    }

                    int index = GetMergeLineOffsetIndex(DNETableHG[h].Index, DNETableLG[l].Index);

                    // These are filled in from the camera
                    if (index < 0)
                    {
                        // Divide the Low gain by the High gain because the numbers are in 1/x form
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = DNETableLG[l].Gain / DNETableHG[h].Gain;
                        // There is no easy way to calculate the slope so I just figured out an equation that yields a "reasonable" value.
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = -10;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 120);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = 5;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 125);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = 20;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 130);
                    }
                    else
                    {
                        // Use the current default
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = MergeLineOffsetList[index].GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = MergeLineOffsetList[index].Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = MergeLineOffsetList[index].LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = MergeLineOffsetList[index].Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = MergeLineOffsetList[index].LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = MergeLineOffsetList[index].Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                    }

                    GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;

                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;
                }
            }

            LowGainValue = LowGainTable[0];
            HighGainValue = HighGainTable[0];
        }

        public void SetDefaultGainTablesCobaltDC230(bool initialize)
        {
            if (initialize)
            {
                DNETableLG = new ObservableCollection<SensorGain>();
                DNETableLG.Add(new SensorGain(0, 1.000, 0));

                DNETableLGFactory = new ObservableCollection<SensorGain>();
                DNETableLGFactory.Add(new SensorGain(0, 1.000, 0));

                DNETableHG = new ObservableCollection<SensorGain>();
                DNETableHG.Add(new SensorGain(0, 1.000, 0));

                DNETableHGFactory = new ObservableCollection<SensorGain>();
                DNETableHGFactory.Add(new SensorGain(0, 1.000, 0));

                LowGainTable = new ObservableCollection<SensorGain>();
                LowGainTable.Add(new SensorGain(0, 1000, 0));
                LowGainValue = LowGainTable[0];

                HighGainTable = new ObservableCollection<SensorGain>();
                HighGainTable.Add(new SensorGain(0, 1000, 0));
                HighGainValue = HighGainTable[0];
            }
        }

        public void SetDefaultGainTablesKepler0400(bool initialize)
        {
            if (initialize)
            {
                DNETableLG = new ObservableCollection<SensorGain>();
                DNETableLG.Add(new SensorGain(0, 47.8469, 0));
                DNETableLG.Add(new SensorGain(4, 24.9377, 1));
                DNETableLG.Add(new SensorGain(1, 17.3611, 2));
                DNETableLG.Add(new SensorGain(2, 12.6904, 3));
                DNETableLG.Add(new SensorGain(3, 8.6059, 4));
                DNETableLG.Add(new SensorGain(5, 8.8652, 5));
                DNETableLG.Add(new SensorGain(6, 6.5789, 6));
                DNETableLG.Add(new SensorGain(7, 4.8544, 7));

                DNETableLGFactory = new ObservableCollection<SensorGain>();
                DNETableLGFactory.Add(new SensorGain(0, 47.8469, 0));
                DNETableLGFactory.Add(new SensorGain(4, 24.9377, 1));
                DNETableLGFactory.Add(new SensorGain(1, 17.3611, 2));
                DNETableLGFactory.Add(new SensorGain(2, 12.6904, 3));
                DNETableLGFactory.Add(new SensorGain(3, 8.6059, 4));
                DNETableLGFactory.Add(new SensorGain(5, 8.8652, 5));
                DNETableLGFactory.Add(new SensorGain(6, 6.5789, 6));
                DNETableLGFactory.Add(new SensorGain(7, 4.8544, 7));

                DNETableHG = new ObservableCollection<SensorGain>();
                DNETableHG.Add(new SensorGain(0, 7.5988, 0));
                DNETableHG.Add(new SensorGain(4, 3.7979, 1));
                DNETableHG.Add(new SensorGain(1, 2.5661, 2));
                DNETableHG.Add(new SensorGain(2, 1.7492, 3));
                DNETableHG.Add(new SensorGain(3, 1.0896, 4));
                DNETableHG.Add(new SensorGain(5, 1.1004, 5));
                DNETableHG.Add(new SensorGain(6, 0.8109, 6));
                DNETableHG.Add(new SensorGain(7, 0.5549, 7));

                DNETableHGFactory = new ObservableCollection<SensorGain>();
                DNETableHGFactory.Add(new SensorGain(0, 7.5988, 0));
                DNETableHGFactory.Add(new SensorGain(4, 3.7979, 1));
                DNETableHGFactory.Add(new SensorGain(1, 2.5661, 2));
                DNETableHGFactory.Add(new SensorGain(2, 1.7492, 3));
                DNETableHGFactory.Add(new SensorGain(3, 1.0896, 4));
                DNETableHGFactory.Add(new SensorGain(5, 1.1004, 5));
                DNETableHGFactory.Add(new SensorGain(6, 0.8109, 6));
                DNETableHGFactory.Add(new SensorGain(7, 0.5549, 7));

                LowGainTable = new ObservableCollection<SensorGain>();
                LowGainTable.Add(new SensorGain(0, 660, 0));
                LowGainTable.Add(new SensorGain(4, 1290, 1));
                LowGainTable.Add(new SensorGain(1, 1850, 2));
                LowGainTable.Add(new SensorGain(2, 2490, 3));
                LowGainTable.Add(new SensorGain(3, 3680, 4));
                LowGainTable.Add(new SensorGain(5, 3700, 5));
                LowGainTable.Add(new SensorGain(6, 4950, 6));
                LowGainTable.Add(new SensorGain(7, 7250, 7));
                LowGainValue = LowGainTable[0];

                HighGainTable = new ObservableCollection<SensorGain>();
                HighGainTable.Add(new SensorGain(0, 660, 0));
                HighGainTable.Add(new SensorGain(4, 1290, 1));
                HighGainTable.Add(new SensorGain(1, 1850, 2));
                HighGainTable.Add(new SensorGain(2, 2490, 3));
                HighGainTable.Add(new SensorGain(3, 3680, 4));
                HighGainTable.Add(new SensorGain(5, 3700, 5));
                HighGainTable.Add(new SensorGain(6, 4950, 6));
                HighGainTable.Add(new SensorGain(7, 7250, 7));
                HighGainValue = HighGainTable[0];
            }

            RestoreMergeSettingsFromFactorDefaults();

            if (ActiveElectronsPerCountLow == 0)
            {
                // Then we haven't initialized this camera yet so do it now
                MergeOverride = false;
                HighGainIndex = 7;
                LowGainIndex = 4;
                ActiveElectronsPerCountHigh = 0.5549;
                ActiveElectronsPerCountLow = 24.9377;
                Temperature1 = -10;
                Temperature2 = 5;
                Temperature3 = 20;
                LineOffset1 = 3500;
                LineOffset2 = 3500;
                LineOffset3 = 3500;
                GainRatio = 40;
                ADUsPerElectron = 1;
                BlackLevelHigh = 15750;
                BlackLevelLow = 15580;
                SaveMergeSettingsAsFactorDefaults();
            }
            GainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];
            ManufacturersDefaultGainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];

            // Because of the way we handle Gain indecies on the 400 we get back the array index in the Metadata.
            // That being the case we need to use the array index to populate the GainOffsetPairArray
            for (int l = 0; l < DNETableLG.Count; l++)
            {
                int epcLowIndex = GetFactoryElectronsPerCountLowIndex(DNETableLG[l].Index);
                if (epcLowIndex >= 0)
                {
                    DNETableLG[l].Gain = FactoryElectronsPerCountLowGainList[epcLowIndex].ElectronsPerCount;
                }

                for (int h = 0; h < DNETableHG.Count; h++)
                {
                    if (l == 0)
                    {
                        int epcHighIndex = GetFactoryElectronsPerCountHighIndex(DNETableHG[h].Index);
                        if (epcHighIndex >= 0)
                        {
                            DNETableHG[h].Gain = FactoryElectronsPerCountHighGainList[epcHighIndex].ElectronsPerCount;
                        }
                    }

                    int index = GetMergeLineOffsetIndex(DNETableHG[h].Index, DNETableLG[l].Index);

                    // These are filled in from the camera
                    if (index < 0)
                    {
                        // Divide the Low gain by the High gain because the numbers are in 1/x form
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = Math.Round(DNETableLG[l].Gain / DNETableHG[h].Gain, 4);
                        // There is no easy way to calculate the slope so I just figured out an equation that yields a "reasonable" value.
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = -10;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 120);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = 5;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 125);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = 20;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 130);
                    }
                    else
                    {
                        // Use the current default
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = MergeLineOffsetList[index].GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = MergeLineOffsetList[index].Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = MergeLineOffsetList[index].LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = MergeLineOffsetList[index].Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = MergeLineOffsetList[index].LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = MergeLineOffsetList[index].Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                    }

                    GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;

                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;
                }
            }
        }

        public void SetDefaultGainTablesKepler4040(bool initialize)
        {
            if (initialize)
            {
                DNETableLG = new ObservableCollection<SensorGain>();
                DNETableHG = new ObservableCollection<SensorGain>();
                LowGainTable = new ObservableCollection<SensorGain>();
                HighGainTable = new ObservableCollection<SensorGain>();

                RestoreMergeSettingsFromFactorDefaults();

                for (int i = 0; i < 64; i++)
                {
                    double value = 0;
                    if ((i % 2) == 0)
                    {
                        value = 1.0 + (i / 2) * 0.5;
                    }
                    else
                    {
                        value = 0.2 + ((i + 1) / 2) * 0.2;
                    }
                    LowGainTable.Add(new SensorGain(i, value, i));
                    HighGainTable.Add(new SensorGain(i, value, i));

                    // We only have a few values from the Sensor manufacturer so use them
                    // and default the rest to Low Index 25 (2.800) and High Index 62 (16.500).
                    double LowElectronsPerCount = ActiveElectronsPerCountLow;
                    double HighElectronsPerCount = ActiveElectronsPerCountHigh;

                    switch ((int)(value * 1000))
                    {
                        case 2800:
                            LowElectronsPerCount = 18.6916;
                            HighElectronsPerCount = 18.6916;
                            break;
                        case 5200:
                            LowElectronsPerCount = 10;
                            HighElectronsPerCount = 10;
                            break;
                        case 6600:
                            LowElectronsPerCount = 8;
                            HighElectronsPerCount = 8;
                            break;
                        case 13500:
                            LowElectronsPerCount = 1;
                            HighElectronsPerCount = 1;
                            break;
                        case 14000:
                            LowElectronsPerCount = 4;
                            HighElectronsPerCount = 4;
                            break;
                        case 16500:
                            LowElectronsPerCount = 0.8194;
                            HighElectronsPerCount = 0.8194;
                            break;
                    }
                    DNETableLG.Add(new SensorGain(i, LowElectronsPerCount, i));
                    DNETableHG.Add(new SensorGain(i, HighElectronsPerCount, i));
                    DNETableLGFactory.Add(new SensorGain(i, LowElectronsPerCount, i));
                    DNETableHGFactory.Add(new SensorGain(i, HighElectronsPerCount, i));
                }
            }

            if (ActiveElectronsPerCountLow == 0)
            {
                // Then we haven't initialized this camera yet so do it now
                MergeOverride = false;
                HighGainIndex = 62;
                LowGainIndex = 25;
                ActiveElectronsPerCountHigh = 0.8194;
                ActiveElectronsPerCountLow = 18.6916;
                Temperature1 = -10;
                Temperature2 = 5;
                Temperature3 = 20;
                LineOffset1 = 1700;
                LineOffset2 = 1750;
                LineOffset3 = 1800;
                GainRatio = 20.1;
                ADUsPerElectron = 1;
                BlackLevelHigh = 15750;
                BlackLevelLow = 15580;
                SaveMergeSettingsAsFactorDefaults();
            }

            GainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];
            ManufacturersDefaultGainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];

            for (int l = 0; l < DNETableLG.Count; l++)
            {
                int epcLowIndex = GetFactoryElectronsPerCountLowIndex(DNETableLG[l].Index);
                if (epcLowIndex >= 0)
                {
                    DNETableLG[l].Gain = FactoryElectronsPerCountLowGainList[epcLowIndex].ElectronsPerCount;
                }

                for (int h = 0; h < DNETableHG.Count; h++)
                {
                    if (l == 0)
                    {
                        int epcHighIndex = GetFactoryElectronsPerCountHighIndex(DNETableHG[h].Index);
                        if (epcHighIndex >= 0)
                        {
                            DNETableHG[h].Gain = FactoryElectronsPerCountHighGainList[epcHighIndex].ElectronsPerCount;
                        }
                    }

                    int index = GetMergeLineOffsetIndex(DNETableHG[h].Index, DNETableLG[l].Index);

                    // These are filled in from the camera
                    if (index < 0)
                    {
                        // There is no default for this pair so use the current settings
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = LineOffset3;
                    }
                    else
                    {
                        // Use the current default
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = MergeLineOffsetList[index].GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = MergeLineOffsetList[index].Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = MergeLineOffsetList[index].LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = MergeLineOffsetList[index].Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = MergeLineOffsetList[index].LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = MergeLineOffsetList[index].Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                    }

                    GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;

                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;
                }
            }

            LowGainValue = GetGainValueFromIndex(LowGainTable, 25);
            HighGainValue = GetGainValueFromIndex(HighGainTable, 62);
        }

        public void UpdateDefaultGainTablesKepler4040()
        {
            DNETableLG = new ObservableCollection<SensorGain>();
            DNETableHG = new ObservableCollection<SensorGain>();

            for (int i = 0; i < 64; i++)
            {
                // We only have these two values from the Sensor manufacturer 
                // which are for Low Index 25 (2.800) and High Index 62 (16.500).
                // So just use them for all the Gains.
                DNETableLG.Add(new SensorGain(i, ActiveElectronsPerCountLow, i));
                DNETableHG.Add(new SensorGain(i, ActiveElectronsPerCountHigh, i));
                DNETableLGFactory.Add(new SensorGain(i, ActiveElectronsPerCountLow, i));
                DNETableHGFactory.Add(new SensorGain(i, ActiveElectronsPerCountHigh, i));
            }

            RestoreMergeSettingsFromFactorDefaults();

            GainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];
            ManufacturersDefaultGainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];

            for (int l = 0; l < DNETableLG.Count; l++)
            {
                int epcLowIndex = GetFactoryElectronsPerCountLowIndex(DNETableLG[l].Index);
                if (epcLowIndex >= 0)
                {
                    DNETableLG[l].Gain = FactoryElectronsPerCountLowGainList[epcLowIndex].ElectronsPerCount;
                }

                for (int h = 0; h < DNETableHG.Count; h++)
                {
                    if (l == 0)
                    {
                        int epcHighIndex = GetFactoryElectronsPerCountHighIndex(DNETableHG[h].Index);
                        if (epcHighIndex >= 0)
                        {
                            DNETableHG[h].Gain = FactoryElectronsPerCountHighGainList[epcHighIndex].ElectronsPerCount;
                        }
                    }

                    int index = GetMergeLineOffsetIndex(DNETableHG[h].Index, DNETableLG[l].Index);

                    // These are filled in from the camera
                    if (index < 0)
                    {
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = DNETableHG[h].Gain / DNETableLG[l].Gain;
                        // There is no easy way to calculate the slope so I just figured out an equation that yields a "reasonable" value.
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = -10;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 120);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = 5;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 125);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = 20;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 130);
                    }
                    else
                    {
                        // Use the current default
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = MergeLineOffsetList[index].GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = MergeLineOffsetList[index].Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = MergeLineOffsetList[index].LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = MergeLineOffsetList[index].Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = MergeLineOffsetList[index].LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = MergeLineOffsetList[index].Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                    }

                    GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;

                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;
                }
            }
        }

        public void SetDefaultSupportedBinRatios()
        {
            SupportedBinRatios.Add(new BinRatio(0, 1, 1));
            BinRatioValue = SupportedBinRatios[0];
        }

        public virtual void SetDefaultSupportedPixelDepths()
        {
            SupportedPixelDepths.Add(12);
            PixelDepth = SupportedPixelDepths[0];
        }
        
        public void SetDefaultSupportedModes()
        {
            SupportedModes = new ObservableCollection<NameValuePair>();
            SupportedModes.Add(new NameValuePair("Rolling LDR - LDC", 1));
            SupportedModes.Add(new NameValuePair("Rolling HDR - LDC", 3));
            SupportedModes.Add(new NameValuePair("Rolling LDR", 0));
            SupportedModes.Add(new NameValuePair("Rolling HDR", 2));
            Mode = 0;
            CameraMode = SupportedModes[Mode].Name;
        }

        public void SetEnabled(bool enabled)
        {
            IsEnabled = enabled;
        }

        public void SetForcePropertyChange(bool isForce)
        {
            if (fliProCameraManager == null)
            {
                fliProCameraManager = FliProCameraManager.Instance;
            }

            ForcePropertyChange = isForce;
        }

        public virtual void SetupOverScan(int overScanLength)
        {
        }

        //-------------------------------------------
        /// <summary>
        /// Start a frame capture sequence
        /// </summary>
        /// <param name="wait">wait for capture to complete if true (default = false)</param>
        /// <returns>false on any error</returns>
        public bool StartCapture(bool wait = false)
        {
            bool ok = true;

            LogEvent("Entering StartCapture");
            while (IsCapturing && !IsAborted)
            {
                LogEvent("StartCapture had to wait for IsCapturing flag");
                Thread.Sleep(2);
            }

            if (!IsCapturing)
            {
                // We need to officially Stop the capture thread here so that we can capture again.
                // If we don't there will be a deadlock in StartStopJob.Start because StartStopJob._stopped remains true.
                StopCapture();

                // Need to wait for the lock to set the IsCapturing flag so that we know the Temperature Request Timer is not active.
                SetIsCapturingLocked(true);
                IsAborted = false;
                MinimumFramePoolCount = MaximumFrameStorage * 2;

                // Make sure we have enough frame buffers to start
                ReplenishFramePool(true);

                if (TrackingFrameCount > 0)
                {
                    ReplenishTrackingFramePool();
                }

                FramesCaptured = 0;

                LogEvent("Start frame capture");

                if (ok)
                {
                    try
                    {
                        // Start Frame Capture thread
                        FrameCaptureJob.Start();
                        if(wait)
                        {
                            //-=-= FIXME: Add wait timeout, stop if exceeded
                            while (!FrameCaptureJob.Wait(TimeSpan.FromSeconds(1)))
                            {
                            }
                        }
                    }
                    catch
                    {
                        ok = false;
                    }
                }

                LogEvent("Start frame capture Complete");

                IsCapturing = ok;
            }
            return ok;
        }


        //-------------------------------------------
        /// <summary>
        /// Stop current frame capture sequence, if any)
        /// </summary>
        public void StopCapture()
        {
            if (IsCapturing == false)
            {
                FrameCaptureJob.Stop();
            }
        }

        private int tempCounter = 0;
        private ObservableCollection<SensorGain> DNETableLGManufacturer;

        private void TemperatureStatusTimerEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            int status = 0;
            string TempString = String.Empty;

            // 
            // We want to only ask for temperature if the image is not capturing. 
            // And we want to prevent asking for image capture while we are getting the temperature. 
            // That means we have to lock during this entire routine but we only have to lock the image capture when setting the IsCapturing flag.
            lock (ImageCaptureLock)
            {
                try
                {
                    if (IsCapturing && !ReadDuringExposureOn)
                    {
                        // We don't want to read the Temperature register while the camera is capturing because this causes the sensor to fail.
                        return;
                    }

                    if ((IsConnected == false) || (ReadErrorEncountered == true))
                    {
                        TempString = "Temp: N/A";
                    }
                    else
                    {
                        double ambientTemp = 0;
                        double baseTemp = 0;
                        double coolerTemp = 0;

                        status = GetTemperatures(ref ambientTemp, ref baseTemp, ref coolerTemp);
                        tempCounter++;
                        Temperature = coolerTemp;
                        TemperatureText = Temperature.ToString("#0.0");

                        if (status < 0)
                        {
                            // Turn off checking for temperature on the timer
                            // TemperatureStatusTimer.Enabled = false;         // We need to keep this on so we can read the Xylinx chip temperature

                            // We don't care if we get an error because we are shutting down
                            try
                            {
                                // I don't think we want to shut down the camera just because we had a status error
                                //CloseDevice();
                            }
                            catch
                            {
                                // No harm in continuing to read but post the error
                                //ReadErrorEncountered = true;
                            }
                        }
                        else
                        {
                            if (fliProCameraManager.IsCheckedAutoForceTraining && !FirstTraining)
                            {
                                if (!fliProCameraManager.AutoForceTrainingTargetReached && ((Temperature > TargetTemperature - 0.2) && (Temperature < TargetTemperature + 0.2)))
                                {
                                    // We have reached our target so set the Target reached flag and run one last Forced Training
                                    fliProCameraManager.AutoForceTrainingTargetReached = true;
                                    fliProCameraManager.AutoForceTrainingLastTargetTemperature = TargetTemperature;
                                    ForceTraining(this);
                                }
                                else if (!fliProCameraManager.AutoForceTrainingTargetReached &&
                                    ((Temperature > fliProCameraManager.AutoForceTrainingLastTargetTemperature - fliProCameraManager.AutoForceRangeTemperature) &&
                                    (Temperature < fliProCameraManager.AutoForceTrainingLastTargetTemperature + fliProCameraManager.AutoForceRangeTemperature)))
                                {
                                    // We haven't reached our target but we are still within AutoForceRangeTemperature so don't ForceTraining 
                                    if (CameraStatusText.Contains("Force Training"))
                                    {
                                        PublishStatus("", false);    // Clear the last status message
                                    }
                                }
                                else if ((Temperature < fliProCameraManager.AutoForceTrainingLastTargetTemperature - fliProCameraManager.AutoForceRangeTemperature) ||
                                    (Temperature > fliProCameraManager.AutoForceTrainingLastTargetTemperature + fliProCameraManager.AutoForceRangeTemperature))
                                {
                                    // We've moved outside of the range so send ForceTraining
                                    fliProCameraManager.AutoForceTrainingTargetReached = false;
                                    fliProCameraManager.AutoForceTrainingLastTargetTemperature = Temperature;
                                    ForceTraining(this);
                                }
                                else
                                {
                                    // Just clear the status 
                                    if (CameraStatusText.Contains("Force Training"))
                                    {
                                        PublishStatus("", false);    // Clear the last status message
                                    }
                                }
                            }

                            FirstTraining = false;

                            int cooler = 0;
                            status = GetCoolerDutyCycle(ref cooler);

                            if (status == 0)
                            {
                                if (baseTemp > (-100.0))
                                {
                                    TempString = "C/B: " + TemperatureText + " / " + baseTemp.ToString("#0.0") +
                                        "(" + Convert.ToInt16(cooler).ToString() + "%)";
                                }
                                else
                                {
                                    TempString = "CCD: " + TemperatureText + "(" + Convert.ToUInt16(cooler).ToString() + "%)";
                                }
                            }
                            else
                            {
                                TempString = "CCD: " + TemperatureText + " BASE: " + baseTemp.ToString();
                            }
                        }

                        // If the State of the GPS has changed we need to update the display
                        if (CheckGPS())
                        {
                            OnPropertyChanged("GPSUnlockedVisibility");
                            OnPropertyChanged("GPSLockedVisibility");
                            OnPropertyChanged("GPSConnectedVisibility");
                            OnPropertyChanged("GPSLockWidth");
                            OnPropertyChanged("GPSWidth");
                        }
                    }
                }
                catch
                {
                    // Turn off checking for temperature on the timer
                    //TemperatureStatusTimer.Enabled = false;           // We need to keep this on so we can read the Xylinx chip temperature

                    // No harm in continuing to read but post the error
                    //ReadErrorEncountered = true;
                    TempString = "Temp: N/A";
                }

                try
                {
                    if (EngineeringBuild)
                    {
                        double boardTemp = ReadBoardTemp();
                        TempString += ", X: " + boardTemp.ToString("#0.0");
                    }
                }
                catch
                {
                    TempString += ", X: N/A";
                }
                TemperatureStatusText = TempString;
            }
        }

        private bool ImageCaptureActive()
        {
            bool isActive = false;

            lock (ImageCaptureLock)
            {
                isActive = IsCapturing;
            }
            return isActive;
        }


        //-------------------------------------------
        /// <summary>
        /// Default string representation for a camera
        /// </summary>
        public override string ToString()
        {
            if (SerialNumber.Length > 0)
            {
                return Model + GetDeviceType() + ", " + SerialNumber + Speed;
            }
            else
            {
                return Model + GetDeviceType();
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Get the description string for the specified camera mode index
        /// </summary>
        /// <param name="mode">camera mode index</param>
        /// <returns>string or "" if mode index is not supported</returns>
        public virtual string GetCameraModeString(int mode)
        {
            return "";
        }

        public void WriteFile(FliFrame frame, string folder, string NameHint = "", bool forceFileWrite = false)
        {
            // Don't allow writing frame to file when this camera is already from a file
            if (((Type == CameraType.FILE) || (Type == CameraType.GENERATED)) && !forceFileWrite)
            {
                return;
            }

            if (!String.IsNullOrEmpty(folder))
            {
                lock (ImageWriteLock)
                {
                    try
                    {
                        //string frameNumberString = (frame.FrameMetaData == null ? "0000" : frame.FrameMetaData.FrameNumber.ToString("0000")) + "_";
                        //string targetSerialNumber = SerialNumber;
                        //if (SerialNumber.StartsWith("Camera_"))
                        //{
                        //    // This is an image from a file so filter out the extra substrings
                        //    targetSerialNumber = targetSerialNumber.Substring(7);
                        //    int index = targetSerialNumber.LastIndexOf(".");
                        //    targetSerialNumber = targetSerialNumber.Substring(0, index);
                        //}
                        if (fliProCameraManager == null)
                        {
                            fliProCameraManager = FliProCameraManager.Instance;
                        }
                        string fileNameBase = fliProCameraManager.CalculateNameString(frame, NameHint);
                        string fileName = String.Empty;
                        BitmapSource sourceLow = null;
                        BitmapSource sourceHigh = null;
                        BitmapSource sourceMerge = null;
                        List<string> nameList = new List<string>();
                        string meanName = String.Empty;

                        if (IsSaveExposureToFITFileChecked)
                        {
                            if (UseLowGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Low_", FILE_EXTENSION_FIT);
                                nameList.Add(fileName);
                                FITSImageProcessing.SaveFitsFile(fileName, frame, FliFrame.FliImageType.LowGainDummy);
                            }
                            if (UseHighGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "High_", FILE_EXTENSION_FIT);
                                nameList.Add(fileName);
                                FITSImageProcessing.SaveFitsFile(fileName, frame, FliFrame.FliImageType.HighGainDummy);
                            }
                            if (UseMergedImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Merge_", FILE_EXTENSION_FIT);
                                nameList.Add(fileName);
                                FITSImageProcessing.SaveFitsFile(fileName, frame, FliFrame.FliImageType.MergedDummy);
                            }
                        }

                        if (IsSaveExposureToJPEGFileChecked)
                        {
                            if (UseLowGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Low_", FILE_EXTENSION_JPEG);
                                nameList.Add(fileName);
                                sourceLow = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, sourceLow, frame.GetTransformedPixelArray(FliFrame.FliImageType.LowGain));
                            }
                            if (UseHighGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "High_", FILE_EXTENSION_JPEG);
                                sourceHigh = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, sourceHigh, frame.GetTransformedPixelArray(FliFrame.FliImageType.HighGain));
                            }
                            if (UseMergedImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Merge_", FILE_EXTENSION_JPEG);
                                sourceMerge = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, sourceMerge, frame.GetTransformedPixelArray(FliFrame.FliImageType.Merged));
                            }
                        }

                        if (IsSaveExposureToPNGFileChecked)
                        {
                            if (UseLowGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Low_", FILE_EXTENSION_PNG);
                                nameList.Add(fileName);
                                sourceLow = SaveImageFile(frame, fileName, ImageFormat.Png, sourceLow, frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                            }
                            if (UseHighGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "High_", FILE_EXTENSION_PNG);
                                sourceHigh = SaveImageFile(frame, fileName, ImageFormat.Png, sourceHigh, frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                            }
                            if (UseMergedImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Merge_", FILE_EXTENSION_PNG);
                                sourceMerge = SaveImageFile(frame, fileName, ImageFormat.Png, sourceMerge, frame.GetPixelArray(FliFrame.FliImageType.Merged));
                            }
                        }

                        if (IsSaveExposureToBINFileChecked)
                        {
                            // First save the low only as a 16 bit file 
                            if (UseLowGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Low_", FILE_EXTENSION_BIN);
                                nameList.Add(fileName);
                                SaveBINFile(frame, fileName, File_Type.Low);
                            }
                            if (UseHighGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "High_", FILE_EXTENSION_BIN);
                                SaveBINFile(frame, fileName, File_Type.High);
                            }
                            if (UseMergedImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Merge_", FILE_EXTENSION_BIN);
                                SaveBINFile(frame, fileName, File_Type.Merged);
                            }
                        }

                        if (IsSaveExposureToRCDFileChecked || !String.IsNullOrEmpty(NameHint))    // We always want to write the final image from Test routines
                        {
                            // Save the Combined file which is exactly how we received it from the board.
                            fileName = GetValidFilename(folder, fileNameBase + "RCDData_", FILE_EXTENSION_RCD);
                            nameList.Add(fileName);
                            SaveBINFile(frame, fileName, File_Type.RawData);
                        }

                        if (IsSaveExposureToRAWFileChecked)
                        {
                            if (UseLowGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Low_", FILE_EXTENSION_RAW);
                                nameList.Add(fileName);
                                SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.LowGain), fileName);
                            }
                            if (UseHighGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "High_", FILE_EXTENSION_RAW);
                                SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.HighGain), fileName);
                            }
                            if (UseMergedImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Merge_", FILE_EXTENSION_RAW);
                                SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.Merged), fileName);
                            }
                        }

                        if (IsSaveExposureToTIFFFileChecked)
                        {
                            if (UseLowGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Low_", FILE_EXTENSION_TIFF);
                                nameList.Add(fileName);
                                sourceLow = SaveImageFile(frame, fileName, ImageFormat.Tiff, sourceLow, frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                            }
                            if (UseHighGainImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "High_", FILE_EXTENSION_TIFF);
                                sourceHigh = SaveImageFile(frame, fileName, ImageFormat.Tiff, sourceHigh, frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                            }
                            if (UseMergedImageChecked)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                fileName = GetValidFilename(folder, fileNameBase + meanName + "Merge_", FILE_EXTENSION_TIFF);
                                sourceMerge = SaveImageFile(frame, fileName, ImageFormat.Tiff, sourceMerge, frame.GetPixelArray(FliFrame.FliImageType.Merged));
                            }
                        }

                        if (IsSaveMetaDataToFileChecked)
                        {
                            fileName = GetValidFilename(folder, fileNameBase + "Metadata_", FILE_EXTENSION_TXT);
                            SaveMetadataFile(frame, fileName);
                        }

                        if (SaveSectorsAsFiles)
                        {
                            WriteFileSectors(frame, folder, nameList);
                        }
                    }
                    catch (Exception ex)
                    {
                        if (!fliProCameraManager.FileWriteError)
                        {
                            MessageBox.Show("Error encountered writing to folder:\n" + folder + "\n\n" + ex.Message);
                            fliProCameraManager.FileWriteError = true;
                        }
                    }
                }
            }
        }

        public string WriteFile(FliFrame frame, string folder, string fileName, string extension)
        {
            string tempFileName = String.Empty;

            if (!String.IsNullOrEmpty(fileName))
            {
                lock (ImageWriteLock)
                {
                    try
                    {
                        if (fliProCameraManager == null)
                        {
                            fliProCameraManager = FliProCameraManager.Instance;
                        }
                        BitmapSource sourceLow = null;
                        BitmapSource sourceHigh = null;
                        BitmapSource sourceMerge = null;
                        ushort[] lowGainArray = frame.GetPixelArray(FliFrame.FliImageType.LowGainDummy);
                        ushort[] highGainArray = frame.GetPixelArray(FliFrame.FliImageType.HighGainDummy);
                        ushort[] mergeGainArray = frame.GetPixelArray(FliFrame.FliImageType.MergedDummy);
                        List<string> nameList = new List<string>();
                        string meanName = String.Empty;

                        if (extension.Equals(FILE_EXTENSION_FIT))
                        {
                            if (lowGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Low_", FILE_EXTENSION_FIT);
                                nameList.Add(tempFileName);
                                FITSImageProcessing.SaveFitsFile(tempFileName, frame, FliFrame.FliImageType.LowGainDummy);
                            }
                            if (highGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_High_", FILE_EXTENSION_FIT);
                                nameList.Add(tempFileName);
                                FITSImageProcessing.SaveFitsFile(tempFileName, frame, FliFrame.FliImageType.HighGainDummy);
                            }
                            if (mergeGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Merge_", FILE_EXTENSION_FIT);
                                nameList.Add(tempFileName);
                                FITSImageProcessing.SaveFitsFile(tempFileName, frame, FliFrame.FliImageType.MergedDummy);
                            }
                        }

                        if (extension.Equals(FILE_EXTENSION_JPEG))
                        {
                            if (lowGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Low_", FILE_EXTENSION_JPEG);
                                nameList.Add(tempFileName);
                                sourceLow = SaveImageFile8Bit(frame, tempFileName, ImageFormat.Jpeg, sourceLow, frame.GetTransformedPixelArray(FliFrame.FliImageType.LowGain));
                            }
                            if (highGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_High_", FILE_EXTENSION_JPEG);
                                nameList.Add(tempFileName);
                                sourceHigh = SaveImageFile8Bit(frame, tempFileName, ImageFormat.Jpeg, sourceHigh, frame.GetTransformedPixelArray(FliFrame.FliImageType.HighGain));
                            }
                            if (mergeGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Merge_", FILE_EXTENSION_JPEG);
                                nameList.Add(tempFileName);
                                sourceMerge = SaveImageFile8Bit(frame, tempFileName, ImageFormat.Jpeg, sourceMerge, frame.GetTransformedPixelArray(FliFrame.FliImageType.Merged));
                            }
                        }

                        if (extension.Equals(FILE_EXTENSION_PNG))
                        {
                            if (lowGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Low_", FILE_EXTENSION_PNG);
                                nameList.Add(tempFileName);
                                sourceLow = SaveImageFile(frame, tempFileName, ImageFormat.Png, sourceLow, frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                            }
                            if (highGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_High_", FILE_EXTENSION_PNG);
                                nameList.Add(tempFileName);
                                sourceHigh = SaveImageFile(frame, tempFileName, ImageFormat.Png, sourceHigh, frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                            }
                            if (mergeGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Merge_", FILE_EXTENSION_PNG);
                                nameList.Add(tempFileName);
                                sourceMerge = SaveImageFile(frame, tempFileName, ImageFormat.Png, sourceMerge, frame.GetPixelArray(FliFrame.FliImageType.Merged));
                            }
                        }

                        if (extension.Equals(FILE_EXTENSION_BIN))
                        {
                            // First save the low only as a 16 bit file 
                            if (lowGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Low_", FILE_EXTENSION_BIN);
                                nameList.Add(tempFileName);
                                SaveBINFile(frame, tempFileName, File_Type.Low);
                            }
                            if (highGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_High_", FILE_EXTENSION_BIN);
                                nameList.Add(tempFileName);
                                SaveBINFile(frame, tempFileName, File_Type.High);
                            }
                            if (mergeGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Merge_", FILE_EXTENSION_BIN);
                                nameList.Add(tempFileName);
                                SaveBINFile(frame, tempFileName, File_Type.Merged);
                            }
                        }

                        if (extension.Equals(FILE_EXTENSION_RCD))
                        {
                            // Save the Combined file which is exactly how we received it from the board.
                            tempFileName = GetValidFilename(folder, fileName, FILE_EXTENSION_RCD);
                            nameList.Add(tempFileName);
                            SaveBINFile(frame, tempFileName, File_Type.RawData);
                        }

                        if (extension.Equals(FILE_EXTENSION_RAW))
                        {
                            if (lowGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Low_", FILE_EXTENSION_RAW);
                                nameList.Add(tempFileName);
                                SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.LowGain), tempFileName);
                            }
                            if (highGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_High_", FILE_EXTENSION_RAW);
                                nameList.Add(tempFileName);
                                SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.HighGain), tempFileName);
                            }
                            if (mergeGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Merge_", FILE_EXTENSION_RAW);
                                nameList.Add(tempFileName);
                                SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.Merged), tempFileName);
                            }
                        }

                        if (extension.Equals(FILE_EXTENSION_TIFF) || extension.Equals(FILE_EXTENSION_TIFF + "f"))
                        {
                            if (lowGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.LowGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Low_", FILE_EXTENSION_TIFF);
                                nameList.Add(tempFileName);
                                sourceLow = SaveImageFile(frame, tempFileName, ImageFormat.Tiff, sourceLow, frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                            }
                            if (highGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.HighGainDummy);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_High_", FILE_EXTENSION_TIFF);
                                nameList.Add(tempFileName);
                                sourceHigh = SaveImageFile(frame, tempFileName, ImageFormat.Tiff, sourceHigh, frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                            }
                            if (mergeGainArray != null)
                            {
                                meanName = fliProCameraManager.GetNameValue(NameEntryType.Mean, frame, String.Empty, FliFrame.FliImageType.Merged);
                                tempFileName = GetValidFilename(folder, fileName + meanName + "_Merge_", FILE_EXTENSION_TIFF);
                                nameList.Add(tempFileName);
                                sourceMerge = SaveImageFile(frame, tempFileName, ImageFormat.Tiff, sourceMerge, frame.GetPixelArray(FliFrame.FliImageType.Merged));
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        if (!fliProCameraManager.FileWriteError)
                        {
                            MessageBox.Show("Error encountered writing to folder:\n" + folder + "\nFilename:\n" + tempFileName + "\n\n" + ex.Message);
                            fliProCameraManager.FileWriteError = true;
                        }
                    }
                }
            }
            return fileName;
        }

        public void WriteFileInvalidMetadata(FliFrame frame, string folder, string NameHint = "", bool forceFileWrite = false)
        {
            // Don't allow writing frame to file when this camera is already from a file
            if ((Type == CameraType.FILE) && !forceFileWrite)
            {
                return;
            }

            if (!String.IsNullOrEmpty(folder))
            {
                //string frameNumberString = (frame.FrameMetaData == null ? "0000" : frame.FrameMetaData.FrameNumber.ToString("0000")) + "_";
                //string targetSerialNumber = SerialNumber;
                //if (SerialNumber.StartsWith("Camera_"))
                //{
                //    // This is an image from a file so filter out the extra substrings
                //    targetSerialNumber = targetSerialNumber.Substring(7);
                //    int index = targetSerialNumber.LastIndexOf(".");
                //    targetSerialNumber = targetSerialNumber.Substring(0, index);
                //}
                string fileNameBase = fliProCameraManager.CalculateNameString(frame);
                string fileName = String.Empty;
                List<string> nameList = new List<string>();

                // The 16 bit arrays have not been generated so only save the raw image data 
                fileName = GetValidFilename(folder, fileNameBase + "RawData_", FILE_EXTENSION_BIN);
                nameList.Add(fileName);
                SaveBINFile(frame, fileName, File_Type.RawData);

                if (IsSaveMetaDataToFileChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Metadata_", FILE_EXTENSION_TXT);
                    SaveMetadataFile(frame, fileName);
                }
            }
        }

        public void WriteFileSkipped(FliFrame frame, string folder, string NameHint = "", bool forceFileWrite = false)
        {
            // Don't allow writing frame to file when this camera is already from a file
            if ((Type == CameraType.FILE) && !forceFileWrite)
            {
                return;
            }

            if (!String.IsNullOrEmpty(folder))
            {
                if (frame.FrameMetaData == null)
                {
                    if (frame.RawData != null)
                    {
                        // Strip off the MetaData and put it into the RawMetaData property
                        frame.RawMetaData = new byte[FliProFrameMetadata.GetMetadataLengthFromArray(frame.RawData)];
                        Buffer.BlockCopy(frame.RawData, 0, frame.RawMetaData, 0, frame.RawMetaData.Length);

                        // Parse the raw bytes into the metadata structure
                        frame.FrameMetaData = new FliProFrameMetadata(frame.RawMetaData);
                        frame.FrameMetaData.TestLowGainAbsolute = GetGainValueAbsoluteFromIndex(LowGainTable, frame.FrameMetaData.LowGain);
                        frame.FrameMetaData.TestHighGainAbsolute = GetGainValueAbsoluteFromIndex(HighGainTable, frame.FrameMetaData.HighGain);

                        // Temporary code
                        frame.FrameMetaData.NumberOfDataChannels = 8;
                        //FrameMetaData.PixelOrderedImage = !UnscrambleImage;

                        frame.IsValid = frame.IsValid && (frame.FrameMetaData.MagicNumber == FliProFrameMetadata.MAGIC_NUMBER);
                        if (!frame.IsValid)
                        {
                            // This is corrupted data so we need to just abort and return
                            LogEvent("ParseRawImageData: Frame data invalid");
                            CancelExposure();
                            WriteFileInvalidMetadata(frame, fliProCameraManager.RawFilePath, "BadMetadataFile");
                            return;
                        }

                        BitsPerPixel = frame.FrameMetaData.DataPixelBitDepth;

                        // Insure that the FrameMetaData.StartingExposureRow is always even. If it is odd we will crash with Index out of range while parsing.
                        // This is a problem in the image generation code that once fixed should never occur.
                        if (frame.FrameMetaData.StartingExposureRow % 2 > 0)
                        {
                            frame.FrameMetaData.StartingExposureRow--;
                        }

                        int totalHeight = frame.FrameMetaData.PreDummyRows + (frame.FrameMetaData.ImageStopRow - frame.FrameMetaData.ImageStartRow + 1) + frame.FrameMetaData.PostDummyRows;
                        frame.FrameAreaWH = new FliPro.FliRectangleWH(0, 0, 
                            (frame.FrameMetaData.ImageStopColumn - frame.FrameMetaData.ImageStartColumn + 1),
                            (frame.FrameMetaData.ImageStopRow - frame.FrameMetaData.ImageStartRow + 1));
                        frame.FrameAreaWHTotal = new FliPro.FliRectangleWH(0, 0, (frame.FrameMetaData.ImageStopColumn - frame.FrameMetaData.ImageStartColumn + 1) +
                            frame.FrameMetaData.DummyPixelsPerRow, totalHeight);
                    }
                }
                //string frameNumberString = (frame.FrameMetaData == null ? "0000" : frame.FrameMetaData.FrameNumber.ToString("0000")) + "_";
                //string targetSerialNumber = SerialNumber;
                //if (SerialNumber.StartsWith("Camera_"))
                //{
                //    // This is an image from a file so filter out the extra substrings
                //    targetSerialNumber = targetSerialNumber.Substring(7);
                //    int index = targetSerialNumber.IndexOf("_");
                //    targetSerialNumber = targetSerialNumber.Substring(0, index);
                //}
                string fileNameBase = fliProCameraManager.CalculateNameString(frame);
                string fileName = String.Empty;
                List<string> nameList = new List<string>();

                // The 16 bit arrays have not been generated so only save the raw image data 
                fileName = GetValidFilename(folder, fileNameBase + "RawData_", FILE_EXTENSION_BIN);
                nameList.Add(fileName);
                SaveBINFile(frame, fileName, File_Type.RawData);

                if (IsSaveMetaDataToFileChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Metadata_", FILE_EXTENSION_TXT);
                    SaveMetadataFile(frame, fileName);
                }
            }
        }

        public void WriteFileWithFileName(FliFrame frame, string folder, string fileName, string extension)
        {
            string targetSerialNumber = SerialNumber;
            string fileNameBase = fileName + "_" + DateTime.Now.ToString("yyyy_MM_dd") + "_";
            string completeFileName = String.Empty;
            BitmapSource sourceLow = null;
            BitmapSource sourceHigh = null;
            BitmapSource sourceMerge = null;
            List<string> nameList = new List<string>();

            if (IsSaveExposureToFITFileChecked || extension.Equals(FILE_EXTENSION_FIT, StringComparison.OrdinalIgnoreCase))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_FIT);
                    nameList.Add(fileName);
                    FITSImageProcessing.SaveFitsFile(fileName, frame, FliFrame.FliImageType.LowGainDummy);
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_FIT);
                    nameList.Add(fileName);
                    FITSImageProcessing.SaveFitsFile(fileName, frame, FliFrame.FliImageType.HighGainDummy);
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_FIT);
                    nameList.Add(fileName);
                    FITSImageProcessing.SaveFitsFile(fileName, frame, FliFrame.FliImageType.MergedDummy);
                }
            }

            if (IsSaveExposureToJPEGFileChecked || extension.Equals(FILE_EXTENSION_JPEG, StringComparison.OrdinalIgnoreCase))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_JPEG);
                    nameList.Add(fileName);
                    sourceLow = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, sourceLow, frame.GetTransformedPixelArray(FliFrame.FliImageType.LowGain));
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_JPEG);
                    sourceHigh = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, sourceHigh, frame.GetTransformedPixelArray(FliFrame.FliImageType.HighGain));
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_JPEG);
                    sourceMerge = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, sourceMerge, frame.GetTransformedPixelArray(FliFrame.FliImageType.Merged));
                }
            }

            if (IsSaveExposureToPNGFileChecked || extension.Equals(FILE_EXTENSION_PNG, StringComparison.OrdinalIgnoreCase))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_PNG);
                    nameList.Add(fileName);
                    sourceLow = SaveImageFile(frame, fileName, ImageFormat.Png, sourceLow, frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_PNG);
                    sourceHigh = SaveImageFile(frame, fileName, ImageFormat.Png, sourceHigh, frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_PNG);
                    sourceMerge = SaveImageFile(frame, fileName, ImageFormat.Png, sourceMerge, frame.GetPixelArray(FliFrame.FliImageType.Merged));
                }
            }

            if (IsSaveExposureToBINFileChecked || extension.Equals(FILE_EXTENSION_BIN, StringComparison.OrdinalIgnoreCase))
            {
                // Then first the low only as a 16 bit file 
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_BIN);
                    nameList.Add(fileName);
                    SaveBINFile(frame, fileName, File_Type.Low);
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_BIN);
                    SaveBINFile(frame, fileName, File_Type.High);
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_BIN);
                    SaveBINFile(frame, fileName, File_Type.Merged);
                }
            }

            if (IsSaveExposureToRCDFileChecked || extension.Equals(FILE_EXTENSION_RCD, StringComparison.OrdinalIgnoreCase))
            {
                // First save the Combined file which is exactly how we received it from the board.
                fileName = GetValidFilename(folder, fileNameBase + "RCDData_", FILE_EXTENSION_RCD);
                nameList.Add(fileName);
                SaveBINFile(frame, fileName, File_Type.RawData);

            }

            if (IsSaveExposureToRAWFileChecked || extension.Equals(FILE_EXTENSION_RAW))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_RAW);
                    nameList.Add(fileName);
                    SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.LowGain), fileName);
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_RAW);
                    SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.HighGain), fileName);
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_RAW);
                    SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.Merged), fileName);
                }
            }

            if (IsSaveExposureToTIFFFileChecked || extension.Equals(FILE_EXTENSION_TIFF, StringComparison.OrdinalIgnoreCase))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_TIFF);
                    nameList.Add(fileName);
                    sourceLow = SaveImageFile(frame, fileName, ImageFormat.Tiff, sourceLow, frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_TIFF);
                    sourceHigh = SaveImageFile(frame, fileName, ImageFormat.Tiff, sourceHigh, frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_TIFF);
                    sourceMerge = SaveImageFile(frame, fileName, ImageFormat.Tiff, sourceMerge, frame.GetPixelArray(FliFrame.FliImageType.Merged));
                }
            }

            if (IsSaveMetaDataToFileChecked)
            {
                fileName = GetValidFilename(folder, fileNameBase + "Metadata_", FILE_EXTENSION_TXT);
                SaveMetadataFile(frame, fileName);
            }

            if (SaveSectorsAsFiles)
            {
                WriteFileSectors(frame, folder, nameList);
            }
        }

        public bool WriteFile(FliFrame.FliImageType imageType, string fileName, FliFrame frame)
        {
            BitmapSource source = null;
            bool fileExtensionValid = false;

            if (!String.IsNullOrEmpty(fileName))
            {
                if (fileName.EndsWith(FILE_EXTENSION_FIT))
                {
                    FITSImageProcessing.SaveFitsFile(fileName, frame, imageType);
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_JPEG))
                {
                    source = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, source, frame.GetTransformedPixelArray(imageType));
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_PNG))
                {
                    source = SaveImageFile(frame, fileName, ImageFormat.Png, source, frame.GetPixelArray(imageType));
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_RAW))
                {
                    SaveRAWFile(frame.GetPixelArray(imageType), fileName);
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_RCD))
                {
                    SaveBINFile(frame, fileName, File_Type.RawData);
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_BIN))
                {
                    File_Type fileType = File_Type.RawData;
                    if ((imageType == FliFrame.FliImageType.HighGain) || (imageType == FliFrame.FliImageType.HighGainDummy))
                    {
                        fileType = File_Type.High;
                    }
                    else if ((imageType == FliFrame.FliImageType.LowGain) || (imageType == FliFrame.FliImageType.LowGainDummy))
                    {
                        fileType = File_Type.Low;
                    }
                    else
                    {
                        fileType = File_Type.Merged;
                    }
                    SaveBINFile(frame, fileName, fileType);
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_TIFF))
                {
                    source = SaveImageFile(frame, fileName, ImageFormat.Tiff, source, frame.GetPixelArray(imageType));
                    fileExtensionValid = true;
                }
            }
            return fileExtensionValid;
        }

        private void WriteFileSectors(FliFrame frame, string folder, List<string> nameList)
        {
            FliFrame subframe = new FliFrame(frame.Camera);
            ushort[,] newLowArray = new ushort[frame.FrameAreaWH.H, frame.FrameAreaWH.W / 8];
            string fileName = String.Empty;
            BitmapSource source = null;
            int fileNameIndex = 0;
            int index = 0;

            subframe.FrameMetaData = frame.FrameMetaData;
            subframe.FrameAreaWH = new FliPro.FliRectangleWH(0, 0, frame.FrameAreaWH.W / 8, frame.FrameAreaWH.H);
            subframe.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
            subframe.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
            subframe.FrameMetaData.HorizontalPixels = subframe.FrameAreaWH.W;
            subframe.FrameMetaData.ImageStartColumn = 255;
            subframe.FrameMetaData.PixelOrderedImage = true;
            subframe.FrameMetaData.HorizontalBinning = 1;
            subframe.FrameMetaData.VerticalBinning = 1;
            subframe.FrameMetaData.HighGain = fliProCameraManager.ActiveCamera.GetMetadataIndexFromGainIndex(
                        fliProCameraManager.ActiveCamera.HighGainTable, fliProCameraManager.ActiveCamera.HighGainIndex);
            subframe.FrameMetaData.LowGain = fliProCameraManager.ActiveCamera.GetMetadataIndexFromGainIndex(
                        fliProCameraManager.ActiveCamera.LowGainTable, fliProCameraManager.ActiveCamera.LowGainIndex);

            for (int i = 0; i < 8; i++)
            {
                int offset = i * subframe.FrameAreaWH.W;
                for (int y = 0; y < frame.FrameAreaWH.H; y++)
                {
                    for (int x = 0; x < subframe.FrameAreaWH.W; x++)
                    {
                        newLowArray[y, x] = frame.GetPixel(FliFrame.FliImageType.LowGainDummy, offset + x, y);
                    }
                }
                BitsPerPixel = FliFrame.BITS_PER_PIXEL_16;
                subframe.LowGainDummyPixelArray = subframe.CopyToOneDimensionalArray(newLowArray);
                fileNameIndex = 0;  // Reset the index

                // Now write the files
                if (IsSaveExposureToFITFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    FITSImageProcessing.SaveFitsFile(fileName, subframe, FliFrame.FliImageType.LowGainDummy);
                }

                if (IsSaveExposureToJPEGFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    source = SaveImageFile8Bit(subframe, fileName, ImageFormat.Jpeg, source, subframe.GetTransformedPixelArray(FliFrame.FliImageType.LowGain));
                }

                if (IsSaveExposureToPNGFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    source = SaveImageFile(subframe, fileName, ImageFormat.Png, source, subframe.GetPixelArray(FliFrame.FliImageType.LowGain));
                }

                if (IsSaveExposureToBINFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    SaveBINFile(subframe, fileName, File_Type.Low);
                }

                if (IsSaveExposureToRCDFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    SaveBINFile(subframe, fileName, File_Type.Low);
                }

                if (IsSaveExposureToRAWFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    SaveRAWFile(subframe.GetPixelArray(FliFrame.FliImageType.LowGain), fileName);
                }

                if (IsSaveExposureToTIFFFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    source = SaveImageFile(subframe, fileName, ImageFormat.Tiff, source, subframe.GetPixelArray(FliFrame.FliImageType.LowGain));
                }
            }
        }

        public void SaveBINFile(FliFrame frame, string fileName, File_Type fileType)
        {
            try
            {
                FileStream fs = new FileStream(fileName, FileMode.CreateNew);
                // Create the writer for data.
                BinaryWriter w = new BinaryWriter(fs);
                ushort[] pixelArray;
                bool tempHDRMode = false;
                bool tempPixelOrderedImage = false;
                int tempDataPixelBitDepth = 0;
                int tempSensorPixelBitDepth = 0;

                if ((frame.RawData == null) && ((fileType == File_Type.RawData) || (fileType == File_Type.Low)))
                {
                    // This is data from an old camera so write the MetaData and the low gain array
                    tempHDRMode = frame.FrameMetaData.HDRMode;
                    tempPixelOrderedImage = frame.FrameMetaData.PixelOrderedImage;
                    tempDataPixelBitDepth = frame.FrameMetaData.DataPixelBitDepth;
                    tempSensorPixelBitDepth = frame.FrameMetaData.SensorPixelBitDepth;
                    if (fileType == File_Type.Low)
                    {
                        frame.FrameMetaData.HDRMode = false;
                    }
                    frame.FrameMetaData.PixelOrderedImage = true;  // Need to set this flag so that the image is read back in properly
                    frame.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                    frame.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                    w.Write(frame.FrameMetaData.GetByteArray());
                    frame.FrameMetaData.HDRMode = tempHDRMode;
                    frame.FrameMetaData.PixelOrderedImage = tempPixelOrderedImage;
                    frame.FrameMetaData.DataPixelBitDepth = tempDataPixelBitDepth;
                    frame.FrameMetaData.SensorPixelBitDepth = tempSensorPixelBitDepth;
                    pixelArray = frame.GetPixelArray(FliFrame.FliImageType.LowGainDummy);
                    byte[] byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                    Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                    w.Write(byteArr);
                    pixelArray = frame.GetPixelArray(FliFrame.FliImageType.HighGainDummy);
                    if ((fileType != File_Type.Low) && (pixelArray != null))
                    {
                        byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                        Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                        w.Write(byteArr);
                    }
                }
                else   // This is a new camera
                {
                    switch (fileType)
                    {
                        case File_Type.RawData:
                            if (frame.RawData != null)
                            {
                                w.Write(frame.RawData);
                            }
                            break;

                        case File_Type.High:
                            pixelArray = frame.GetPixelArray(FliFrame.FliImageType.HighGainDummy);
                            if (pixelArray != null)
                            {
                                tempHDRMode = frame.FrameMetaData.HDRMode;
                                tempPixelOrderedImage = frame.FrameMetaData.PixelOrderedImage;
                                tempDataPixelBitDepth = frame.FrameMetaData.DataPixelBitDepth;
                                tempSensorPixelBitDepth = frame.FrameMetaData.SensorPixelBitDepth;
                                frame.FrameMetaData.HDRMode = false;
                                frame.FrameMetaData.PixelOrderedImage = true;  // Need to set this flag so that the image is read back in properly
                                frame.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                frame.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                w.Write(frame.FrameMetaData.GetByteArray());
                                frame.FrameMetaData.HDRMode = tempHDRMode;
                                frame.FrameMetaData.PixelOrderedImage = tempPixelOrderedImage;
                                frame.FrameMetaData.DataPixelBitDepth = tempDataPixelBitDepth;
                                frame.FrameMetaData.SensorPixelBitDepth = tempSensorPixelBitDepth;
                                byte[] byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                                Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                                w.Write(byteArr);
                            }
                            break;

                        case File_Type.Low:
                            pixelArray = frame.GetPixelArray(FliFrame.FliImageType.LowGainDummy);
                            if (pixelArray != null)
                            {
                                tempHDRMode = frame.FrameMetaData.HDRMode;
                                tempPixelOrderedImage = frame.FrameMetaData.PixelOrderedImage;
                                tempDataPixelBitDepth = frame.FrameMetaData.DataPixelBitDepth;
                                tempSensorPixelBitDepth = frame.FrameMetaData.SensorPixelBitDepth;
                                frame.FrameMetaData.HDRMode = false;
                                frame.FrameMetaData.PixelOrderedImage = true;  // Need to set this flag so that the image is read back in properly
                                frame.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                frame.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                w.Write(frame.FrameMetaData.GetByteArray());
                                frame.FrameMetaData.HDRMode = tempHDRMode;
                                frame.FrameMetaData.PixelOrderedImage = tempPixelOrderedImage;
                                frame.FrameMetaData.DataPixelBitDepth = tempDataPixelBitDepth;
                                frame.FrameMetaData.SensorPixelBitDepth = tempSensorPixelBitDepth;
                                byte[] byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                                Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                                w.Write(byteArr);
                            }
                            break;

                        case File_Type.Merged:
                            pixelArray = frame.GetPixelArray(FliFrame.FliImageType.MergedDummy);
                            if (pixelArray != null)
                            {
                                tempHDRMode = frame.FrameMetaData.HDRMode;
                                tempPixelOrderedImage = frame.FrameMetaData.PixelOrderedImage;
                                tempDataPixelBitDepth = frame.FrameMetaData.DataPixelBitDepth;
                                tempSensorPixelBitDepth = frame.FrameMetaData.SensorPixelBitDepth;
                                frame.FrameMetaData.HDRMode = false;
                                frame.FrameMetaData.PixelOrderedImage = true;  // Need to set this flag so that the image is read back in properly
                                frame.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                frame.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                w.Write(frame.FrameMetaData.GetByteArray());
                                frame.FrameMetaData.HDRMode = tempHDRMode;
                                frame.FrameMetaData.PixelOrderedImage = tempPixelOrderedImage;
                                frame.FrameMetaData.DataPixelBitDepth = tempDataPixelBitDepth;
                                frame.FrameMetaData.SensorPixelBitDepth = tempSensorPixelBitDepth;
                                byte[] byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                                Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                                w.Write(byteArr);
                            }
                            break;
                    }
                }
                w.Close();
                fs.Close();
            }
            catch { }
        }

        /// <summary>
        /// Creates the target file type under the given filename and returns the generated BitmapSource for reuse
        /// if more than one file is requested.
        /// </summary>
        /// <param name="frame"></param>
        /// <param name="fileName"></param>
        /// <param name="imageFormat"></param>
        /// <param name="source"></param>
        /// <returns></returns>
        public BitmapSource SaveImageFile(FliFrame frame, string fileName, ImageFormat imageFormat, BitmapSource source, ushort[] targetArray)
        {
            // Create an Image from the byte array
            if (source == null)
            {
                if (targetArray != null)
                {
                    byte[] byteArr = new byte[targetArray.Length * 2];
                    Buffer.BlockCopy(targetArray, 0, byteArr, 0, byteArr.Length);
                    Bitmap bitmap = new Bitmap(frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H,
                        System.Drawing.Imaging.PixelFormat.Format16bppGrayScale);
                    var bitmap_data = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                        ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format16bppGrayScale);
                    Marshal.Copy(byteArr, 0, bitmap_data.Scan0, byteArr.Length);
                    bitmap.UnlockBits(bitmap_data);

                    Rectangle rect = new Rectangle(0, 0, frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H);
                    BitmapData bitmapData = bitmap.LockBits(rect, ImageLockMode.ReadOnly, bitmap.PixelFormat);

                    source = BitmapSource.Create(bitmap.Width,
                                                 bitmap.Height,
                                                 bitmap.HorizontalResolution,
                                                 bitmap.VerticalResolution,
                                                 PixelFormats.Gray16,
                                                 null,
                                                 bitmapData.Scan0,
                                                 bitmapData.Stride * bitmap.Height,
                                                 bitmapData.Stride);

                    bitmap.UnlockBits(bitmapData);
                }
            }

            if (source != null)
            { 
                FileStream stream = new FileStream(fileName, FileMode.Create);

                if (imageFormat == ImageFormat.Tiff)
                {
                    TiffBitmapEncoder tiffEncoder = new TiffBitmapEncoder();

                    tiffEncoder.Compression = TiffCompressOption.Zip;
                    tiffEncoder.Frames.Add(BitmapFrame.Create(source));
                    tiffEncoder.Save(stream);
                }
                else if (imageFormat == ImageFormat.Jpeg)
                {
                    JpegBitmapEncoder jpegEncoder = new JpegBitmapEncoder();

                    jpegEncoder.QualityLevel = 100;  // Save 100% of the data
                    jpegEncoder.Frames.Add(BitmapFrame.Create(source));
                    jpegEncoder.Save(stream);
                }
                else if (imageFormat == ImageFormat.Png)
                {
                    PngBitmapEncoder pngEncoder = new PngBitmapEncoder();

                    pngEncoder.Frames.Add(BitmapFrame.Create(source));
                    pngEncoder.Save(stream);
                }

                stream.Close();
            }

            return source;
        }

        /// <summary>
        /// Creates the target file type under the given filename and returns the generated BitmapSource for reuse
        /// if more than one file is requested.
        /// </summary>
        /// <param name="frame"></param>
        /// <param name="fileName"></param>
        /// <param name="imageFormat"></param>
        /// <param name="source"></param>
        /// <returns></returns>
        public BitmapSource SaveImageFile8Bit(FliFrame frame, string fileName, ImageFormat imageFormat, BitmapSource source, byte[] targetArray)
        {
            // Create an Image from the byte array
            if (source == null)
            {
                if (targetArray != null)
                {
                    byte[] byteArr = new byte[targetArray.Length];
                    Buffer.BlockCopy(targetArray, 0, byteArr, 0, byteArr.Length);
                    Bitmap bitmap = new Bitmap(frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H,
                        System.Drawing.Imaging.PixelFormat.Format8bppIndexed);
                    var bitmap_data = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                        ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format8bppIndexed);
                    Marshal.Copy(byteArr, 0, bitmap_data.Scan0, byteArr.Length);
                    bitmap.UnlockBits(bitmap_data);

                    Rectangle rect = new Rectangle(0, 0, frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H);
                    BitmapData bitmapData = bitmap.LockBits(rect, ImageLockMode.ReadOnly, bitmap.PixelFormat);

                    source = BitmapSource.Create(bitmap.Width,
                                                 bitmap.Height,
                                                 bitmap.HorizontalResolution,
                                                 bitmap.VerticalResolution,
                                                 PixelFormats.Gray8,
                                                 null,
                                                 bitmapData.Scan0,
                                                 bitmapData.Stride * bitmap.Height,
                                                 bitmapData.Stride);

                    bitmap.UnlockBits(bitmapData);
                }
            }

            if (source != null)
            {
                FileStream stream = new FileStream(fileName, FileMode.Create);

                if (imageFormat == ImageFormat.Tiff)
                {
                    TiffBitmapEncoder tiffEncoder = new TiffBitmapEncoder();

                    tiffEncoder.Compression = TiffCompressOption.Zip;
                    tiffEncoder.Frames.Add(BitmapFrame.Create(source));
                    tiffEncoder.Save(stream);
                }
                else if (imageFormat == ImageFormat.Jpeg)
                {
                    JpegBitmapEncoder jpegEncoder = new JpegBitmapEncoder();

                    jpegEncoder.QualityLevel = 100;  // Save 100% of the data
                    jpegEncoder.Frames.Add(BitmapFrame.Create(source));
                    jpegEncoder.Save(stream);
                }
                else if (imageFormat == ImageFormat.Png)
                {
                    PngBitmapEncoder pngEncoder = new PngBitmapEncoder();

                    pngEncoder.Frames.Add(BitmapFrame.Create(source));
                    pngEncoder.Save(stream);
                }

                stream.Close();
            }

            return source;
        }

        /// <summary>
        /// Creates the JPeg file type under the given filename
        /// </summary>
        /// <param name="frame"></param>
        /// <param name="fileName"></param>
        /// <param name="imageFormat"></param>
        /// <param name="source"></param>
        /// <returns></returns>
        public void SaveJPEGFile(FliFrame frame, string fileName, ImageFormat imageFormat)
        {
            // Create an Image from the byte array
            BitmapSource source = null;
            ushort[] targetArray = frame.LowGainDummyPixelArray;

            if (UseMergedImageChecked && (frame.MergedDummyPixelArray != null))
            {
                targetArray = frame.MergedDummyPixelArray;
            }
            if (targetArray != null)
            {
                byte[] byteArr = new byte[targetArray.Length];
                for (int i = 0; i < targetArray.Length; i++)
                {
                    byteArr[i] = (byte)(targetArray[i] / 16);
                }
                Bitmap bitmap = new Bitmap(frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H);
                var bitmap_data = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                    ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format8bppIndexed);
                Marshal.Copy(byteArr, 0, bitmap_data.Scan0, byteArr.Length);
                bitmap.UnlockBits(bitmap_data);

                Rectangle rect = new Rectangle(0, 0, frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H);
                BitmapData bitmapData = bitmap.LockBits(rect, ImageLockMode.ReadOnly, bitmap.PixelFormat);

                source = BitmapSource.Create(bitmap.Width,
                                             bitmap.Height,
                                             bitmap.HorizontalResolution,
                                             bitmap.VerticalResolution,
                                             PixelFormats.Gray8,
                                             null,
                                             bitmapData.Scan0,
                                             bitmapData.Stride * bitmap.Height,
                                             bitmapData.Stride);

                bitmap.UnlockBits(bitmapData);
            }

            if (source != null)
            {
                FileStream stream = new FileStream(fileName, FileMode.Create);

                JpegBitmapEncoder jpegEncoder = new JpegBitmapEncoder();

                jpegEncoder.QualityLevel = 100;  // Save 100% of the data
                jpegEncoder.Frames.Add(BitmapFrame.Create(source));
                jpegEncoder.Save(stream);

                stream.Close();
            }
        }

        public void SaveMetadataFile(FliFrame frame, string fileName)
        {
            string formatedMetadata = frame.FrameMetaData.ToString();
            string formattedRawData = Environment.NewLine + "Raw Meta Data:" + Environment.NewLine + "[";

            if (frame.RawMetaData == null)
            {
                frame.RawMetaData = frame.FrameMetaData.GetByteArray();
            }

            for (int i = 0; i < frame.RawMetaData.Length; i++)
            {
                formattedRawData += frame.RawMetaData[i].ToString("X2");
                if (i < frame.RawMetaData.Length - 1)
                {
                    // Do this for all but the last one
                    if (((i + 1) % 20) == 0)
                    {
                        formattedRawData += "," + Environment.NewLine + " ";
                    }
                    else
                    {
                        formattedRawData += ", ";
                    }
                }
            }
            formattedRawData += "]";

            try
            {
                System.IO.File.WriteAllText(fileName, formatedMetadata + formattedRawData);
            }
            catch { }
        }

        public void SaveRAWFile(ushort[] targetArray, string fileName)
        {
            if (targetArray != null)
            {
                FileStream fs = new FileStream(fileName, FileMode.CreateNew);
                // Create the writer for data.
                BinaryWriter w = new BinaryWriter(fs);
                byte[] twoBytes = new byte[2];

                // Copy the 16 bit pixel data from the target gain array into a byte array
                byte[] byteArr = new byte[targetArray.Length * sizeof(ushort)];
                //Buffer.BlockCopy(targetArray, 0, byteArr, 0, byteArr.Length);

                for (int i = 0; i < targetArray.Length; ++i)
                {
                    twoBytes = BitConverter.GetBytes(targetArray[i]);
                    Array.Reverse(twoBytes);
                    Array.Copy(twoBytes, 0, byteArr, i * 2, 2);
                }

                w.Write(byteArr);
                w.Close();
                fs.Close();
            }
        }

        #endregion Public Methods

        //**************************************************************
        #region Protected Methods

        //-------------------------------------------
        /// <summary>
        /// Cleanup prior to disposing unmanaged objects
        /// </summary>
        protected override void OnDispose()
        {
            if (StatusMonitorJob != null)
            {
                StatusMonitorJob.Stop();
            }
            Disconnect();
            base.OnDispose();
        }

        protected void OnConnectBase()
        {
            Shutter = FliShutterControl.CLOSE;

            // Turn On The Temperature Timer
            ReadErrorEncountered = false;
            TemperatureStatusTimer.Enabled = true;
        }

        public virtual FliRectangle OverScanData(FliRectangle rec)
        {
            return rec;
        }

        protected void SetIsCapturingLocked(bool isCapturing)
        {
            lock(ImageCaptureLock)
            {
                IsCapturing = isCapturing;
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Update all dynamic device properties and status
        /// </summary>
        protected void UpdateCameraStatus()
        {
            try
            {
                lock (DeviceLock)
                {
                    if (IsConnected)
                    {
                        UpdateStatus();
                        StatusUpdateCount++;
                    }
                }
            }
            catch (Exception)
            {
                // Ignore for now
            }
        }

        protected virtual void Initialize()
        {
        }

        protected virtual int OpenDevice()
        {
            return 0;
        }

        protected virtual void CloseDevice() { }
        protected virtual void ReadNVStorage(uint uiOffset, byte[] pData, uint uiLength) { }
        protected virtual void WriteNVStorage(uint uiOffset, byte[] pData, uint uiLength) { }

        //-------------------------------------------
        /// <summary>
        /// Update dynamically changing values
        /// </summary>
        protected virtual void UpdateStatus()
        {
        }

        //-------------------------------------------
        /// <summary>
        /// Device Specific actions when connection is established
        /// </summary>
        protected virtual unsafe void OnConnect()
        {
            SerialNumber = "";
            HwVersion = 0;
            FwVersion = 0;
            Model = "";

            ArrayArea = new FliRectangle();
            VisibleArea = new FliRectangle();
            ImageArea = VisibleArea;
            PixelSize = new FliPixelSize();
            ColorPattern = FliBayerPattern.NONE;
            Shutter = FliShutterControl.CLOSE;
            TDIRate = 0;
            HBin = 1;
            VBin = 1;
            FrameCountType = FliFrameCount.SINGLE;
            FrameType = FliFrameType.NORMAL;
            NFlushes = 1;
            VideoModeEnabled = false;
        }

        public double ReadBoardTemp()
        {
            // get the Device Channel/Register offset
            long address = 0x75;
            int bufferLength = 0;
            byte[] buffer = new byte[8];
            string valueStr = String.Empty;
            double value = 0;

            buffer[0] = (byte)((I2C_DEVICE_ID_FPGA << 1) | 0x01);

            buffer[1] = (byte)address;
            bufferLength = 2;

            // Send and receive...
            byte[] results = SendRecvRaw(buffer, bufferLength);

            valueStr = "0x";
            for (int i = 0; i < 4; i++)
            {
                valueStr += results[i].ToString("X2");
            }

            value = Convert.ToInt32(valueStr, 16);
            value /= 1000;

            return value;
        }



        #endregion Protected Methods

        //**************************************************************
        #region Public Events

        public event EventHandler<EventArgs> ADUsPerElectronChangeAction;

        protected virtual void OnADUsPerElectronChangeAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = ADUsPerElectronChangeAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> CaptureCompleteAction;

        protected virtual void OnCaptureCompleteAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = CaptureCompleteAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }
        
        public event EventHandler<EventArgs> GainRatioChangeAction;

        protected virtual void OnGainRatioChangeAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = GainRatioChangeAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> LEDAction;

        protected virtual void OnLEDAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = LEDAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> MergeRangeChangeAction;

        protected virtual void OnMergeRangeChangeAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = MergeRangeChangeAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> LineOffsetChangeAction;

        protected virtual void OnLineOffsetChangeAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = LineOffsetChangeAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        

        #endregion Public Events
    }
}
