﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using FliUtils.Jobs;
using System.IO;
using System.Drawing.Imaging;
using System.Linq;
using System.Drawing;
using System.Windows.Media.Imaging;
using System.Windows.Media;
using System.Windows;
using System.IO.Compression;
using static FliProCameraLib.FliProCameraManager;

namespace FliProCameraLib
{
    /// <summary>
    /// Each instance of the FliProCamera class represents one Finger Lakes Imaging camera.
    /// This file contains methods and properties common to all classes derived from FliProCamera.
    /// </summary>
    public partial class FliProCamera : FliPro
    {
        //**************************************************************
        #region Public Constants

        public const int FLI_TEMPERATURE_INTERNAL = 0x0000;
        public const int FLI_TEMPERATURE_EXTERNAL = 0x0001;
        public const int FLI_TEMPERATURE_CCD = 0x0000;
        public const int FLI_TEMPERATURE_BASE = 0x0001;
        public const int FLI_TEMPERATURE_AMBIENT = 0x0002;
        public const int FLI_DEFAULT_TEMPERATURE_READ_INTERVAL = 5000;   // Check temperature every 5 seconds

        public const string FILE_EXTENSION_BIN = ".bin";
        public const string FILE_EXTENSION_FIT = ".fit";
        public const string FILE_EXTENSION_JPEG = ".jpg";
        public const string FILE_EXTENSION_PNG = ".png";
        public const string FILE_EXTENSION_RAW = ".raw";
        public const string FILE_EXTENSION_RCD = ".rcd";
        public const string FILE_EXTENSION_TIFF = ".tif";
        public const string FILE_EXTENSION_TXT = ".txt";

        public const string CAMERA_MODEL_KEPLER_0400 = "KEPLER KL400";
        public const string CAMERA_MODEL_KEPLER_0400_2 = "KEPLER 0400";
        public const string CAMERA_MODEL_KEPLER_0400_3 = "KEPLEPKL400";
        public const string CAMERA_MODEL_KEPLER_4040 = "KEPLER KL4040";
        public const string CAMERA_MODEL_KEPLER_2020 = "KEPLER KL2020";
        public const string CAMERA_MODEL_KODAK_29050 = "KODAK 29050";
        public const string CAMERA_MODEL_KODAK_47051 = "KODAK 47051";

        public const string CAMERA_MODEL_COBALT_DC230_42 = "Cobalt DC230-42";
        public const string CAMERA_MODEL_COBALT_DC230_84 = "Cobalt DC230-84";

        private const int I2C_DEVICE_ID_FPGA = 0x31;

        public const string CAMERAL_MODEL_FLAT_FIELD_CORRECTED_RESULTS = "FlatFieldCorrectedResults";
        public const string CAMERAL_MODEL_MERGED_RESULTS = "MergedResults";
        public const string CAMERAL_MODEL_SUBTRACT_RESULTS = "DifferenceResults";

        public const int MAX_THUMBNAIL_SIZE = 0x1000;
        public const int THUMBNAIL_DIMENSION_SIZE = 64;

        public const bool DO_SIMPLE_CONNECTION = true;

        #endregion Public Constants

        //**************************************************************
        #region Constructor and Initialization

        //-------------------------------------------
        /// <summary>
        /// Constructor
        /// </summary>
        public FliProCamera(FPRODEVICEINFO DevInfo)
        {
            ForcePropertyChange = true;

            DeviceLock = new object();
            IOPortLock = new object();
            ImageWriteLock = new object();
            ImageCaptureLock = new object();
            FrameListLock = new object();
            Handle = INVALID_HANDLE;
            IsConnected = false;
            IsCapturing = false;
            Type = CameraType.UNKNOWN;

            StatusUpdateCount = 0;
            SetDeviceInfo(DevInfo);
            InitializeBase();
            Initialize();
            SetDefaultGainTables();

            ForcePropertyChange = false;
        }

        //-------------------------------------------
        /// <summary>
        /// We need an empty Constructor to create the property backers for the entire UI
        /// </summary>
        public FliProCamera() : this(new FPRODEVICEINFO())
        {
            DeviceLock = new object();
            IOPortLock = new object();
            ImageCaptureLock = new object();
            FrameListLock = new object();
            Handle = INVALID_HANDLE;
            IsConnected = false;
            IsCapturing = false;
            Type = CameraType.UNKNOWN;

            SetDefaultFliFrameCountList();
            SetDefaultFliFrameTypeList();
            SetDefaultSupportedModes();
            //SetDefaultSupportedBinRatios();
            SetDefaultSupportedPixelDepths();
            SetDefaultGainTables();

            LowGainValue = LowGainTable[0];
            HighGainValue = HighGainTable[0];

            StatusUpdateCount = 0;
        }

        //-------------------------------------------
        /// <summary>
        /// Called to set the DevInfo for a Camera that was created as the empty startup camera
        /// </summary>
        /// <param name="DevInfo"></param>
        public virtual void SetDeviceInfo(FPRODEVICEINFO DevInfo)
        {
            DeviceInfo = DevInfo;

            //-=-= TODO: What do we really need for Model?
            if (String.IsNullOrEmpty(DevInfo.cFriendlyName))
            {
                Model = EMPTY_MODEL;
            }
            else
            {
                Model = DeviceInfo.cFriendlyName;
            }

            if (String.IsNullOrEmpty(DevInfo.cSerialNo))
            {
                SerialNumber = EMPTY_SERIAL_NUMBER;
            }
            else
            {
                SerialNumber = DevInfo.cSerialNo;
            }

            if (DevInfo.eConnType == FProConnection.FPRO_CONNECTION_FIBER)
            {
                Speed = " (QSFP)";
            }
            else
            {
                switch (DevInfo.USBSpeed)
                {
                    case FProUSBSpeed.FULLSPEED:
                        Speed = " (Full Speed)";
                        break;
                    case FProUSBSpeed.HIGHSPEED:
                        Speed = " (High Speed)";
                        break;
                    case FProUSBSpeed.SUPERSPEED:
                        Speed = " (Super Speed)";
                        break;
                    default:
                        Speed = String.Empty;
                        break;
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Initialize all properties
        /// </summary>
        protected virtual void InitializeBase()
        {
            StatusMonitorInterval = TimeSpan.FromMilliseconds(2000);

            HwVersion = 0;
            FwVersion = 0;
            ColorPattern = FliBayerPattern.NONE;
            IsConnected = false;

            //CameraControlSettings = new FliProCameraControlSettings();
            //FrameSettings = new FliProFrameSettings();
            SensorSettings = new FliProSensorSettings();

            SupportedPixelDepths = new ObservableCollection<uint>();
            SupportedBinRatios = new ObservableCollection<BinRatio>();
            SupportedModes = new ObservableCollection<NameValuePair>();
            LowGainTable = new ObservableCollection<SensorGain>();
            HighGainTable = new ObservableCollection<SensorGain>();

            ModeCount = 0;
            BinRatioCount = 0;
            LowGainCount = 0;
            HighGainCount = 0;
            HInverseScanSupported = false;
            VInverseScanSupported = false;

            MaxHBin = 1000;
            MaxVBin = 1000;
            HBin = 1;
            VBin = 1;
            MaxPixelDepth = DEFAULT_MAX_PIXEL_DEPTH;

            MaxImageHeight = MAX_IMAGE_HEIGHT;
            MaxImageWidth = MAX_IMAGE_WIDTH;

            FrameTimeout = 0;

            FramePool = new ConcurrentQueue<FliFrame>();
            MinimumFramePoolCount = DEFAULT_MIN_FRAME_POOL_COUNT;
            FramePoolBufSize = 0;

            TrackingFramePool = new ConcurrentQueue<FliFrame>();
            MinimumTrackingFramePoolCount = DEFAULT_MIN_TRACKING_FRAME_POOL_COUNT;
            TrackingFramePoolBufSize = 0;

            FrameSequenceNumber = 0;

            // Initialize The Temperature Timer
            TemperatureStatusTimer.Enabled = false;
            TemperatureStatusTimer.Elapsed += TemperatureStatusTimerEvent;
            TemperatureStatusTimer.Interval = FLI_DEFAULT_TEMPERATURE_READ_INTERVAL;
            TemperatureStatusTimer.AutoReset = true;

        }

        #endregion Constructor and Initialization

        //**************************************************************
        #region Support structures and classes

        public struct GainOffsetPair
        {
            public double gainRatio;
            public double temperature1;
            public int lineOffset1;
            public double temperature2;
            public int lineOffset2;
            public double temperature3;
            public int lineOffset3;
            public double stdDev;
        }

        #endregion Support structures and classes

        //**************************************************************
        #region Public Types

        public enum Exposure_Enum
        {
            exposure_0 = 0,
            exposure_5 = 5,
            exposure_10 = 10,
            exposure_30 =30,
            exposure_60 = 60,
            exposure_100 = 100,
            exposure_120 = 120,
            exposure_600 = 600
        }

        public enum Exposure_Triggers
        {
            Normal,
            External_Low,
            External_High
        }

        public enum ExposureUnits_Enum
        {
            units_uSecs,
            units_mSecs,
            units_secs,
            units_mins,
            units_hours
        }

        public enum File_Type
        {
            RawData,
            Low,
            High,
            Merged
        }

        public enum FrameInterval_Enum
        {
            interval_100 = 100,
            interval_600 = 600,
            interval_Other = 10000
        }

        public enum Resolution_Enum
        {
            row_period,
            clock_period
        }

        #endregion Public Types

        //**************************************************************
        #region Public Constants

        public const int MILLISECS_PER_HOUR = 3600 * 1000;
        public const int MILLISECS_PER_MIN = 60 * 1000;
        public const int MILLISECS_PER_SEC = 1000;

        #endregion Public Constants

        //**************************************************************
        #region Protected Properties

        protected object DeviceLock;
        protected object FrameListLock;
        protected object ImageCaptureLock;
        protected object ImageWriteLock;
        protected object IOPortLock;

        protected int Handle
        {
            get { return _Handle; }
            set { _Handle = value; }
        }
        protected int _Handle;

        protected TimeSpan StatusMonitorInterval { get; set; }

        protected double DelayCounter { get; set; }
        protected double FrameDelay { get; set; }

        protected ConcurrentQueue<FliFrame> FramePool { get; set; }

        protected bool ReadErrorEncountered { get; set; }

        protected ConcurrentQueue<FliFrame> TrackingFramePool { get; set; }

        protected System.Timers.Timer TemperatureStatusTimer = new System.Timers.Timer();

        #endregion

        //**************************************************************
        #region Public Properties
        //public string Name { get { return DeviceId.Name; } }

        public int BitsPerPixel { get; set; }
        public string SerialNumber { get; set; }
        public string Speed { get; protected set; }
        public int HwVersion { get; protected set; }
        public int FwVersion { get; protected set; }
        public string Model { get; protected set; }
        public CameraType Type { get; protected set; }
        public uint FrameSequenceNumber { get; protected set; }

        public int LEDOnTime { get; set; }

        public int DeviceType { get; set; }
        public int MaxImageHeight { get; set; }
        public int MaxImageWidth { get; set; }
        public int MetaDataSize = MAX_FRAME_METADATA_BYTES;

        public int MinimumFramePoolCount { get; set; }
        public int MinimumTrackingFramePoolCount { get; set; }

        public GainOffsetPair[,] GainOffsetPairArray = null;
        public GainOffsetPair[,] ManufacturersDefaultGainOffsetPairArray = null;

        #endregion Public Properties

        //**************************************************************
        #region Observable Collections
        #endregion Observable Collections

        //**************************************************************
        #region Observable Properties

        //-------------------------------------------
        public FPRODEVICEINFO DeviceInfo
        {
            get { return _DeviceInfo; }
            set { SetValue<FPRODEVICEINFO>(ref _DeviceInfo, value); }
        }
        protected FPRODEVICEINFO _DeviceInfo;

        public Visibility DeviceTypeNot400Visibility
        {
            get { return (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400.GetHashCode() ? Visibility.Visible : Visibility.Collapsed); }
        }

        //-------------------------------------------
        /// <summary>
        /// "User friendly" display string
        /// </summary>
        public string DisplayString
        {
            get { return _DisplayString; }
            protected set { SetValue<string>(ref _DisplayString, value); }
        }
        protected string _DisplayString;

        public bool FirstTraining = true;         // So we don't train the first time the temperature timer fires

        //-------------------------------------------
        /// <summary>
        /// Flag to signify that the device Grab is currently being Aborted
        /// </summary>
        public bool IsAborted
        {
            get { return _IsAborted; }
            set { SetValue<bool>(ref _IsAborted, value); }
        }
        protected bool _IsAborted = false;

        //-------------------------------------------
        /// <summary>
        /// True if device is connected
        /// </summary>
        public bool IsConnected
        {
            get { return _IsConnected; }
            protected set { SetValue<bool>(ref _IsConnected, value); }
        }
        protected bool _IsConnected;

        //-------------------------------------------
        /// <summary>
        /// True if device is currently capturing frames
        /// </summary>
        public bool IsCapturing
        {
            get { return _IsCapturing; }
            protected set
            {
                if (SetValue<bool>(ref _IsCapturing, value))
                {
                    if (value == false)
                    {
                        OnCaptureCompleteAction(new EventArgs());
                    }
                }
            }
        }
        protected bool _IsCapturing = true;

        //-------------------------------------------
        /// <summary>
        /// False if device is currently capturing frames
        /// </summary>
        public bool IsEnabled
        {
            get { return _IsEnabled; }
            protected set { SetValue<bool>(ref _IsEnabled, value); }
        }
        protected bool _IsEnabled = true;

        //-------------------------------------------
        /// <summary>
        /// Used to prevent multiple initializations during power up
        /// </summary>
        public bool IsInitializedFromSettings
        {
            get { return _IsInitializedFromSettings; }
            set { SetValue<bool>(ref _IsInitializedFromSettings, value); }
        }
        protected bool _IsInitializedFromSettings = false;

        public virtual bool IsLegacyCamera { get { return false; } }

        //-------------------------------------------
        /// <summary>
        /// Incremented on each status update
        /// </summary>
        public int StatusUpdateCount
        {
            get { return _StatusUpdateCount; }
            protected set { SetValue<int>(ref _StatusUpdateCount, value); }
        }
        protected int _StatusUpdateCount;

        //-------------------------------------------
        /// <summary>
        /// Property to display camera status in the bottom of the Main Window
        /// </summary>
        public string CameraStatusText
        {
            get { return _StatusText; }
            set { SetValue<string>(ref _StatusText, value); }
        }
        protected string _StatusText;

        //-------------------------------------------
        /// <summary>
        /// Flag to signify that reference data pixels should be included in the statistics
        /// </summary>
        public bool TotalStats
        {
            get { return _TotalStats; }
            set { SetValue<bool>(ref _TotalStats, value); }
        }
        protected bool _TotalStats = true;

        //-------------------------------------------
        /// <summary>
        /// Video Mode Capable
        /// </summary>
        public bool VideoModeCapable
        {
            get { return _VideoModeCapable; }
            set { SetValue<bool>(ref _VideoModeCapable, value); }
        }
        protected bool _VideoModeCapable = false;

        //-------------------------------------------
        /// <summary>
        /// Video Mode 
        /// </summary>
        public bool VideoModeEnabled
        {
            get { return _VideoModeEnabled; }
            set
            {
                SetValue<bool>(ref _VideoModeEnabled, value);
                SetVideoModeEnabled(value);
            }
        }
        protected bool _VideoModeEnabled;

        #endregion Observable Properties

        //**************************************************************
        #region Observable Settings Properties

        //-------------------------------------------
        /// <summary>
        /// The current settings of the sensor
        /// </summary>
        public FliProSensorSettings SensorSettings
        {
            get { return _SensorSettings; }
            set { SetValue(ref _SensorSettings, value); }
        }
        protected FliProSensorSettings _SensorSettings;

        #endregion Observable Settings Properties

        //**************************************************************
        #region Public Methods

        //-------------------------------------------
        /// <summary>
        /// Open the connection to this device if not already open, 
        /// and obtain basic information if available
        /// (FliOpen, FliGetSerialString, FliGetHWRevision, FliGetFWRevision)
        /// </summary>
        public virtual int Connect()
        {
            int status = -1;
            lock (DeviceLock)
            {
                if (!IsConnected)
                {
                    Handle = INVALID_HANDLE;
                    status = OpenDevice();

                    if (status < 0)
                    {
                        return status;
                    }

                    if (Handle != INVALID_HANDLE)
                    {
                        IsConnected = true;

                        if (DO_SIMPLE_CONNECTION)
                        {
                            // The only reason I made this function was so that we could get the image area programmed into the camera.
                            // We need that so that we can allocate the buffer to store the image. If desired you can simple set 
                            // the image size and exposure time and then just do the Grab. To do a more complete connection change
                            // DO_SIMPLE_CONNECTION to false. This will kick off the thread that monitors the temperature, etc.
                            OnConnectSimple();
                        }
                        else
                        {
                            OnConnect();
                            OnConnectBase();
                            UpdateCameraStatus();
                        }
                    }
                }
                else
                {
                    // We're already connected
                    status = 0;
                }
            }

            return status;
        }

        public void CopyCameraProperties(FliProCamera copyCamera)
        {

        }

        public byte[] Compress(byte[] data)
        {
            byte[] compressArray = null;
            using (MemoryStream memoryStream = new MemoryStream())
            {
                using (DeflateStream deflateStream = new DeflateStream(memoryStream, CompressionMode.Compress))
                {
                    deflateStream.Write(data, 0, data.Length);
                }
                compressArray = memoryStream.ToArray();
            }
            return compressArray;
        }

        public byte[] Decompress(byte[] data)
        {
            byte[] decompressedArray = null;
            using (MemoryStream decompressedStream = new MemoryStream())
            {
                using (MemoryStream compressStream = new MemoryStream(data))
                {
                    using (DeflateStream deflateStream = new DeflateStream(compressStream, CompressionMode.Decompress))
                    {
                        deflateStream.CopyTo(decompressedStream);
                    }
                }
                decompressedArray = decompressedStream.ToArray();
            }

            return decompressedArray;
        }

        //-------------------------------------------
        /// <summary>
        /// Close the connection to this device, if open
        /// </summary>
        public void Disconnect()
        {
            lock (DeviceLock)
            {
                if (IsConnected)
                {
                    try
                    {
                        // We don't care about any errors from Close. The camera could be already disconnected
                        CloseDevice();
                    }
                    catch { }
                    Handle = INVALID_HANDLE;
                    IsConnected = false;

                    if (CurrentGPSState == FliGPSState.FPRO_GPS_DETECTED_AND_SAT_LOCK)
                    {
                        CurrentGPSState = FliGPSState.FPRO_GPS_DETECTED_NO_SAT_LOCK;
                        OnPropertyChanged("GPSUnlockedVisibility");
                        OnPropertyChanged("GPSLockedVisibility");
                    }
                }
            }
        }

        public virtual void EnableDriverDebugLogActive(bool enabled)
        {
            DebugEnable = enabled;
            if (enabled)
            {
                DebugLevel = DBGLEVEL.TRACE;
            }
            else
            {
                DebugLevel = DBGLEVEL.INFO;
            }
        }

        public void EnableTemperatureTimer(bool enabled)
        {
            TemperatureStatusTimer.Enabled = enabled;
        }

        public virtual FliRectangle GetBaseExposureArea()
        {
            return new FliRectangle();
        }

        public static string GetClockFromTime(long value, FliProCamera.ExposureUnits_Enum unit)
        {
            string result = String.Empty;
            long factor = 1;
            long hours = 0;
            long mins = 0;
            long secs = 0;
            string valueFormat = "000";
            double newValue = value;
            int divisorFactor = 1;

            switch (unit)
            {
                case ExposureUnits_Enum.units_uSecs:
                    factor = 1;
                    valueFormat = "000000";
                    divisorFactor = 1000;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = MILLISECS_PER_SEC;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = MILLISECS_PER_MIN;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = MILLISECS_PER_HOUR;
                    break;
            }
            newValue *= factor;

            // Value is now in millisecs
            hours = (long)(newValue / (MILLISECS_PER_HOUR * divisorFactor));
            newValue -= hours * (MILLISECS_PER_HOUR * divisorFactor);
            mins = (long)(newValue / (MILLISECS_PER_MIN * divisorFactor));
            newValue -= mins * (MILLISECS_PER_MIN * divisorFactor);
            secs = (long)(newValue / (MILLISECS_PER_SEC * divisorFactor));
            newValue -= secs * (MILLISECS_PER_SEC * divisorFactor);
            result = hours.ToString("00") + ":" + mins.ToString("00") + ":" + secs.ToString("00") + "." + newValue.ToString(valueFormat);
            return result;
        }

        /// <summary>
        /// Read the Cooler Duty Cycle
        /// </summary>
        /// <param name="DutyCycle">The percentage at which the Cooler is running</param>
        /// <returns></returns>
        public virtual int GetCoolerDutyCycle(ref int DutyCycle)
        {
            int status = 0;
            return status;
        }

        public bool GetEnableTemperatureTimer()
        {
            return TemperatureStatusTimer.Enabled;
        }

        public virtual int GetHDREnable(ref bool bIsHDR)
        {
            int status = 0;
            bIsHDR = true;
            return status;
        }

        public virtual int GetTemperatureSetPoint(ref double SetPoint)
        {
            SetPoint = 25;
            return 0;
        }

        public static long GetTimeFromUnits(long value, FliProCamera.ExposureUnits_Enum unit)
        {
            long factor = 1;
            long newValue = value;

            switch (unit)
            {
                case ExposureUnits_Enum.units_uSecs:
                    factor = 1;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1000;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = MILLISECS_PER_SEC * 1000;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = MILLISECS_PER_MIN * 1000;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = (long)MILLISECS_PER_HOUR * 1000;
                    break;
            }
            newValue *= factor;

            // Value is now in millisecs
            return newValue;
        }

        public virtual uint GetFWRevision()
        {
            if ((fliProCameraManager.ActiveCamera.Type == CameraType.GEN1) ||
                (fliProCameraManager.ActiveCamera.Type == CameraType.GEN2))
            {
                return fliProCameraManager.ActiveCamera.GetFWRevision();
            }
            return 0;
        }

        public virtual string GetLibVersion()
        {
            return String.Empty;
        }

        public virtual void GetDeviceVersion(ref string firmwareVersion, ref string fpgaVersion, ref string controllerVersion)
        {
            firmwareVersion = "0.0.0";
            fpgaVersion = "0.0.0";
        }

        /// <summary>
        /// Gets all of the Temperatures on the camera
        /// </summary>
        /// <param name="AmbientTemp">The current Ambient temperature on the sensor</param>
        /// <param name="BaseTemp">The current Base temperature on the sensor</param>
        /// <param name="CoolerTemp">The temperature of the Cooler unit on the camera</param>
        /// <returns></returns>
        public virtual int GetTemperatures(ref double AmbientTemp, ref double BaseTemp, ref double CoolerTemp)
        {
            int status = 0;
            return status;
        }

        public Bitmap BitmapFromSource(BitmapSource bitmapsource)
        {
            Bitmap bitmap;
            using (var outStream = new MemoryStream())
            {
                BitmapEncoder enc = new BmpBitmapEncoder();
                enc.Frames.Add(BitmapFrame.Create(bitmapsource));
                enc.Save(outStream);
                bitmap = new Bitmap(outStream);
            }
            return bitmap;
        }

        public virtual bool CheckGPS()
        {
            return false;
        }

        public string GetValidFilename(string folder, string filename, string extension)
        {
            int fileIndex = 1;
            string newFileName = filename;
            DirectoryInfo di = new DirectoryInfo(folder);
            FileInfo[] TXTFiles = null;

            try
            {
                TXTFiles = di.GetFiles(filename + "*" + extension);

                List<int> indexList = new List<int>();
                for (int i = 0; i < TXTFiles.Length; i++)
                {
                    string tempName = TXTFiles[i].Name.Substring(filename.Length, TXTFiles[i].Name.Length - (filename.Length + 4)); // subtract off ".bin"
                    try
                    {
                        indexList.Add(Convert.ToInt32(tempName));
                    }
                    catch { }
                }
                while (indexList.Contains(fileIndex))
                {
                    fileIndex++;
                }
            }
            catch { }

            return folder + "\\" + newFileName + ((TXTFiles != null) && (TXTFiles.Length > 0) ? fileIndex.ToString() : String.Empty) + extension;
        }

        private int originalStatusLength = 0;
        public void PublishStatus(string text, bool isAppend)
        {
            if (isAppend && !String.IsNullOrEmpty(CameraStatusText))
            {
                CameraStatusText = CameraStatusText.Substring(0, originalStatusLength);
                CameraStatusText += text;
            }
            else
            {
                CameraStatusText = text;
                originalStatusLength = text.Length;
            }
        }

        /// <summary>
        /// Reads the Temperature of the given channel
        /// </summary>
        /// <param name="channel">The channel from which the temperature is to be read</param>
        /// <param name="temp">The returned temperature</param>
        /// <returns></returns>
        public virtual int ReadTemperature(int channel, ref double temp)
        {
            int status = 0;
            return status;
        }

        public void ResetToFactoryDefaults()
        {
            OnConnect();
        }

        public void SetDefaultFliFrameCountList()
        {
            FliFrameCountList = new System.Collections.ObjectModel.ObservableCollection<FliFrameCountNameValueClass>();
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Single", FliFrameCount.SINGLE));
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Multiple", FliFrameCount.MULTIPLE));
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Continuous", FliFrameCount.CONTINUOUS));

            FrameCountType = FliFrameCount.SINGLE;
        }

        public void SetDefaultFliFrameTypeList()
        {
            FliFrameTypeList = new System.Collections.ObjectModel.ObservableCollection<FliFrameTypeNameValueClass>();
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Normal", FliFrameType.NORMAL));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Flat", FliFrameType.FLAT));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Dark", FliFrameType.DARK));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Flood", FliFrameType.FLOOD));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("RBI_Flush", FliFrameType.RBI_FLUSH));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Bias", FliFrameType.BIAS));

            FrameType = FliFrameType.NORMAL;
        }

        public void SetDefaultGainTables()
        {
            DNETableLG = new ObservableCollection<SensorGain>();
            DNETableLG.Add(new SensorGain(0, 0.0201, 0));
            DNETableLG.Add(new SensorGain(4, 0.0401, 1));
            DNETableLG.Add(new SensorGain(1, 0.0576, 2));
            DNETableLG.Add(new SensorGain(2, 0.0788, 3));
            DNETableLG.Add(new SensorGain(3, 0.1162, 4));
            DNETableLG.Add(new SensorGain(5, 0.1128, 5));
            DNETableLG.Add(new SensorGain(6, 0.152, 6));
            DNETableLG.Add(new SensorGain(7, 0.206, 7));

            DNETableLGFactory = new ObservableCollection<SensorGain>();
            DNETableLGFactory.Add(new SensorGain(0, 0.0201, 0));
            DNETableLGFactory.Add(new SensorGain(4, 0.0401, 1));
            DNETableLGFactory.Add(new SensorGain(1, 0.0576, 2));
            DNETableLGFactory.Add(new SensorGain(2, 0.0788, 3));
            DNETableLGFactory.Add(new SensorGain(3, 0.1162, 4));
            DNETableLGFactory.Add(new SensorGain(5, 0.1128, 5));
            DNETableLGFactory.Add(new SensorGain(6, 0.152, 6));
            DNETableLGFactory.Add(new SensorGain(7, 0.206, 7));

            DNETableHG = new ObservableCollection<SensorGain>();
            DNETableHG.Add(new SensorGain(0, 0.1316, 0));
            DNETableHG.Add(new SensorGain(4, 0.2633, 1));
            DNETableHG.Add(new SensorGain(1, 0.3897, 2));
            DNETableHG.Add(new SensorGain(2, 0.5717, 3));
            DNETableHG.Add(new SensorGain(3, 0.9178, 4));
            DNETableHG.Add(new SensorGain(5, 0.9088, 5));
            DNETableHG.Add(new SensorGain(6, 1.2332, 6));
            DNETableHG.Add(new SensorGain(7, 1.8022, 7));

            DNETableHGFactory = new ObservableCollection<SensorGain>();
            DNETableHGFactory.Add(new SensorGain(0, 0.1316, 0));
            DNETableHGFactory.Add(new SensorGain(4, 0.2633, 1));
            DNETableHGFactory.Add(new SensorGain(1, 0.3897, 2));
            DNETableHGFactory.Add(new SensorGain(2, 0.5717, 3));
            DNETableHGFactory.Add(new SensorGain(3, 0.9178, 4));
            DNETableHGFactory.Add(new SensorGain(5, 0.9088, 5));
            DNETableHGFactory.Add(new SensorGain(6, 1.2332, 6));
            DNETableHGFactory.Add(new SensorGain(7, 1.8022, 7));

            LowGainTable = new ObservableCollection<SensorGain>();
            LowGainTable.Add(new SensorGain(0, 660, 0));
            LowGainTable.Add(new SensorGain(4, 1290, 1));
            LowGainTable.Add(new SensorGain(1, 1850, 2));
            LowGainTable.Add(new SensorGain(2, 2490, 3));
            LowGainTable.Add(new SensorGain(3, 3680, 4));
            LowGainTable.Add(new SensorGain(5, 3700, 5));
            LowGainTable.Add(new SensorGain(6, 4950, 6));
            LowGainTable.Add(new SensorGain(7, 7250, 7));

            HighGainTable = new ObservableCollection<SensorGain>();
            HighGainTable.Add(new SensorGain(0, 660, 0));
            HighGainTable.Add(new SensorGain(4, 1290, 1));
            HighGainTable.Add(new SensorGain(1, 1850, 2));
            HighGainTable.Add(new SensorGain(2, 2490, 3));
            HighGainTable.Add(new SensorGain(3, 3680, 4));
            HighGainTable.Add(new SensorGain(5, 3700, 5));
            HighGainTable.Add(new SensorGain(6, 4950, 6));
            HighGainTable.Add(new SensorGain(7, 7250, 7));

            RestoreMergeSettingsFromFactorDefaults();

            GainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];
            ManufacturersDefaultGainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];

            for (int l = 0; l < DNETableLG.Count; l++)
            {
                int epcLowIndex = GetFactoryElectronsPerCountLowIndex(DNETableLG[l].Index);
                if (epcLowIndex >= 0)
                {
                    DNETableLG[l].Gain = FactoryElectronsPerCountLowGainList[epcLowIndex].ElectronsPerCount;
                }

                for (int h = 0; h < DNETableHG.Count; h++)
                {
                    if (l == 0)
                    {
                        int epcHighIndex = GetFactoryElectronsPerCountHighIndex(DNETableHG[h].Index);
                        if (epcHighIndex >= 0)
                        {
                            DNETableHG[h].Gain = FactoryElectronsPerCountHighGainList[epcHighIndex].ElectronsPerCount;
                        }
                    }

                    int index = GetMergeLineOffsetIndex(DNETableHG[h].Index, DNETableLG[l].Index);

                    // These are filled in from the camera
                    if (index < 0)
                    {
                        // Divide the Low gain by the High gain because the numbers are in 1/x form
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = DNETableLG[l].Gain / DNETableHG[h].Gain;
                        // There is no easy way to calculate the slope so I just figured out an equation that yields a "reasonable" value.
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = -10;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 120);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = 5;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 125);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = 20;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 130);
                    }
                    else
                    {
                        // Use the current default
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = MergeLineOffsetList[index].GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = MergeLineOffsetList[index].Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = MergeLineOffsetList[index].LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = MergeLineOffsetList[index].Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = MergeLineOffsetList[index].LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = MergeLineOffsetList[index].Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                    }

                    GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;

                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;
                }
            }

            LowGainValue = LowGainTable[0];
            HighGainValue = HighGainTable[0];
        }

        public void SetDefaultGainTablesCobaltDC230(bool initialize)
        {
            if (initialize)
            {
                DNETableLG = new ObservableCollection<SensorGain>();
                DNETableLG.Add(new SensorGain(0, 1.000, 0));

                DNETableLGFactory = new ObservableCollection<SensorGain>();
                DNETableLGFactory.Add(new SensorGain(0, 1.000, 0));

                DNETableHG = new ObservableCollection<SensorGain>();
                DNETableHG.Add(new SensorGain(0, 1.000, 0));

                DNETableHGFactory = new ObservableCollection<SensorGain>();
                DNETableHGFactory.Add(new SensorGain(0, 1.000, 0));

                LowGainTable = new ObservableCollection<SensorGain>();
                LowGainTable.Add(new SensorGain(0, 1000, 0));
                LowGainValue = LowGainTable[0];

                HighGainTable = new ObservableCollection<SensorGain>();
                HighGainTable.Add(new SensorGain(0, 1000, 0));
                HighGainValue = HighGainTable[0];
            }
        }

        public void SetDefaultGainTablesKepler0400(bool initialize)
        {
            if (initialize)
            {
                DNETableLG = new ObservableCollection<SensorGain>();
                DNETableLG.Add(new SensorGain(0, 47.8469, 0));
                DNETableLG.Add(new SensorGain(4, 24.9377, 1));
                DNETableLG.Add(new SensorGain(1, 17.3611, 2));
                DNETableLG.Add(new SensorGain(2, 12.6904, 3));
                DNETableLG.Add(new SensorGain(3, 8.6059, 4));
                DNETableLG.Add(new SensorGain(5, 8.8652, 5));
                DNETableLG.Add(new SensorGain(6, 6.5789, 6));
                DNETableLG.Add(new SensorGain(7, 4.8544, 7));

                DNETableLGFactory = new ObservableCollection<SensorGain>();
                DNETableLGFactory.Add(new SensorGain(0, 47.8469, 0));
                DNETableLGFactory.Add(new SensorGain(4, 24.9377, 1));
                DNETableLGFactory.Add(new SensorGain(1, 17.3611, 2));
                DNETableLGFactory.Add(new SensorGain(2, 12.6904, 3));
                DNETableLGFactory.Add(new SensorGain(3, 8.6059, 4));
                DNETableLGFactory.Add(new SensorGain(5, 8.8652, 5));
                DNETableLGFactory.Add(new SensorGain(6, 6.5789, 6));
                DNETableLGFactory.Add(new SensorGain(7, 4.8544, 7));

                DNETableHG = new ObservableCollection<SensorGain>();
                DNETableHG.Add(new SensorGain(0, 7.5988, 0));
                DNETableHG.Add(new SensorGain(4, 3.7979, 1));
                DNETableHG.Add(new SensorGain(1, 2.5661, 2));
                DNETableHG.Add(new SensorGain(2, 1.7492, 3));
                DNETableHG.Add(new SensorGain(3, 1.0896, 4));
                DNETableHG.Add(new SensorGain(5, 1.1004, 5));
                DNETableHG.Add(new SensorGain(6, 0.8109, 6));
                DNETableHG.Add(new SensorGain(7, 0.5549, 7));

                DNETableHGFactory = new ObservableCollection<SensorGain>();
                DNETableHGFactory.Add(new SensorGain(0, 7.5988, 0));
                DNETableHGFactory.Add(new SensorGain(4, 3.7979, 1));
                DNETableHGFactory.Add(new SensorGain(1, 2.5661, 2));
                DNETableHGFactory.Add(new SensorGain(2, 1.7492, 3));
                DNETableHGFactory.Add(new SensorGain(3, 1.0896, 4));
                DNETableHGFactory.Add(new SensorGain(5, 1.1004, 5));
                DNETableHGFactory.Add(new SensorGain(6, 0.8109, 6));
                DNETableHGFactory.Add(new SensorGain(7, 0.5549, 7));

                LowGainTable = new ObservableCollection<SensorGain>();
                LowGainTable.Add(new SensorGain(0, 660, 0));
                LowGainTable.Add(new SensorGain(4, 1290, 1));
                LowGainTable.Add(new SensorGain(1, 1850, 2));
                LowGainTable.Add(new SensorGain(2, 2490, 3));
                LowGainTable.Add(new SensorGain(3, 3680, 4));
                LowGainTable.Add(new SensorGain(5, 3700, 5));
                LowGainTable.Add(new SensorGain(6, 4950, 6));
                LowGainTable.Add(new SensorGain(7, 7250, 7));
                LowGainValue = LowGainTable[0];

                HighGainTable = new ObservableCollection<SensorGain>();
                HighGainTable.Add(new SensorGain(0, 660, 0));
                HighGainTable.Add(new SensorGain(4, 1290, 1));
                HighGainTable.Add(new SensorGain(1, 1850, 2));
                HighGainTable.Add(new SensorGain(2, 2490, 3));
                HighGainTable.Add(new SensorGain(3, 3680, 4));
                HighGainTable.Add(new SensorGain(5, 3700, 5));
                HighGainTable.Add(new SensorGain(6, 4950, 6));
                HighGainTable.Add(new SensorGain(7, 7250, 7));
                HighGainValue = HighGainTable[0];
            }

            RestoreMergeSettingsFromFactorDefaults();

            if (ActiveElectronsPerCountLow == 0)
            {
                // Then we haven't initialized this camera yet so do it now
                HighGainIndex = 7;
                LowGainIndex = 4;
                ActiveElectronsPerCountHigh = 0.5549;
                ActiveElectronsPerCountLow = 24.9377;
                Temperature1 = -10;
                Temperature2 = 5;
                Temperature3 = 20;
                LineOffset1 = 3500;
                LineOffset2 = 3500;
                LineOffset3 = 3500;
                GainRatio = 40;
                ADUsPerElectron = 1;
                BlackLevelHigh = 15750;
                BlackLevelLow = 15580;
                SaveMergeSettingsAsFactorDefaults();
            }
            GainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];
            ManufacturersDefaultGainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];

            // Because of the way we handle Gain indecies on the 400 we get back the array index in the Metadata.
            // That being the case we need to use the array index to populate the GainOffsetPairArray
            for (int l = 0; l < DNETableLG.Count; l++)
            {
                int epcLowIndex = GetFactoryElectronsPerCountLowIndex(DNETableLG[l].Index);
                if (epcLowIndex >= 0)
                {
                    DNETableLG[l].Gain = FactoryElectronsPerCountLowGainList[epcLowIndex].ElectronsPerCount;
                }

                for (int h = 0; h < DNETableHG.Count; h++)
                {
                    if (l == 0)
                    {
                        int epcHighIndex = GetFactoryElectronsPerCountHighIndex(DNETableHG[h].Index);
                        if (epcHighIndex >= 0)
                        {
                            DNETableHG[h].Gain = FactoryElectronsPerCountHighGainList[epcHighIndex].ElectronsPerCount;
                        }
                    }

                    int index = GetMergeLineOffsetIndex(DNETableHG[h].Index, DNETableLG[l].Index);

                    // These are filled in from the camera
                    if (index < 0)
                    {
                        // Divide the Low gain by the High gain because the numbers are in 1/x form
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = Math.Round(DNETableLG[l].Gain / DNETableHG[h].Gain, 4);
                        // There is no easy way to calculate the slope so I just figured out an equation that yields a "reasonable" value.
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = -10;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 120);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = 5;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 125);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = 20;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 130);
                    }
                    else
                    {
                        // Use the current default
                       GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = MergeLineOffsetList[index].GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = MergeLineOffsetList[index].Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = MergeLineOffsetList[index].LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = MergeLineOffsetList[index].Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = MergeLineOffsetList[index].LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = MergeLineOffsetList[index].Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                    }

                    GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;

                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;
                }
            }
        }

        public void SetDefaultGainTablesKepler4040(bool initialize)
        {
            if (initialize)
            {
                DNETableLG = new ObservableCollection<SensorGain>();
                DNETableHG = new ObservableCollection<SensorGain>();
                LowGainTable = new ObservableCollection<SensorGain>();
                HighGainTable = new ObservableCollection<SensorGain>();

                RestoreMergeSettingsFromFactorDefaults();

                for (int i = 0; i < 64; i++)
                {
                    double value = 0;
                    if ((i % 2) == 0)
                    {
                        value = 1.0 + (i / 2) * 0.5;
                    }
                    else
                    {
                        value = 0.2 + ((i + 1) / 2) * 0.2;
                    }
                    LowGainTable.Add(new SensorGain(i, value, i));
                    HighGainTable.Add(new SensorGain(i, value, i));

                    // We only have a few values from the Sensor manufacturer so use them
                    // and default the rest to Low Index 25 (2.800) and High Index 62 (16.500).
                    double LowElectronsPerCount = ActiveElectronsPerCountLow;
                    double HighElectronsPerCount = ActiveElectronsPerCountHigh;

                    switch ((int)(value * 1000))
                    {
                        case 2800:
                            LowElectronsPerCount = 18.6916;
                            HighElectronsPerCount = 18.6916;
                            break;
                        case 5200:
                            LowElectronsPerCount = 10;
                            HighElectronsPerCount = 10;
                            break;
                        case 6600:
                            LowElectronsPerCount = 8;
                            HighElectronsPerCount = 8;
                            break;
                        case 13500:
                            LowElectronsPerCount = 1;
                            HighElectronsPerCount = 1;
                            break;
                        case 14000:
                            LowElectronsPerCount = 4;
                            HighElectronsPerCount = 4;
                            break;
                        case 16500:
                            LowElectronsPerCount = 0.8194;
                            HighElectronsPerCount = 0.8194;
                            break;
                }
                    DNETableLG.Add(new SensorGain(i, LowElectronsPerCount, i));
                    DNETableHG.Add(new SensorGain(i, HighElectronsPerCount, i));
                    DNETableLGFactory.Add(new SensorGain(i, LowElectronsPerCount, i));
                    DNETableHGFactory.Add(new SensorGain(i, HighElectronsPerCount, i));
            }
            }

            if (ActiveElectronsPerCountLow == 0)
            {
                // Then we haven't initialized this camera yet so do it now
                HighGainIndex = 62;
                LowGainIndex = 25;
                ActiveElectronsPerCountHigh = 0.8194;
                ActiveElectronsPerCountLow = 18.6916;
                Temperature1 = -10;
                Temperature2 = 5;
                Temperature3 = 20;
                LineOffset1 = 1700;
                LineOffset2 = 1750;
                LineOffset3 = 1800;
                GainRatio = 20.1;
                ADUsPerElectron = 1;
                BlackLevelHigh = 15750;
                BlackLevelLow = 15580;
                SaveMergeSettingsAsFactorDefaults();
            }

            GainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];
            ManufacturersDefaultGainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];

            for (int l = 0; l < DNETableLG.Count; l++)
            {
                int epcLowIndex = GetFactoryElectronsPerCountLowIndex(DNETableLG[l].Index);
                if (epcLowIndex >= 0)
                {
                    DNETableLG[l].Gain = FactoryElectronsPerCountLowGainList[epcLowIndex].ElectronsPerCount;
                }

                for (int h = 0; h < DNETableHG.Count; h++)
                {
                    if (l == 0)
                    {
                        int epcHighIndex = GetFactoryElectronsPerCountHighIndex(DNETableHG[h].Index);
                        if (epcHighIndex >= 0)
                        {
                            DNETableHG[h].Gain = FactoryElectronsPerCountHighGainList[epcHighIndex].ElectronsPerCount;
                        }
                    }

                    int index = GetMergeLineOffsetIndex(DNETableHG[h].Index, DNETableLG[l].Index);

                    // These are filled in from the camera
                    if (index < 0)
                    {
                        // There is no default for this pair so use the current settings
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = LineOffset3;
                    }
                    else
                    {
                       // Use the current default
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = MergeLineOffsetList[index].GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = MergeLineOffsetList[index].Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = MergeLineOffsetList[index].LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = MergeLineOffsetList[index].Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = MergeLineOffsetList[index].LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = MergeLineOffsetList[index].Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                    }

                    GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;

                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;
                }
            }

            LowGainValue = GetGainValueFromIndex(LowGainTable, 25);
            HighGainValue = GetGainValueFromIndex(HighGainTable, 62);
        }

        public SensorGain GetGainValueFromIndex(ObservableCollection<SensorGain> gainTable, int gainIndex)
        {
            SensorGain sensorGain = new SensorGain(0, 1.0, 0); // Need to create a default in case the target index is not found
            if (gainTable != null)
            {
                for (int i = 0; i < gainTable.Count; i++)
                {
                    if (gainTable[i].Index == gainIndex)
                    {
                        sensorGain = gainTable[i];
                        break;
                    }
                }
            }

            return sensorGain;
        }

        public void UpdateDefaultGainTablesKepler4040()
        {
            DNETableLG = new ObservableCollection<SensorGain>();
            DNETableHG = new ObservableCollection<SensorGain>();

            for (int i = 0; i < 64; i++)
            {
                // We only have these two values from the Sensor manufacturer 
                // which are for Low Index 25 (2.800) and High Index 62 (16.500).
                // So just use them for all the Gains.
                DNETableLG.Add(new SensorGain(i, ActiveElectronsPerCountLow, i));
                DNETableHG.Add(new SensorGain(i, ActiveElectronsPerCountHigh, i));
                DNETableLGFactory.Add(new SensorGain(i, ActiveElectronsPerCountLow, i));
                DNETableHGFactory.Add(new SensorGain(i, ActiveElectronsPerCountHigh, i));
            }

            RestoreMergeSettingsFromFactorDefaults();

            GainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];
            ManufacturersDefaultGainOffsetPairArray = new GainOffsetPair[DNETableLG.Count, DNETableHG.Count];

            for (int l = 0; l < DNETableLG.Count; l++)
            {
                int epcLowIndex = GetFactoryElectronsPerCountLowIndex(DNETableLG[l].Index);
                if (epcLowIndex >= 0)
                {
                    DNETableLG[l].Gain = FactoryElectronsPerCountLowGainList[epcLowIndex].ElectronsPerCount;
                }

                for (int h = 0; h < DNETableHG.Count; h++)
                {
                    if (l == 0)
                    {
                        int epcHighIndex = GetFactoryElectronsPerCountHighIndex(DNETableHG[h].Index);
                        if (epcHighIndex >= 0)
                        {
                            DNETableHG[h].Gain = FactoryElectronsPerCountHighGainList[epcHighIndex].ElectronsPerCount;
                        }
                    }

                    int index = GetMergeLineOffsetIndex(DNETableHG[h].Index, DNETableLG[l].Index);

                    // These are filled in from the camera
                    if (index < 0)
                    {
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = DNETableHG[h].Gain / DNETableLG[l].Gain;
                        // There is no easy way to calculate the slope so I just figured out an equation that yields a "reasonable" value.
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = -10;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 120);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = 5;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 125);
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = 20;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = (int)(GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio * 130);
                    }
                    else
                    {
                        // Use the current default
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = MergeLineOffsetList[index].GainRatio;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = MergeLineOffsetList[index].Temperature1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = MergeLineOffsetList[index].LineOffset1;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = MergeLineOffsetList[index].Temperature2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = MergeLineOffsetList[index].LineOffset2;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = MergeLineOffsetList[index].Temperature3;
                        GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = MergeLineOffsetList[index].LineOffset3;
                    }

                    GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;

                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].gainRatio;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset1;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset2;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].temperature3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3 = GainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].lineOffset3;
                    ManufacturersDefaultGainOffsetPairArray[DNETableLG[l].Index, DNETableHG[h].Index].stdDev = 0;
                }
            }
        }

        public void SetDefaultSupportedBinRatios()
        {
            SupportedBinRatios.Add(new BinRatio(0, 1, 1));
            BinRatioValue = SupportedBinRatios[0];
        }

        public virtual void SetDefaultSupportedPixelDepths()
        {
            SupportedPixelDepths.Add(12);
            PixelDepth = SupportedPixelDepths[0];
        }
        
        public void SetDefaultSupportedModes()
        {
            SupportedModes = new ObservableCollection<NameValuePair>();
            SupportedModes.Add(new NameValuePair("Rolling LDR - LDC", 1));
            SupportedModes.Add(new NameValuePair("Rolling HDR - LDC", 3));
            SupportedModes.Add(new NameValuePair("Rolling LDR", 0));
            SupportedModes.Add(new NameValuePair("Rolling HDR", 2));
            Mode = 0;
            CameraMode = SupportedModes[Mode].Name;
        }

        public void SetEnabled(bool enabled)
        {
            IsEnabled = enabled;
        }

        public void SetForcePropertyChange(bool isForce)
        {
            if (fliProCameraManager == null)
            {
                fliProCameraManager = FliProCameraManager.Instance;
            }

            ForcePropertyChange = isForce;
        }

        private int tempCounter = 0;

        private void TemperatureStatusTimerEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            int status = 0;
            string TempString = String.Empty;

            // 
            // We want to only ask for temperature if the image is not capturing. 
            // And we want to prevent asking for image capture while we are getting the temperature. 
            // That means we have to lock during this entire routine but we only have to lock the image capture when setting the IsCapturing flag.
            lock (ImageCaptureLock)
            {
                try
                {
                    if (IsCapturing && !ReadDuringExposureOn)
                    {
                        // We don't want to read the Temperature register while the camera is capturing because this causes the sensor to fail.
                        return;
                    }

                    if ((IsConnected == false) || (ReadErrorEncountered == true))
                    {
                        TempString = "Temp: N/A";
                    }
                    else
                    {
                        double ambientTemp = 0;
                        double baseTemp = 0;
                        double coolerTemp = 0;

                        status = GetTemperatures(ref ambientTemp, ref baseTemp, ref coolerTemp);
                        tempCounter++;
                        Temperature = coolerTemp;
                        TemperatureText = Temperature.ToString("#0.0");

                        if (status < 0)
                        {
                            // Turn off checking for temperature on the timer
                            // TemperatureStatusTimer.Enabled = false;         // We need to keep this on so we can read the Xylinx chip temperature

                            // We don't care if we get an error because we are shutting down
                            try
                            {
                                // I don't think we want to shut down the camera just because we had a status error
                                //CloseDevice();
                            }
                            catch
                            {
                                // No harm in continuing to read but post the error
                                //ReadErrorEncountered = true;
                            }
                        }
                        else
                        {
                            if (fliProCameraManager.IsCheckedAutoForceTraining && !FirstTraining)
                            {
                                if (!fliProCameraManager.AutoForceTrainingTargetReached && ((Temperature > TargetTemperature - 0.2) && (Temperature < TargetTemperature + 0.2)))
                                {
                                    // We have reached our target so set the Target reached flag and run one last Forced Training
                                    fliProCameraManager.AutoForceTrainingTargetReached = true;
                                    fliProCameraManager.AutoForceTrainingLastTargetTemperature = TargetTemperature;
                                    ForceTraining(this);
                                }
                                else if (!fliProCameraManager.AutoForceTrainingTargetReached &&
                                    ((Temperature > fliProCameraManager.AutoForceTrainingLastTargetTemperature - fliProCameraManager.AutoForceRangeTemperature) &&
                                    (Temperature < fliProCameraManager.AutoForceTrainingLastTargetTemperature + fliProCameraManager.AutoForceRangeTemperature)))
                                {
                                    // We haven't reached our target but we are still within AutoForceRangeTemperature so don't ForceTraining 
                                    if (CameraStatusText.Contains("Force Training"))
                                    {
                                        PublishStatus("", false);    // Clear the last status message
                                    }
                                }
                                else if ((Temperature < fliProCameraManager.AutoForceTrainingLastTargetTemperature - fliProCameraManager.AutoForceRangeTemperature) ||
                                    (Temperature > fliProCameraManager.AutoForceTrainingLastTargetTemperature + fliProCameraManager.AutoForceRangeTemperature))
                                {
                                    // We've moved outside of the range so send ForceTraining
                                    fliProCameraManager.AutoForceTrainingTargetReached = false;
                                    fliProCameraManager.AutoForceTrainingLastTargetTemperature = Temperature;
                                    ForceTraining(this);
                                }
                                else
                                {
                                    // Just clear the status 
                                    if (CameraStatusText.Contains("Force Training"))
                                    {
                                        PublishStatus("", false);    // Clear the last status message
                                    }
                                }
                            }

                            FirstTraining = false;

                            int cooler = 0;
                            status = GetCoolerDutyCycle(ref cooler);

                            if (status == 0)
                            {
                                if (baseTemp > (-100.0))
                                {
                                    TempString = "C/B: " + TemperatureText + " / " + baseTemp.ToString("#0.0") +
                                        "(" + Convert.ToInt16(cooler).ToString() + "%)";
                                }
                                else
                                {
                                    TempString = "CCD: " + TemperatureText + "(" + Convert.ToUInt16(cooler).ToString() + "%)";
                                }
                            }
                            else
                            {
                                TempString = "CCD: " + TemperatureText + " BASE: " + baseTemp.ToString();
                            }
                        }

                        // If the State of the GPS has changed we need to update the display
                        if (CheckGPS())
                        {
                            OnPropertyChanged("GPSUnlockedVisibility");
                            OnPropertyChanged("GPSLockedVisibility");
                            OnPropertyChanged("GPSConnectedVisibility");
                            OnPropertyChanged("GPSLockWidth");
                            OnPropertyChanged("GPSWidth");
                        }
                    }
                }
                catch
                {
                    // Turn off checking for temperature on the timer
                    //TemperatureStatusTimer.Enabled = false;           // We need to keep this on so we can read the Xylinx chip temperature

                    // No harm in continuing to read but post the error
                    //ReadErrorEncountered = true;
                    TempString = "Temp: N/A";
                }

                TemperatureStatusText = TempString;
            }
        }

        private bool ImageCaptureActive()
        {
            bool isActive = false;

            lock (ImageCaptureLock)
            {
                isActive = IsCapturing;
            }
            return isActive;
        }


        //-------------------------------------------
        /// <summary>
        /// Default string representation for a camera
        /// </summary>
        public override string ToString()
        {
            if (SerialNumber.Length > 0)
            {
                return Model + GetDeviceType() + ", " + SerialNumber + Speed;
            }
            else
            {
                return Model + GetDeviceType();
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Get the description string for the specified camera mode index
        /// </summary>
        /// <param name="mode">camera mode index</param>
        /// <returns>string or "" if mode index is not supported</returns>
        public virtual string GetCameraModeString(int mode)
        {
            return "";
        }

        public void WriteFileBasic(FliFrame frame, string folder)
        {
            if (!String.IsNullOrEmpty(folder))
            {
                lock (ImageWriteLock)
                {
                    try
                    {
                        string fileNameBase = fliProCameraManager.CalculateNameString(frame);
                        string fileName = String.Empty;
                        // Save the Combined file which is exactly how we received it from the board.
                        fileName = GetValidFilename(folder, fileNameBase + "RCDData_", FILE_EXTENSION_RCD);
                        SaveBINFile(frame, fileName, File_Type.RawData);
                    }
                    catch
                    {
                    }
                }
            }
        }

        public void SaveBINFile(FliFrame frame, string fileName, File_Type fileType)
        {
            try
            {
                FileStream fs = new FileStream(fileName, FileMode.CreateNew);
                // Create the writer for data.
                BinaryWriter w = new BinaryWriter(fs);

                if ((frame.RawData == null) && ((fileType == File_Type.RawData) || (fileType == File_Type.Low)))
                {
                    return;
                }
                else   // This is a new camera
                {
                    switch (fileType)
                    {
                        case File_Type.RawData:
                            if (frame.RawData != null)
                            {
                                w.Write(frame.RawData);
                            }
                            break;
                    }
                }
                w.Close();
                fs.Close();
            }
            catch { }
        }

        #endregion Public Methods

        //**************************************************************
        #region Protected Methods

        //-------------------------------------------
        /// <summary>
        /// Cleanup prior to disposing unmanaged objects
        /// </summary>
        protected override void OnDispose()
        {
            Disconnect();
            base.OnDispose();
        }

        protected void OnConnectBase()
        {
            Shutter = FliShutterControl.CLOSE;

            // Turn On The Temperature Timer
            ReadErrorEncountered = false;
            TemperatureStatusTimer.Enabled = true;
        }

        public void SetIsCapturingLocked(bool isCapturing)
        {
            lock(ImageCaptureLock)
            {
                IsCapturing = isCapturing;
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Update all dynamic device properties and status
        /// </summary>
        protected void UpdateCameraStatus()
        {
            try
            {
                lock (DeviceLock)
                {
                    if (IsConnected)
                    {
                        UpdateStatus();
                        StatusUpdateCount++;
                    }
                }
            }
            catch (Exception)
            {
                // Ignore for now
            }
        }

        protected virtual void Initialize()
        {
        }

        protected virtual int OpenDevice()
        {
            return 0;
        }

        protected virtual void CloseDevice() { }
        protected virtual void ReadNVStorage(uint uiOffset, byte[] pData, uint uiLength) { }
        protected virtual void WriteNVStorage(uint uiOffset, byte[] pData, uint uiLength) { }

        //-------------------------------------------
        /// <summary>
        /// Update dynamically changing values
        /// </summary>
        protected virtual void UpdateStatus()
        {
        }

        //-------------------------------------------
        /// <summary>
        /// Device Specific actions when connection is established
        /// </summary>
        protected virtual unsafe void OnConnect()
        {
            SerialNumber = "";
            HwVersion = 0;
            FwVersion = 0;
            Model = "";

            ArrayArea = new FliRectangle();
            VisibleArea = new FliRectangle();
            ImageArea = VisibleArea;
            PixelSize = new FliPixelSize();
            ColorPattern = FliBayerPattern.NONE;
            Shutter = FliShutterControl.CLOSE;
            TDIRate = 0;
            HBin = 1;
            VBin = 1;
            FrameCountType = FliFrameCount.SINGLE;
            FrameType = FliFrameType.NORMAL;
            NFlushes = 1;
            VideoModeEnabled = false;
        }

        protected virtual unsafe void OnConnectSimple()
        {
            SerialNumber = "";
            HwVersion = 0;
            FwVersion = 0;
            Model = "";

            ArrayArea = new FliRectangle();
            VisibleArea = new FliRectangle();
            ImageArea = VisibleArea;
            PixelSize = new FliPixelSize();
            ColorPattern = FliBayerPattern.NONE;
            Shutter = FliShutterControl.CLOSE;
            TDIRate = 0;
            HBin = 1;
            VBin = 1;
            FrameCountType = FliFrameCount.SINGLE;
            FrameType = FliFrameType.NORMAL;
            NFlushes = 1;
            VideoModeEnabled = false;
        }


        #endregion Protected Methods

        //**************************************************************
        #region Public Events

        public event EventHandler<EventArgs> ADUsPerElectronChangeAction;

        protected virtual void OnADUsPerElectronChangeAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = ADUsPerElectronChangeAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> CaptureCompleteAction;

        protected virtual void OnCaptureCompleteAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = CaptureCompleteAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }
        
        public event EventHandler<EventArgs> GainRatioChangeAction;

        protected virtual void OnGainRatioChangeAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = GainRatioChangeAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> LEDAction;

        protected virtual void OnLEDAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = LEDAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> MergeRangeChangeAction;

        protected virtual void OnMergeRangeChangeAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = MergeRangeChangeAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> LineOffsetChangeAction;

        protected virtual void OnLineOffsetChangeAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = LineOffsetChangeAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        

        #endregion Public Events
    }
}
