﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Runtime.InteropServices;
using System.Windows;
using System.Windows.Data;
using FliUtils.Mvvm;
using FliUtils.Usb;
using System.Threading;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Windows.Input;
using System.Reflection;
using System.Xml.Serialization;
using System.Security.AccessControl;
using System.Security.Principal;
using static FliProCameraLib.FliProCamera;

namespace FliProCameraLib
{
    /// <summary>
    /// This is a singleton class that manages enumeration, opening, and closing
    /// of Finger Lakes Imaging cameras. It includes methods and properties
    /// that provide dotnet friendly wrappers around LibFLi Native functions.
    /// 
    /// All error codes returned by the native methods are thrown
    /// as FliExceptions 
    /// 
    /// </summary>
    public class FliProCameraManager : FliPro
    {
        //**************************************************************
        #region Constants and Enums

        // Maximum number of devices in a device list
        private const int MAX_DEVICE_COUNT = 20;
        public const string TOOL_SETTINGS_DEFAULT_FILE = "DefaultSettings.xml";
        public const string TOOL_SETTINGS_PREVIOUS_FILE = "PreviousSettings.xml";
        private const string TOOL_SETTINGS_FOLDER = @"FliPilot\Settings";
        private const string TOOL_SEQUENCES_FOLDER = @"FliPilot\Sequences";
        private const string TOOL_TOOL_SETTINGS_FOLDER = @"FliPilot\ToolSettings";

        public bool USE_D3D11_IMAGE = false;

        // File Name Setup 
        public static int DEFAULT_ORDER_BINNING = 7;
        public static int DEFAULT_ORDER_CAMERA_MODEL = 4;
        public static int DEFAULT_ORDER_DATE = 11;
        public static int DEFAULT_ORDER_DIAG_ROUTINE = 1;
        public static int DEFAULT_ORDER_EXPOSURE_TIME = 9;
        public static int DEFAULT_ORDER_FRAME_NUMBER = 10;
        public static int DEFAULT_ORDER_FRAME_TYPE = 2;
        public static int DEFAULT_ORDER_GAIN = 8;
        public static int DEFAULT_ORDER_MODE = 5;
        public static int DEFAULT_ORDER_SENSOR_TEMPERATURE = 6;
        public static int DEFAULT_ORDER_SERIAL_NUMBER = 3;

        public static bool DEFAULT_IS_CHECKED_BINNING = false;
        public static bool DEFAULT_IS_CHECKED_CAMERA_MODEL = false;
        public static bool DEFAULT_IS_CHECKED_DATE = false;
        public static bool DEFAULT_IS_CHECKED_DIAG_ROUTINE = true;
        public static bool DEFAULT_IS_CHECKED_EXPOSURE_TIME = false;
        public static bool DEFAULT_IS_CHECKED_FRAME_NUMBER = false;
        public static bool DEFAULT_IS_CHECKED_FRAME_TYPE = false;
        public static bool DEFAULT_IS_CHECKED_GAIN = false;
        public static bool DEFAULT_IS_CHECKED_MEAN = false;
        public static bool DEFAULT_IS_CHECKED_MODE = false;
        public static bool DEFAULT_IS_CHECKED_PREAMBLE = false;
        public static bool DEFAULT_IS_CHECKED_SENSOR_TEMPERATURE = false;
        public static bool DEFAULT_IS_CHECKED_SERIAL_NUMBER = false;

        public static string FILTER_WHEEL_CONNECT = "Connect";
        public static string FILTER_WHEEL_DISCONNECT = "Disconnect";

        #endregion Constants and Enums

        //**************************************************************
        #region Constructor/Destructor

        //-------------------------------------------
        /// <summary>
        /// Private Constructor - 
        /// </summary>
        private FliProCameraManager()
        {
            ActiveCamera = new FliProCamera();

            DisplayCamera = ActiveCamera;
            InitializeEmptyCamera();
            AvailableCameras = new ObservableCollection<FliProCameraListEntry>();
            BindingOperations.EnableCollectionSynchronization(AvailableCameras, _lock);

            string pathName = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
        }

        /// <summary>
        /// Initialize all properties
        /// </summary>
        public void Initialize()
        {
#if false
            CSFENG_CFGCAM cscfgCam;
            StringBuilder strFileName;
            StringBuilder strSaveFileName;
            cscfgCam = new CSFENG_CFGCAM();
            strFileName = new StringBuilder("C:\\TomsStuff\\Work\\AppliedLogix\\GoogleDrive\\Client Projects\\FLICamera\\FLICameraSettings_appsaved_atod.csv");
            NativeMethods.CSFENGCam_GetConfigFromFile(strFileName, ref cscfgCam);

            strSaveFileName = new StringBuilder("C:\\TomsStuff\\Work\\AppliedLogix\\GoogleDrive\\Client Projects\\FLICamera\\FLICameraSettings_appsaved_new_atod.csv");
            NativeMethods.CSFENGCam_SaveConfigToCSV(strSaveFileName, ref cscfgCam);

            CSFENG_CFGCAM cscfgCam;
            cscfgCam = new CSFENG_CFGCAM();
            cscfgCam.cfgCamInfo.uiDeviceType = 1234;
            cscfgCam.cfgCamInfo.uiSerialNumLength = 5;
            cscfgCam.cfgCamInfo.wcSerialNumber = "Hello";
            NativeMethods.CSFENGCam_SetConfig(0, ref cscfgCam);
#endif
            if (Process.GetProcessesByName(Process.GetCurrentProcess().ProcessName).Length > 1)
            {
                return;
            }

            LoadToolSettings(TOOL_SETTINGS_PREVIOUS_FILE, false, false);

            monitor.Connected += Monitor_Connected;
            monitor.Disconnected += Monitor_Disconnected;

            // Define some device ids we are interested in
            monitor.DeviceIDs.Add(new USBDeviceID(FLI_LEGACY_CAMERA_KEY, FLI_LEGACY_CAMERA_VID, FLI_LEGACY_CAMERA_PID));
            monitor.DeviceIDs.Add(new USBDeviceID(FLI_PRO_CAMERA_KEY, FLI_PRO_CAMERA_VID, FLI_PRO_CAMERA_KEPLER_PID));
            monitor.DeviceIDs.Add(new USBDeviceID(FLI_PRO_CAMERA_KEY, FLI_PRO_CAMERA_VID, FLI_PRO_CAMERA_COBALT_PID));
            monitor.DeviceIDs.Add(new USBDeviceID(FLI_PRO_CAMERA_KEY, FLI_PRO_CAMERA_CYPRESS_VID, FLI_PRO_CAMERA_CYPRESS_PID));
            monitor.DeviceIDs.Add(new USBDeviceID(FLI_LEGACY_FILTER_WHEEL_KEY, FLI_LEGACY_FILTER_WHEEL_VID, FLI_LEGACY_FILTER_WHEEL_PID));
            //monitor.DeviceIDs.Add(new USBDeviceID(SI_LABS_KEY, SI_LABS_VID));

            // Set up the bit mapping arrays for use during image capture
            for (int i = 0; i < 256; i++)
            {
                LeadByteArray[i] = (ushort)(i << 4);                 // Shift i to the left 4
                MiddleByteLeadArray[i] = (ushort)((i & 0xF0) >> 4);  // Mask out the lower nibble and shift right 4
                MiddleByteTailArray[i] = (ushort)((i & 0x0F) << 8);  // Mask out the upper nibble and shift right 8
            }

            // Initialize The Status Progress Timer
            StatusProgressTimer.Enabled = false;
            StatusProgressTimer.Elapsed += StatusProgressTimerEvent;
            StatusProgressTimer.Interval = 100;  // Post progress every 100 milliseconds
            StatusProgressTimer.AutoReset = true;

            monitor.Start();

            // Need to make this call after FliProCameraManager has been initialized
            PopulateCameraWithToolSettings(ActiveCamera);
        }

        /// <summary>
        /// This function is used to provide defaults so that the screen is populated even when there is no camera attached.
        /// </summary>
        private void InitializeEmptyCamera()
        {
            ActiveCamera.SetDefaultFliFrameCountList();
            ActiveCamera.SetDefaultFliFrameTypeList();
            ActiveCamera.SetDefaultSupportedModes();
            //ActiveCamera.SetDefaultSupportedBinRatios();
            ActiveCamera.SetDefaultSupportedPixelDepths();
            ActiveCamera.SetDefaultGainTables();
        }

        public void Close()
        {
            UpdateToolSettings();
            monitor.Stop();
        }

        #endregion Constructor/Destructor

        //**************************************************************
        #region Private Constants

        private const string FORMATTED_CONNECTED = " - Connected";
        private const string FORMATTED_NOT_CONNECTED = " - Not Connected";
        private const string FORMATTED_FILE_PATTERN = "File - ";
        private const string FORMATTED_GENERATED_PATTERN = "Generated - ";
        private const string FORMATTED_GEN1_PATTERN = "Legacy - ";
        private const string FORMATTED_GEN2_PATTERN = "FliPro - ";

        private const string FLI_LEGACY_CAMERA_VID = "0F18";
        private const string FLI_LEGACY_CAMERA_PID = "000A";
        private const string FLI_LEGACY_FILTER_WHEEL_VID = "0F18";
        private const string FLI_LEGACY_FILTER_WHEEL_PID = "0007";
        private const string FLI_PRO_CAMERA_CYPRESS_VID = "04B4";
        private const string FLI_PRO_CAMERA_CYPRESS_PID = "00F0";
        private const string FLI_PRO_CAMERA_VID = "0F18";
        private const string FLI_PRO_CAMERA_KEPLER_PID = "000E";
        private const string FLI_PRO_CAMERA_COBALT_PID = "000F";
        private const string SI_LABS_VID = "10c4";

        private const string FLI_LEGACY_FILTER_WHEEL_KEY = "FLI_FILTER_WHEEL";
        private const string FLI_LEGACY_CAMERA_KEY = "FLI_LEGACY";
        private const string FLI_PRO_CAMERA_KEY = "FLI_PRO";
        private const string SI_LABS_KEY = "SILABS";

        #endregion Private Constants

        //**************************************************************
        #region Public Classes

        public class ActiveCameraEventArgs : EventArgs
        {
            public ActiveCameraEventArgs() { }

            public ActiveCameraEventArgs(FliProCamera camera)
            {
                Camera = camera;
            }
            public FliProCamera Camera { get; set; }
        }

        public class ActiveJobCompleteEventArgs : EventArgs
        {
            public ActiveJobCompleteEventArgs() { }

            public ActiveJobCompleteEventArgs(bool success, FliFrame frame)
            {
                Success = success;
                Frame = frame;
            }
            public FliFrame Frame { get; set; }
            public bool Success { get; set; }
        }

        //-=-= TODO: Refactor: Merge these properties into FliProCamera class and delete
        /// <summary>
        /// Class to store the various fields need to maintain a list of cameras
        /// </summary>
        public class FliProCameraListEntry : ObservableProperties
        {
            public FliProCameraListEntry(FliProCamera camera)
            {
                Camera = camera;
                Type = camera.Type;
                DeviceInfo = (camera.DeviceInfo);
                insertDate = DateTime.Now;
            }

            public CameraType Type { get; protected set; }
            public bool IsLegacyCamera { get { return (Type == CameraType.GEN1); } }
            public bool IsGen2Camera { get { return (Type == CameraType.GEN2); } }
            public bool IsFromFile { get { return (Type == CameraType.FILE); } }
            public bool IsGeneratedBitmap { get { return (Type == CameraType.GENERATED); } }
            public DateTime insertDate { get; internal set; }

            // The information returned from the USBMonitor Event
            public USBDeviceInfo usbDeviceInfo { get; set; }

            // The information returned from the FliPro library enumerate call
            public FPRODEVICEINFO DeviceInfo { get; set; }

            public FliDeviceId LegacyDeviceId { get; set; }

            public FliProCamera Camera
            {
                get { return _Camera; }
                protected set
                {
                    SetValue(ref _Camera, value);
                }
            }
            private FliProCamera _Camera;

            public string Foreground
            {
                get
                {
                    return ((
                              ((Type == CameraType.GEN2) && ((DeviceInfo.USBSpeed != FProUSBSpeed.SUPERSPEED) || !Camera.IsConnected)) ||
                              ((Type == CameraType.GEN1) && !Camera.IsConnected)
                            ) ? "Red" : "Black");
                }
            }

            public void Refresh()
            {
                OnPropertyChanged("Text");
                OnPropertyChanged("Foreground");
            }

            public string Text
            {
                get
                {
                    string displayStr = String.Empty;
                    switch (Type)
                    {
                        case CameraType.GEN1:
                            displayStr = Camera.Model + Camera.GetDeviceType() + " - " +
                                         Camera.SerialNumber + (Camera.IsConnected ? FORMATTED_CONNECTED : FORMATTED_NOT_CONNECTED) + Camera.Speed;
                            break;
                        case CameraType.GEN2:
                            displayStr = Camera.Model + Camera.GetDeviceType() + " - " +
                                         Camera.SerialNumber + (Camera.IsConnected ? FORMATTED_CONNECTED : FORMATTED_NOT_CONNECTED) + Camera.Speed;
                            break;
                        default:
                            displayStr = "UNKNOWN";
                            break;
                    }
                    return displayStr;
                }
            }
        }

        public class NameOrder
        {
            public NameOrder() { }
            public NameOrder(NameEntryType nameEntry, bool isChecked)
            {
                NameEntry = nameEntry;
                IsChecked = isChecked;
            }

            public NameEntryType NameEntry { get; set; }
            public bool IsChecked { get; set; }
        }

        public enum NameEntryType
        {
            Binning,
            CameraModel,
            Date,
            DiagRoutine,
            ExposureTime,
            FrameNumber,
            FrameType,
            Gain,
            Mode,
            Mean,
            SensorTemperature,
            SerialNumber
        }

        public class StringListEntry : ObservableProperties
        {
            public StringListEntry(string text)
            {
                Text = text;
            }
            public string Text { get; set; }
        }

        #endregion Public Classes

        //**************************************************************
        #region Private Properties

        private static object _lock = new object();
        private static volatile FliProCameraManager _instance;
        private static object syncRoot = new Object();
        private bool LoggingActive = false;
        private USBMonitor monitor = new USBMonitor();
        protected DateTime ProgressTimeStartTime = DateTime.Now;
        protected int ProgressTimeMaximum = 0;
        protected int ProgressTimeValue = 0;
        protected bool RecursiveSet = false;
        protected System.Timers.Timer StatusProgressTimer = new System.Timers.Timer();
        protected bool UseStatusBar { get; set; }
        private NameOrder[] NameOrderArray = new NameOrder[Enum.GetNames(typeof(NameEntryType)).Length];
        private NameOrder[] SavedNameOrderArray = new NameOrder[Enum.GetNames(typeof(NameEntryType)).Length];

        #endregion Private Properties

        //**************************************************************
        #region Public Properties

        //-------------------------------------------
        /// <summary>
        /// Returns a reference to the single instance of this class.
        /// </summary>
        public static FliProCameraManager Instance
        {
            get
            {
                if (_instance == null)
                {
                    lock (syncRoot)
                    {
                        if (_instance == null)
                        {
                            _instance = new FliProCameraManager();
                        }
                    }
                }
                return _instance;
            }
        }

        // These arrays will be used during image capture to speed the conversion from 12 bits per pixel packed to 16 bits
        public ushort[] LeadByteArray = new ushort[256];
        public ushort[] MiddleByteLeadArray = new ushort[256];
        public ushort[] MiddleByteTailArray = new ushort[256];
        public FliProToolSettings ToolSettings = new FliProToolSettings();
        public ushort[] BiasFrameHigh = null;
        public ushort[] BiasFrameLow = null;
        public ushort[] FlatFrameHigh = null;
        public ushort[] FlatFrameLow = null;

        #endregion

        //**************************************************************
        #region Tool Tip Constants

        private const string TOOL_TIP_RAW_FILE_NAME = "The pattern of the file name to be used when saving captured frame files.";
        private const string TOOL_TIP_RAW_FILE_PATH = "The path in which to save captured frame files in the selected formats.";

        #endregion Tool Tip Constants

        //**************************************************************
        #region Observable Tool Tip Properties


        public string RawFileNameToolTip { get { return TOOL_TIP_RAW_FILE_NAME; } }
        public string RawFileNameTextBoxToolTip { get { return (RawFileName == String.Empty ? TOOL_TIP_RAW_FILE_NAME : RawFileName); } }
        public string RawFilePathToolTip { get { return TOOL_TIP_RAW_FILE_PATH; } }
        public string RawFilePathTextBoxToolTip { get { return (RawFilePath == String.Empty ? TOOL_TIP_RAW_FILE_PATH : RawFilePath); } }

        #endregion Observable Tool Tip Properties

        //**************************************************************
        #region Observable Properties

        //-------------------------------------------
        /// <summary>
        /// The current settings of the Camera.
        /// If an actual camera is attached this will remain pointing to that camera.
        /// </summary>
        public FliProCamera ActiveCamera
        {
            get { return _ActiveCamera; }
            set { SetValue(ref _ActiveCamera, value); }
        }
        private FliProCamera _ActiveCamera;

        public int ActiveIndex
        {
            get { return _ActiveIndex; }
            set
            {
                if (SetValue(ref _ActiveIndex, value) && (_ActiveIndex >= 0))
                {
                    DisplayCamera = AvailableCameras[_ActiveIndex].Camera;
                    if ((ActiveCamera.Type == CameraType.FILE) || (ActiveCamera.Type == CameraType.GENERATED) || (ActiveCamera.Type == CameraType.UNKNOWN) ||
                        ((DisplayCamera.Type != CameraType.FILE) && (DisplayCamera.Type != CameraType.GENERATED)))
                    {
                        // Only assign the ActiveCamera when it currently is a file or if the new camera is a real camera.
                        ActiveCamera = DisplayCamera;
                    }
                    
                    // Display the active camera and it's most recent frame
                    OnNewActiveCameraAction(new ActiveCameraEventArgs(DisplayCamera));
                }
            }
        }
        private int _ActiveIndex = -1;

        public int ActualStatusDisplayProgressValue
        {
            get { return _ActualStatusDisplayProgressValue; }
            set { SetValue(ref _ActualStatusDisplayProgressValue, value); }
        }
        private int _ActualStatusDisplayProgressValue = 0;

        public string ActualStatusFrameRate
        {
            get { return _ActualStatusFrameRate; }
            set { SetValue(ref _ActualStatusFrameRate, value); }
        }
        private string _ActualStatusFrameRate = String.Empty;

        public int ActualStatusProgressValue
        {
            get { return _ActualStatusProgressValue; }
            set { SetValue(ref _ActualStatusProgressValue, value); }
        }
        private int _ActualStatusProgressValue = 0;

                /// <summary>
                                        /// True means Subtract the selected Reference File from the display frames.
                                        /// </summary>
        public double AutoForceRangeTemperature
        {
            get { return _AutoForceRangeTemperature; }
            set { SetValue(ref _AutoForceRangeTemperature, value); }
        }
        private double _AutoForceRangeTemperature = 5;

        public double AutoForceTrainingLastTargetTemperature
        {
            get { return _AutoForceTrainingLastTargetTemperature; }
            set { SetValue(ref _AutoForceTrainingLastTargetTemperature, value); }
        }
        private double _AutoForceTrainingLastTargetTemperature = 0;

        public bool AutoForceTrainingTargetReached
        {
            get { return _AutoForceTrainingTargetReached; }
            set { SetValue(ref _AutoForceTrainingTargetReached, value); }
        }
        private bool _AutoForceTrainingTargetReached = false;

        public ObservableCollection<FliProCameraListEntry> AvailableCameras
        {
            get { return _AvailableCameras; }
            set { SetValue(ref _AvailableCameras, value); }
        }
        private ObservableCollection<FliProCameraListEntry> _AvailableCameras;

        public int AvailableCameraCount { get { return AvailableCameras.Count; } }

        public int ConnectedCameraCount
        {
            get
            {
                int count = 0;
                foreach (var cam in AvailableCameras)
                {
                    if (cam.Camera.IsConnected) count++;
                }
                return count;
            }
        }

        /// <summary>
        /// The camera currently display. It can be a File or Actual camera.
        /// </summary>
        public FliProCamera DisplayCamera
        {
            get { return _DisplayCamera; }
            set { SetValue(ref _DisplayCamera, value); }
        }
        private FliProCamera _DisplayCamera;

        public bool ExecuteSequenceActive
        {
            get { return _ExecuteSequenceActive; }
            set { SetValue(ref _ExecuteSequenceActive, value); }
        }
        private bool _ExecuteSequenceActive = false;

        //-------------------------------------------
        /// <summary>
        /// Property to display camera status in the bottom of the Main Window
        /// </summary>
        public string ExposureStatusText
        {
            get { return _ExposureStatusText; }
            set { SetValue<string>(ref _ExposureStatusText, value); }
        }
        protected string _ExposureStatusText;
                
        public bool FileWriteError
        {
            get { return _FileWriteError; }
            set { SetValue(ref _FileWriteError, value); }
        }
        private bool _FileWriteError = false;

        public int FramesCaptured
        {
            get { return _FramesCaptured; }
            set { SetValue<int>(ref _FramesCaptured, value); }
        }
        protected int _FramesCaptured = 0;

        public bool GrabPaused
        {
            get { return _GrabPaused; }
            set { SetValue<bool>(ref _GrabPaused, value); }
        }
        protected bool _GrabPaused = false;

        public bool IsCheckedAutoForceTraining
        {
            get { return _IsCheckedAutoForceTraining; }
            set
            {
                if (SetValue<bool>(ref _IsCheckedAutoForceTraining, value))
                {
                }
            }
        }
        protected bool _IsCheckedAutoForceTraining = true;

        /// <summary>
        /// Flag to determine if the 12-Bit Images should be left shifted to make the 16-Bit Image
        /// </summary>
        public bool LeftShift12BitImages
        {
            get { return _LeftShift12BitImages; }
            set { SetValue(ref _LeftShift12BitImages, value); }
        }
        private bool _LeftShift12BitImages = false;

        public FliFrame.FliImageType MainImageType { get; set; }

        public int MainWindowHeight
        {
            get { return _MainWindowHeight; }
            set { SetValue(ref _MainWindowHeight, value); }
        }
        private int _MainWindowHeight = 600;

        public int MainWindowLeft
        {
            get { return _MainWindowLeft; }
            set { SetValue(ref _MainWindowLeft, value); }
        }
        private int _MainWindowLeft = 0;

        public WindowState MainWindowState
        {
            get { return _MainWindowState; }
            set { SetValue(ref _MainWindowState, value); }
        }
        private WindowState _MainWindowState = WindowState.Normal;

        public int MainWindowTop
        {
            get { return _MainWindowTop; }
            set { SetValue(ref _MainWindowTop, value); }
        }
        private int _MainWindowTop = 0;

        public int MainWindowWidth
        {
            get { return _MainWindowWidth; }
            set { SetValue(ref _MainWindowWidth, value); }
        }
        private int _MainWindowWidth = 600;

        public string ProgressTimerClock
        {
            get { return _ProgressTimerClock; }
            set { SetValue(ref _ProgressTimerClock, value); }
        }
        private string _ProgressTimerClock = String.Empty;
        
        public string RawFilePath
        {
            get { return _RawFilePath; }
            set
            {
                if (SetValue<string>(ref _RawFilePath, value))
                {
                    ActiveCamera.SetSaveToFileActive();
                    OnPropertyChanged("RawFilePathTextBoxToolTip");
                }
            }
        }
        protected string _RawFilePath = String.Empty;

        public int StatusProgressMaximum
        {
            get { return _StatusProgressMaximum; }
            set { SetValue(ref _StatusProgressMaximum, value); }
        }
        private int _StatusProgressMaximum = 1;

        public int StatusProgressValue
        {
            get { return _StatusProgressValue; }
            set { SetValue(ref _StatusProgressValue, value); }
        }
        private int _StatusProgressValue = 0;

        public bool StatusProgressVisible
        {
            get { return _StatusProgressVisible; }
            set { SetValue(ref _StatusProgressVisible, value); }
        }
        private bool _StatusProgressVisible = false;

        public DateTime StatusStartTime
        {
            get { return _StatusStartTime; }
            set { SetValue(ref _StatusStartTime, value); }
        }
        private DateTime _StatusStartTime = DateTime.Now;

        #endregion Observable Properties

        //**************************************************************
        #region File Name Setup Observable Properties

        public int BinningOrder
        {
            get { return _BinningOrder; }
            set
            {
                if (SetValue(ref _BinningOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.Binning, value);
                }
            }
        }
        private int _BinningOrder = DEFAULT_ORDER_BINNING;

        public int CameraModelOrder
        {
            get { return _CameraModelOrder; }
            set
            {
                if (SetValue(ref _CameraModelOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.CameraModel, value);
                }
            }
        }
        private int _CameraModelOrder = DEFAULT_ORDER_CAMERA_MODEL;

        public int DateOrder
        {
            get { return _DateOrder; }
            set
            {
                if (SetValue(ref _DateOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.Date, value);
                }
            }
        }
        private int _DateOrder = DEFAULT_ORDER_DATE;

        public int DiagRoutineOrder
        {
            get { return _DiagRoutineOrder; }
            set
            {
                if (SetValue(ref _DiagRoutineOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.DiagRoutine, value);
                }
            }
        }
        private int _DiagRoutineOrder = DEFAULT_ORDER_DATE;

        public int ExposureTimeOrder
        {
            get { return _ExposureTimeOrder; }
            set
            {
                if (SetValue(ref _ExposureTimeOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.ExposureTime, value);
                }
            }
        }
        private int _ExposureTimeOrder = DEFAULT_ORDER_EXPOSURE_TIME;

        public int FrameNumberOrder
        {
            get { return _FrameNumberOrder; }
            set
            {
                if (SetValue(ref _FrameNumberOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.FrameNumber, value);
                }
            }
        }
        private int _FrameNumberOrder = DEFAULT_ORDER_FRAME_NUMBER;

        public int FrameTypeOrder
        {
            get { return _FrameTypeOrder; }
            set
            {
                if (SetValue(ref _FrameTypeOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.FrameType, value);
                }
            }
        }
        private int _FrameTypeOrder = DEFAULT_ORDER_FRAME_TYPE;

        public int GainOrder
        {
            get { return _GainOrder; }
            set
            {
                if (SetValue(ref _GainOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.Gain, value);
                }
            }
        }
        private int _GainOrder = DEFAULT_ORDER_GAIN;

        public bool IsCheckedBinning
        {
            get { return _IsCheckedBinning; }
            set
            {
                if (SetValue(ref _IsCheckedBinning, value))
                {
                    SetIsCheckedFlag(NameEntryType.Binning, value);
                }
            }
        }
        private bool _IsCheckedBinning = DEFAULT_IS_CHECKED_BINNING;

        public bool IsCheckedCameraModel
        {
            get { return _IsCheckedCameraModel; }
            set
            {
                if (SetValue(ref _IsCheckedCameraModel, value))
                {
                    SetIsCheckedFlag(NameEntryType.CameraModel, value);
                }
            }
        }
        private bool _IsCheckedCameraModel = DEFAULT_IS_CHECKED_CAMERA_MODEL;

        public bool IsCheckedDate
        {
            get { return _IsCheckedDate; }
            set
            {
                if (SetValue(ref _IsCheckedDate, value))
                {
                    SetIsCheckedFlag(NameEntryType.Date, value);
                }
            }
        }
        private bool _IsCheckedDate = DEFAULT_IS_CHECKED_DATE;

        public bool IsCheckedDiagRoutine
        {
            get { return _IsCheckedDiagRoutine; }
            set
            {
                if (SetValue(ref _IsCheckedDiagRoutine, value))
                {
                    SetIsCheckedFlag(NameEntryType.DiagRoutine, value);
                }
            }
        }
        private bool _IsCheckedDiagRoutine = DEFAULT_IS_CHECKED_DIAG_ROUTINE;

        public bool IsCheckedExposureTime
        {
            get { return _IsCheckedExposureTime; }
            set
            {
                if (SetValue(ref _IsCheckedExposureTime, value))
                {
                    SetIsCheckedFlag(NameEntryType.ExposureTime, value);
                }
            }
        }
        private bool _IsCheckedExposureTime = DEFAULT_IS_CHECKED_EXPOSURE_TIME;

        public bool IsCheckedFrameNumber
        {
            get { return _IsCheckedFrameNumber; }
            set
            {
                if (SetValue(ref _IsCheckedFrameNumber, value))
                {
                    SetIsCheckedFlag(NameEntryType.FrameNumber, value);
                }
            }
        }
        private bool _IsCheckedFrameNumber = DEFAULT_IS_CHECKED_FRAME_NUMBER;

        public bool IsCheckedFrameType
        {
            get { return _IsCheckedFrameType; }
            set
            {
                if (SetValue(ref _IsCheckedFrameType, value))
                {
                    SetIsCheckedFlag(NameEntryType.FrameType, value);
                }
            }
        }
        private bool _IsCheckedFrameType = DEFAULT_IS_CHECKED_FRAME_TYPE;

        public bool IsCheckedGain
        {
            get { return _IsCheckedGain; }
            set
            {
                if (SetValue(ref _IsCheckedGain, value))
                {
                    SetIsCheckedFlag(NameEntryType.Gain, value);
                }
            }
        }
        private bool _IsCheckedGain = DEFAULT_IS_CHECKED_GAIN;
        
        public bool IsCheckedMean
        {
            get { return _IsCheckedMean; }
            set
            {
                if (SetValue(ref _IsCheckedMean, value))
                {
                    SetIsCheckedFlag(NameEntryType.Mean, value);
                }
            }
        }
        private bool _IsCheckedMean = DEFAULT_IS_CHECKED_MEAN;

        public bool IsCheckedMode
        {
            get { return _IsCheckedMode; }
            set
            {
                if (SetValue(ref _IsCheckedMode, value))
                {
                    SetIsCheckedFlag(NameEntryType.Mode, value);
                }
            }
        }
        private bool _IsCheckedMode = DEFAULT_IS_CHECKED_MODE;

        public bool IsCheckedPreamble
        {
            get { return _IsCheckedPreamble; }
            set
            {
                if (SetValue(ref _IsCheckedPreamble, value))
                {
                    CalculateNameString(null);
                }
            }
        }
        private bool _IsCheckedPreamble = DEFAULT_IS_CHECKED_PREAMBLE;

        public bool IsCheckedSensorTemperature
        {
            get { return _IsCheckedSensorTemperature; }
            set
            {
                if (SetValue(ref _IsCheckedSensorTemperature, value))
                {
                    SetIsCheckedFlag(NameEntryType.SensorTemperature, value);
                }
            }
        }
        private bool _IsCheckedSensorTemperature = DEFAULT_IS_CHECKED_SENSOR_TEMPERATURE;

        public bool IsCheckedSerialNumber
        {
            get { return _IsCheckedSerialNumber; }
            set
            {
                if (SetValue(ref _IsCheckedSerialNumber, value))
                {
                    SetIsCheckedFlag(NameEntryType.SerialNumber, value);
                }
            }
        }
        private bool _IsCheckedSerialNumber = DEFAULT_IS_CHECKED_SERIAL_NUMBER;

        public int ModeOrder
        {
            get { return _ModeOrder; }
            set
            {
                if (SetValue(ref _ModeOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.Mode, value);
                }
            }
        }
        private int _ModeOrder = DEFAULT_ORDER_MODE;

        public string Preamble
        {
            get { return _Preamble; }
            set
            {
                if (SetValue(ref _Preamble, value))
                {
                    CalculateNameString(null);
                    OnPropertyChanged("PreambleTextBoxToolTip");
                }
            }
        }
        private string _Preamble = String.Empty;

        public string RawFileName
        {
            get { return _RawFileName; }
            set
            {
                if (SetValue(ref _RawFileName, value))
                {
                    OnPropertyChanged("RawFileNameTextBoxToolTip");
                }
            }
        }
        private string _RawFileName = String.Empty;

        public int SensorTemperatureOrder
        {
            get { return _SensorTemperatureOrder; }
            set
            {
                if (SetValue(ref _SensorTemperatureOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.SensorTemperature, value);
                }
            }
        }
        private int _SensorTemperatureOrder = DEFAULT_ORDER_SENSOR_TEMPERATURE;

        public int SerialNumberOrder
        {
            get { return _SerialNumberOrder; }
            set
            {
                if (SetValue(ref _SerialNumberOrder, value) && !RecursiveSet)
                {
                    SetOrder(NameEntryType.SerialNumber, value);
                }
            }
        }
        private int _SerialNumberOrder = DEFAULT_ORDER_SERIAL_NUMBER;

        #endregion File Name Setup Observable Properties

        //**************************************************************
        #region File Name Setup Public Methods

        public string CalculateNameString(FliFrame frame, string DiagSubName = "")
        {
            string NameString = String.Empty;
            if (IsCheckedPreamble)
            {
                NameString = Preamble;
            }

            for (int i = 0; i < NameOrderArray.Length; i++)
            {
                if ((NameOrderArray[i] != null) && NameOrderArray[i].IsChecked)
                {
                    string entryString = GetNameValue(NameOrderArray[i].NameEntry, frame, DiagSubName);
                    if (!String.IsNullOrEmpty(NameString) && !String.IsNullOrEmpty(entryString))
                    {
                        NameString += ",";
                    }
                    NameString += entryString;
                }
            }

            RawFileName = NameString;

            // Add an underscore to make the file number more visible 
            NameString += "_";

            return NameString;
        }

        public void InitializeOrderArray()
        {
            FliProCameraManager.NameOrder entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.Binning, IsCheckedBinning);

            NameOrderArray[DEFAULT_ORDER_BINNING - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.CameraModel, IsCheckedCameraModel);
            NameOrderArray[DEFAULT_ORDER_CAMERA_MODEL - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.Date, IsCheckedDate);
            NameOrderArray[DEFAULT_ORDER_DATE - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.DiagRoutine, IsCheckedDiagRoutine);
            NameOrderArray[DEFAULT_ORDER_DIAG_ROUTINE - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.ExposureTime, IsCheckedExposureTime);
            NameOrderArray[DEFAULT_ORDER_EXPOSURE_TIME - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.FrameNumber, IsCheckedFrameNumber);
            NameOrderArray[DEFAULT_ORDER_FRAME_NUMBER - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.Gain, IsCheckedGain);
            NameOrderArray[DEFAULT_ORDER_GAIN - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.FrameType, IsCheckedFrameType);
            NameOrderArray[DEFAULT_ORDER_FRAME_TYPE - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.Mode, IsCheckedMode);
            NameOrderArray[DEFAULT_ORDER_MODE - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.SensorTemperature, IsCheckedSensorTemperature);
            NameOrderArray[DEFAULT_ORDER_SENSOR_TEMPERATURE - 1] = entry;
            entry = new FliProCameraManager.NameOrder(FliProCameraManager.NameEntryType.SerialNumber, IsCheckedSerialNumber);
            NameOrderArray[DEFAULT_ORDER_SERIAL_NUMBER - 1] = entry;
        }

        public void PopulateFileNameArrayWithToolSettings()
        {
            if (NameOrderArray[0] == null)
            {
                InitializeOrderArray();
            }

            RecursiveSet = true;
            BinningOrder = (ToolSettings.BinningOrder == 0 ? DEFAULT_ORDER_BINNING : ToolSettings.BinningOrder);
            CameraModelOrder = (ToolSettings.CameraModelOrder == 0 ? DEFAULT_ORDER_CAMERA_MODEL : ToolSettings.CameraModelOrder);
            DateOrder = (ToolSettings.DateOrder == 0 ? DEFAULT_ORDER_DATE : ToolSettings.DateOrder);
            DiagRoutineOrder = (ToolSettings.DiagRoutineOrder == 0 ? DEFAULT_ORDER_DIAG_ROUTINE : ToolSettings.DiagRoutineOrder);
            ExposureTimeOrder = (ToolSettings.ExposureTimeOrder == 0 ? DEFAULT_ORDER_EXPOSURE_TIME : ToolSettings.ExposureTimeOrder);
            FrameNumberOrder = (ToolSettings.FrameNumberOrder == 0 ? DEFAULT_ORDER_FRAME_NUMBER : ToolSettings.FrameNumberOrder);
            GainOrder = (ToolSettings.GainOrder == 0 ? DEFAULT_ORDER_GAIN : ToolSettings.GainOrder);
            FrameTypeOrder = (ToolSettings.GrabTypeOrder == 0 ? DEFAULT_ORDER_FRAME_TYPE : ToolSettings.GrabTypeOrder);
            IsCheckedBinning = ToolSettings.IsCheckedBinning;
            IsCheckedCameraModel = ToolSettings.IsCheckedCameraModel;
            IsCheckedDate = ToolSettings.IsCheckedDate;
            IsCheckedExposureTime = ToolSettings.IsCheckedExposureTime;
            IsCheckedFrameNumber = ToolSettings.IsCheckedFrameNumber;
            IsCheckedGain = ToolSettings.IsCheckedGain;
            IsCheckedMean = ToolSettings.IsCheckedMean;
            IsCheckedMode = ToolSettings.IsCheckedMode;
            IsCheckedPreamble = ToolSettings.IsCheckedPreamble;
            IsCheckedSensorTemperature = ToolSettings.IsCheckedSensorTemperature;
            IsCheckedSerialNumber = ToolSettings.IsCheckedSerialNumber;
            ModeOrder = (ToolSettings.ModeOrder == 0 ? DEFAULT_ORDER_MODE : ToolSettings.ModeOrder);
            Preamble = ToolSettings.Preamble;
            RawFileName = ToolSettings.RawFileName;
            SensorTemperatureOrder = (ToolSettings.SensorTemperatureOrder == 0 ? DEFAULT_ORDER_SENSOR_TEMPERATURE : ToolSettings.SensorTemperatureOrder);
            SerialNumberOrder = (ToolSettings.SerialNumberOrder == 0 ? DEFAULT_ORDER_SERIAL_NUMBER : ToolSettings.SerialNumberOrder);
            RecursiveSet = false;

            // Set the Array to reflect the new property values
            SetOrderAbsolute(NameEntryType.Binning, BinningOrder);
            SetOrderAbsolute(NameEntryType.CameraModel, CameraModelOrder);
            SetOrderAbsolute(NameEntryType.Date, DateOrder);
            SetOrderAbsolute(NameEntryType.DiagRoutine, DiagRoutineOrder);
            SetOrderAbsolute(NameEntryType.ExposureTime, ExposureTimeOrder);
            SetOrderAbsolute(NameEntryType.FrameNumber, FrameNumberOrder);
            SetOrderAbsolute(NameEntryType.Gain, GainOrder);
            SetOrderAbsolute(NameEntryType.FrameType, FrameTypeOrder);
            SetOrderAbsolute(NameEntryType.Mode, ModeOrder);
            SetOrderAbsolute(NameEntryType.SensorTemperature, SensorTemperatureOrder);
            SetOrderAbsolute(NameEntryType.SerialNumber, SerialNumberOrder);
        }

        #endregion File Name Setup Public Methods

        //**************************************************************
        #region File Name Setup Private Methods

        public string GetNameValue(NameEntryType nameEntry, FliFrame frame, string DiagSubName = "", FliFrame.FliImageType imageType = FliFrame.FliImageType.LowGainDummy)
        {
            return String.Empty;
        }

        private void SetIsCheckedFlag(NameEntryType nameEntry, bool isChecked)
        {
            for (int i = 0; i < NameOrderArray.Length; i++)
            {
                if ((NameOrderArray[i] != null) && (NameOrderArray[i].NameEntry == nameEntry))
                {
                    NameOrderArray[i].IsChecked = isChecked;
                    break;
                }
            }

            if (!RecursiveSet)
            {
                CalculateNameString(null);
            }
        }

        private void SetOrder(NameEntryType nameEntry, int index)
        {
            int oldIndex = 0;
            NameOrder nameOrder = null;

            RecursiveSet = true;

            for (int i = 0; i < NameOrderArray.Length; i++)
            {
                if ((NameOrderArray[i] != null) && (NameOrderArray[i].NameEntry == nameEntry))
                {
                    oldIndex = i;
                    nameOrder = NameOrderArray[i];
                    break;
                }
            }

            if (index <= oldIndex)
            {
                // we are moving this entry up
                for (int i = oldIndex; i >= index; i--)
                {
                    NameOrderArray[i] = NameOrderArray[i - 1];
                    SetOrderOnDisplay(NameOrderArray[i].NameEntry, i + 1);
                }
                NameOrderArray[index - 1] = nameOrder;
                if (nameOrder.IsChecked)
                {
                    CalculateNameString(null);
                }
            }
            else if (index > oldIndex + 1)
            {
                // we are moving this entry down
                for (int i = oldIndex; i < index - 1; i++)
                {
                    NameOrderArray[i] = NameOrderArray[i + 1];
                    SetOrderOnDisplay(NameOrderArray[i].NameEntry, i + 1);
                }
                NameOrderArray[index - 1] = nameOrder;
                if (nameOrder.IsChecked)
                {
                    CalculateNameString(null);
                }
            }
            // else index == oldIndex + 1  // We are not moving this entry so do nothing

            RecursiveSet = false;
        }

        private void SetOrderAbsolute(NameEntryType nameEntry, int index)
        {
            int oldIndex = 0;
            NameOrder nameOrder = null;
            NameOrder tempEntry = null;

            RecursiveSet = true;

            for (int i = 0; i < NameOrderArray.Length; i++)
            {
                if ((NameOrderArray[i] != null) && (NameOrderArray[i].NameEntry == nameEntry))
                {
                    oldIndex = i;
                    nameOrder = NameOrderArray[i];
                    break;
                }
            }

            if (oldIndex != index - 1)
            {
                // Swap this entry with old location
                tempEntry = NameOrderArray[index - 1];
                NameOrderArray[index - 1] = nameOrder;
                NameOrderArray[oldIndex] = tempEntry;
            }

            RecursiveSet = false;
        }

        private void SetOrderOnDisplay(NameEntryType nameEntry, int index)
        {
            switch (nameEntry)
            {
                case NameEntryType.Binning:
                    BinningOrder = index;
                    break;
                case NameEntryType.CameraModel:
                    CameraModelOrder = index;
                    break;
                case NameEntryType.Date:
                    DateOrder = index;
                    break;
                case NameEntryType.DiagRoutine:
                    DiagRoutineOrder = index;
                    break;
                case NameEntryType.ExposureTime:
                    ExposureTimeOrder = index;
                    break;
                case NameEntryType.FrameNumber:
                    FrameNumberOrder = index;
                    break;
                case NameEntryType.Gain:
                    GainOrder = index;
                    break;
                case NameEntryType.FrameType:
                    FrameTypeOrder = index;
                    break;
                case NameEntryType.Mode:
                    ModeOrder = index;
                    break;
                case NameEntryType.SensorTemperature:
                    SensorTemperatureOrder = index;
                    break;
                case NameEntryType.SerialNumber:
                    SerialNumberOrder = index;
                    break;
            }
        }

        #endregion File Name Setup Private Methods

        //**************************************************************
        #region Public Methods

        public void ActivateLogging(bool isActive)
        {
            LoggingActive = isActive;
        }

        public bool DeleteSequenceFile(StringListEntry item)
        {
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_SEQUENCES_FOLDER);
            bool success = false;

            if (File.Exists(pathName + "\\" + item.Text + ".xml"))
            {
                try
                {
                    File.Delete(pathName + "\\" + item.Text + ".xml");
                    success = true;
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Error encountered Deleting file: " + pathName + "\\" + item.Text + ".xml" +
                        "\n\nError:\n" + ex.Message, "Warning", MessageBoxButton.OK,
                        MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                }
            }
            return success;
        }

        public bool DeleteSettingsFile(StringListEntry item)
        {
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_TOOL_SETTINGS_FOLDER);
            bool success = false;

            if (File.Exists(pathName + "\\" + item.Text + ".xml"))
            {
                try
                {
                    File.Delete(pathName + "\\" + item.Text + ".xml");
                    success = true;
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Error encountered Deleting file: " + pathName + "\\" + item.Text + ".xml" +
                        "\n\nError:\n" + ex.Message, "Warning", MessageBoxButton.OK,
                        MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                }
            }
            return success;
        }

        //-------------------------------------------
        /// <summary>
        /// Get a list of all connected cameras
        /// </summary>
        /// <returns>a list of Device Info structs representing all connected cameras</returns>
        public List<FPRODEVICEINFO> GetConnectedCameras()
        {
            FPRODEVICEINFO[] DevInfo = new FPRODEVICEINFO[FPRO_MAX_CONNECTED_CAMERAS];
            uint nCameras = FPRO_MAX_CONNECTED_CAMERAS;
            try
            {
                int value = NativeMethods.FPROCam_GetCameraList(DevInfo, ref nCameras);
            }
            catch (Exception ex)
            {
                // just return the empty list on any error
                nCameras = 0;
            }

            var devinfolist = new List<FPRODEVICEINFO>();   // start with an empty list
            for (int i = 0; i < nCameras; i++)
            {
                devinfolist.Add(DevInfo[i]);
            }

            return devinfolist;
        }

        public string GetFileName()
        {
            string fileName = String.Empty;
            return fileName;
        }

        //-------------------------------------------
        /// <summary>
        /// Return the nth connected camera of the specified type
        /// </summary>
        /// <param name="type">camera type</param>
        /// <param name="isconnected">true if connected, false if disconnected</param>
        /// <param name="n">desired instance in available cameras list</param>
        /// <returns>reference to specified camera or null if not found</returns>
        public FliProCamera FindCamera(CameraType type, bool isconnected = true, int n = 1)
        {
            FliProCamera camera = null;
            int i = 0;
            foreach (var cam in AvailableCameras)
            {
                if ((cam.Camera.IsConnected == isconnected) && (cam.Type == type))
                {
                    if (++i == n)
                    {
                        camera = cam.Camera;
                    }
                }
            }
            return camera;
        }

        public string GetFliPilotVersion()
        {
            return Assembly.GetExecutingAssembly().GetName().Version.ToString();
        }

        //-------------------------------------------
        /// <summary>
        /// Intermediate struct to make it easier to marshal the char*** returned by FliList.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        private class StringBox
        {
            public string str;
        }

        public string GetAPIVersion()
        {
            StringBuilder version = new StringBuilder(25);

            NativeMethods.FPROCam_GetAPIVersion(version, 25);

            return version.ToString();
        }

        //-------------------------------------------
        /// <summary>
        /// Return a list of device ids for all available devices
        /// matching the specified domain.
        /// (FliList, FliFreeList)
        /// </summary>
        /// <param name="domain">desired device domain (USB_CAMERA by default)</param>
        /// <returns>list of DeviceInfo entries (possibly empty)</returns>
        public List<FliDeviceId> GetDeviceList(DeviceDomain domain = DeviceDomain.USB_CAMERA)
        {
            var devinfolist = new List<FliDeviceId>();   // start with an empty list
            IntPtr listPtr = IntPtr.Zero;

            try
            {
                NativeMethods.FLIList((int)domain, out listPtr);
            }
            catch (Exception)
            {
                // just return empty list on any error
                listPtr = IntPtr.Zero;
            }

            if (listPtr != IntPtr.Zero)
            {
                // Iterate through and parse all device info strings
                IntPtr devPtr = listPtr;
                bool done = false;
                while (!done && (devinfolist.Count < MAX_DEVICE_COUNT))
                {
                    StringBox strbox = (StringBox)Marshal.PtrToStructure(devPtr, typeof(StringBox));
                    string str = strbox.str;
                    if (str != null)
                    {
                        var devinfo = new FliDeviceId();
                        int sep = strbox.str.IndexOf(";");
                        if (sep >= 0)
                        {
                            devinfo.Name = strbox.str.Substring(0, sep);
                            devinfo.Model = strbox.str.Substring(sep + 1);
                        }
                        else
                        {
                            // No model description
                            devinfo.Name = strbox.str;
                            devinfo.Model = "";
                        }
                        devinfo.Domain = domain;
                        devinfolist.Add(devinfo);
                        devPtr += Marshal.SizeOf(devPtr);
                    }
                    else
                    {
                        done = true;
                    }
                }
                NativeMethods.FLIFreeList(listPtr);
            }

            return devinfolist;
        }

        public void GetDeviceVersion(ref string firmwareVersion, ref string fpgaVersion, ref string controllerVersion)
        {
            ActiveCamera.GetDeviceVersion(ref firmwareVersion, ref fpgaVersion, ref controllerVersion);
        }

        //-=-= FIXME: This will break if a file name matches a camera serial number or a test pattern name, etc.
        public FliProCamera GetFliProCamera(string CameraID)
        {
            FliProCamera camera = null;

            foreach (FliProCameraListEntry cameraListClass in AvailableCameras)
            {
                if (CameraID.Equals(cameraListClass.Camera.SerialNumber))
                {
                    camera = cameraListClass.Camera;
                    break;
                }
            }
            return camera;
        }

        public string GetFliProCameraModel(int DeviceType)
        {
            string deviceTypeStr = "Unknown Model";

            switch ((FPRODEVICETYPE)DeviceType)
            {
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400:
                    deviceTypeStr = CAMERA_MODEL_KEPLER_0400;
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE4040:
                    deviceTypeStr = CAMERA_MODEL_KEPLER_4040;
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE2020:
                    deviceTypeStr = CAMERA_MODEL_KEPLER_2020;
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_KODAK29050:
                    deviceTypeStr = CAMERA_MODEL_KODAK_29050;
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_KODAK47051:
                    deviceTypeStr = CAMERA_MODEL_KODAK_47051;
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42:
                    deviceTypeStr = CAMERA_MODEL_COBALT_DC230_42;
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84:
                    deviceTypeStr = CAMERA_MODEL_COBALT_DC230_84;
                    break;
                default:
                    deviceTypeStr = "Unknown Model";
                    break;
            }
            return deviceTypeStr;
        }

        public ObservableCollection<StringListEntry> GetToolSequenceNameList()
        {
            ObservableCollection<StringListEntry> nameList = new ObservableCollection<StringListEntry>();
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_SEQUENCES_FOLDER);

            if (!Directory.Exists(pathName))
            {
                // This is the first time this app has been run for this user so create their sequence files
                Directory.CreateDirectory(pathName);

                // Change the security level so that everyone has access to this directory
                DirectoryInfo dInfo = new DirectoryInfo(pathName);
                DirectorySecurity dSecurity = dInfo.GetAccessControl();
                dSecurity.AddAccessRule(new FileSystemAccessRule(new SecurityIdentifier(WellKnownSidType.WorldSid, null),
                    FileSystemRights.FullControl, InheritanceFlags.ObjectInherit | InheritanceFlags.ContainerInherit, PropagationFlags.NoPropagateInherit, AccessControlType.Allow));
                dInfo.SetAccessControl(dSecurity);
            }

            DirectoryInfo dirInfo = new DirectoryInfo(pathName);

            FileInfo[] info = dirInfo.GetFiles("*.xml");
            foreach (FileInfo f in info)
            {
                nameList.Add(new StringListEntry(Path.GetFileNameWithoutExtension(f.Name)));
            }
            return nameList;
        }

        public ObservableCollection<StringListEntry> GetToolSettingsNameList()
        {
            ObservableCollection<StringListEntry> nameList = new ObservableCollection<StringListEntry>();

            try
            {
                string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_TOOL_SETTINGS_FOLDER);
                if (!Directory.Exists(pathName))
                {
                    Directory.CreateDirectory(pathName);
                }
                DirectoryInfo dirInfo = new DirectoryInfo(pathName);

                FileInfo[] info = dirInfo.GetFiles("*.xml");
                foreach (FileInfo f in info)
                {
                    if (!String.Equals(f.Name, TOOL_SETTINGS_DEFAULT_FILE) && !String.Equals(f.Name, TOOL_SETTINGS_PREVIOUS_FILE))
                    {
                        nameList.Add(new StringListEntry(Path.GetFileNameWithoutExtension(f.Name)));
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error encountered opening Settings folder:\n" + ex.Message);
            }
            return nameList;
        }

        public ObservableCollection<string> LoadToolSequence(string fileName)
        {
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_SEQUENCES_FOLDER);
            ObservableCollection<string> sequenceList = new ObservableCollection<string>();

            try
            {
                TextWriter WriteFileStream;

                // Create a new XmlSerializer instance with the type of the test class
                XmlSerializer SerializerObj = new XmlSerializer(typeof(ObservableCollection<string>));

                if (!Directory.Exists(pathName))
                {
                    // This is the first time this app has been run for this user so create their settings files
                    Directory.CreateDirectory(pathName);

                    // Change the security level so that everyone has access to this directory
                    DirectoryInfo dInfo = new DirectoryInfo(pathName);
                    DirectorySecurity dSecurity = dInfo.GetAccessControl();
                    dSecurity.AddAccessRule(new FileSystemAccessRule(new SecurityIdentifier(WellKnownSidType.WorldSid, null),
                        FileSystemRights.FullControl, InheritanceFlags.ObjectInherit | InheritanceFlags.ContainerInherit, PropagationFlags.NoPropagateInherit, AccessControlType.Allow));
                    dInfo.SetAccessControl(dSecurity);
                }

                // Now see if the file being asked for exists. This will only be the case for a user file being asked for.
                if (!File.Exists(pathName + @"\" + fileName))
                {
                    if (MessageBox.Show("The settings file: " + fileName + " does not exist. Would you like to create it?", "File Does Not Exist",
                                        MessageBoxButton.YesNo, MessageBoxImage.Warning, MessageBoxResult.OK,
                                        MessageBoxOptions.DefaultDesktopOnly) == MessageBoxResult.OK)
                    {
                        // Create the Previous Settings file stream to write the serialized object to a file
                        WriteFileStream = new StreamWriter(pathName + @"\" + fileName);
                        SerializerObj.Serialize(WriteFileStream, ToolSettings);
                        WriteFileStream.Close();
                    }
                    else
                    {
                        return sequenceList;
                    }
                }

                // Retreive the previous settings file and load it into toolSettings
                FileStream ReadFileStream = new FileStream(pathName + @"\" + fileName, FileMode.Open, FileAccess.Read, FileShare.Read);

                // Load the object saved above by using the Deserialize function
                sequenceList = (ObservableCollection<string>)SerializerObj.Deserialize(ReadFileStream);

                // Cleanup
                ReadFileStream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("The following error was encountered when opening the settings file:\n" + pathName + @"\" + fileName +
                    "\n\n" + ex.Message, "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }

            return sequenceList;
        }

        public int GetDeviceTypeFromSettings(string fileName)
        {
            int deviceType = 0;
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_TOOL_SETTINGS_FOLDER);
            // Create a new XmlSerializer instance with the type of the test class
            XmlSerializer SerializerObj = new XmlSerializer(typeof(FliProToolSettings));

            try
            {
                if (!File.Exists(pathName + @"\" + fileName))
                {
                    MessageBox.Show("The settings file: " + fileName + " does not exist.");
                    return deviceType;
                }

                // Retreive the previous settings file and load it into toolSettings
                FileStream ReadFileStream = new FileStream(pathName + @"\" + fileName, FileMode.Open, FileAccess.Read, FileShare.Read);

                // Load the object saved above by using the Deserialize function
                FliProToolSettings TempToolSettings = (FliProToolSettings)SerializerObj.Deserialize(ReadFileStream);

                deviceType = TempToolSettings.DeviceType;

                // Cleanup
                ReadFileStream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("The following error was encountered when opening the settings file:\n" + pathName + @"\" + fileName +
                    "\n\n" + ex.Message, "Warning", MessageBoxButton.OK, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
            return deviceType;
        }

        public bool LoadToolSettings(string fileName, bool isSequence, bool raiseMessage = true)
        {
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData),
                (isSequence ? TOOL_TOOL_SETTINGS_FOLDER : TOOL_SETTINGS_FOLDER));
            string defaultPathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_SETTINGS_FOLDER);

            try
            {
                TextWriter WriteFileStream;

                // Create a new XmlSerializer instance with the type of the test class
                XmlSerializer SerializerObj = new XmlSerializer(typeof(FliProToolSettings));

                if (!Directory.Exists(pathName))
                {
                    // This is the first time this app has been run for this user so create their settings files
                    Directory.CreateDirectory(pathName);

                    // Change the security level so that everyone has access to this directory
                    DirectoryInfo dInfo = new DirectoryInfo(pathName);
                    DirectorySecurity dSecurity = dInfo.GetAccessControl();
                    dSecurity.AddAccessRule(new FileSystemAccessRule(new SecurityIdentifier(WellKnownSidType.WorldSid, null),
                        FileSystemRights.FullControl, InheritanceFlags.ObjectInherit | InheritanceFlags.ContainerInherit, PropagationFlags.NoPropagateInherit, AccessControlType.Allow));
                    dInfo.SetAccessControl(dSecurity);
                }

                // See if the Default Settings file exists
                if (!File.Exists(defaultPathName + @"\" + TOOL_SETTINGS_DEFAULT_FILE))
                {
                    // Create the Default Settings file stream to write the serialized object to a file
                    WriteFileStream = new StreamWriter(defaultPathName + @"\" + TOOL_SETTINGS_DEFAULT_FILE);
                    SerializerObj.Serialize(WriteFileStream, new FliProToolSettings());
                    WriteFileStream.Close();
                }

                // See if the Previous Settings file exists
                if (!File.Exists(defaultPathName + @"\" + TOOL_SETTINGS_PREVIOUS_FILE))
                {
                    // Create the Previous Settings file stream to write the serialized object to a file
                    WriteFileStream = new StreamWriter(defaultPathName + @"\" + TOOL_SETTINGS_PREVIOUS_FILE);
                    SerializerObj.Serialize(WriteFileStream, ToolSettings);
                    WriteFileStream.Close();
                }

                // Now see if the file being asked for exists. This will only be the case for a user file being asked for.
                if (!File.Exists(pathName + @"\" + fileName))
                {
                    if (raiseMessage)
                    {
                        if (MessageBox.Show("The settings file: " + fileName + " does not exist. Would you like to create it?", "File Does Not Exist",
                            MessageBoxButton.YesNo, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly) == MessageBoxResult.OK)
                        {
                            // Create the Previous Settings file stream to write the serialized object to a file
                            WriteFileStream = new StreamWriter(pathName + @"\" + fileName);
                            SerializerObj.Serialize(WriteFileStream, ToolSettings);
                            WriteFileStream.Close();
                        }
                        else
                        {
                            return false;
                        }
                    }
                    else
                    {
                        // Just create the Previous Settings file stream to write the serialized object to a file
                        WriteFileStream = new StreamWriter(pathName + @"\" + fileName);
                        SerializerObj.Serialize(WriteFileStream, ToolSettings);
                        WriteFileStream.Close();
                    }
                }

                // Retreive the previous settings file and load it into toolSettings
                FileStream ReadFileStream = new FileStream(pathName + @"\" + fileName, FileMode.Open, FileAccess.Read, FileShare.Read);

                // Load the object saved above by using the Deserialize function
                ToolSettings = (FliProToolSettings)SerializerObj.Deserialize(ReadFileStream);

                // Cleanup
                ReadFileStream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("The following error was encountered when opening the settings file:\n" + pathName + @"\" + fileName +
                    "\n\n" + ex.Message, "Warning", MessageBoxButton.OK, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                return false;
            }
            return true;
        }

        public void PopulateCameraWithToolSettings(FliProCamera camera)
        {
            if (camera.IsInitializedFromSettings)
            {
                return;
            }

            camera.IsInitializedFromSettings = true;

            camera.SetForcePropertyChange(true);

            if (!ToolSettingsFileExists(camera.SerialNumber + ".xml"))
            {
                // We need to get the factory settings that are on the camera and use them here
                double SetPoint = 25;
                camera.GetTemperatureSetPoint(ref SetPoint);
                camera.TargetTemperature = SetPoint;

                // Save all the currently loaded values in the camera to ToolSettings
                UpdateToolSettings();
            }

            try
            {
                // Only repopulate the fields if this is the same camera type that was saved
                // Otherwise we'll just use the camera settings
                if (ToolSettings.DeviceType == ActiveCamera.DeviceType)
                {

                    // Setting the Mode reads many other fields from the camera so we want to set it first
                    // then overwrite the read settings with those saved here
                    if (camera.Mode != ToolSettings.Mode)
                    {
                        camera.Mode = ToolSettings.Mode;
                    }

                    // Camera Control Panel Properties
                    //camera.CoolerDutyCycleIsEnabledOn = ToolSettings.CoolerDutyCycleIsEnabledOn;    // only set from the camera or engineering mode
                    //camera.CoolerDutyCyclePower = ToolSettings.CoolerDutyCyclePower;          // only set from the camera or engineering mode
                    camera.CoolerTemperatureOn = ToolSettings.CoolerTemperatureOn;
                    camera.FanIsEnabledOn = ToolSettings.FanIsEnabledOn;
                    //camera.FanPower = ToolSettings.FanPower;          // Fan power needs to only be set from the camera or engineering mode
                    camera.HeaterEnabledOn = ToolSettings.HeaterEnabledOn;
                    camera.HeaterPower = ToolSettings.HeaterPower;
                    camera.IlluminationOffDelay = ToolSettings.IlluminationOffDelay;
                    camera.IlluminationOn = ToolSettings.IlluminationOn;
                    camera.IlluminationOnDelay = ToolSettings.IlluminationOnDelay;
                    camera.LEDOn = ToolSettings.LEDOn;
                    camera.MaximumFrameStorage = ToolSettings.MaximumFrameStorage;
                    camera.ReadDuringExposureOn = ToolSettings.ReadDuringExposureOn;
                    camera.ShutterClose = ToolSettings.ShutterClose;
                    camera.ShutterOpen = ToolSettings.ShutterOpen;

                    // This needs to go below ShutterOpen so that ShutterOpen doesn't override the value
                    camera.ManualShutterControlEnabled = ToolSettings.ManualShutterControlEnabled;

                    camera.TemperatureReadInterval = ToolSettings.TemperatureReadInterval;
                    camera.TargetTemperature = ToolSettings.TargetTemperature;

                    // Exposure Panel Properties
                    camera.ExposureTime = ToolSettings.ExposureTime;
                    camera.ExposureUnits = (FliProCamera.ExposureUnits_Enum)ToolSettings.ExposureUnits;
                    camera.ExposureDelayTime = ToolSettings.ExposureDelayTime;
                    camera.ExposureDelayUnits = (FliProCamera.ExposureUnits_Enum)ToolSettings.ExposureDelayUnits;
                    camera.ExposureFrameIntervalTime = ToolSettings.ExposureFrameIntervalTime;
                    camera.ExposureFrameIntervalUnits = (FliProCamera.ExposureUnits_Enum)ToolSettings.ExposureFrameIntervalUnits;
                    camera.Immediate = ToolSettings.Immediate;
                    camera.UseExposureDelay = ToolSettings.UseExposureDelay;

                    // File Name Setup Properties
                    PopulateFileNameArrayWithToolSettings();

                    // File Panel Properties
                    RawFilePath = ToolSettings.RawFilePath;

                    // Force Training Panel Properties
                    IsCheckedAutoForceTraining = ToolSettings.IsCheckedAutoForceTraining;
                    AutoForceRangeTemperature = ToolSettings.AutoForceRangeTemperature;

                    // Frame Panel Properties
                    camera.ImageDataEnabled = ToolSettings.ImageDataEnabled;
                    camera.ImageUpperLeftX = ToolSettings.ImageUpperLeftX;
                    camera.ImageUpperLeftY = ToolSettings.ImageUpperLeftY;
                    camera.ImageWidth = ToolSettings.ImageWidth;
                    camera.ImageHeight = ToolSettings.ImageHeight;

                    if ((ToolSettings.ImageWidth == 1) && (ToolSettings.ImageHeight == 1))
                    {
                        camera.ResetImageArea();
                    }

                    // Grab Panel Properties
                    camera.FrameCountType = FliPro.FliFrameCount.SINGLE;
                    camera.FrameType = (FliFrameType)ToolSettings.FrameType;

                    // Merge Panel Properties
                    camera.ADUsPerElectron = ToolSettings.ADUsPerElectron;
                    camera.GainRatio = ToolSettings.GainRatio;
                    camera.HighLowThreshold = ToolSettings.HighLowThreshold;
                    camera.LineOffset = ToolSettings.LineOffset;
                    camera.LineOffset = ToolSettings.LineOffset;
                    camera.LineOffset = ToolSettings.LineOffset;

                    // Sensor Panel Properties

                    // Decided to put these back in so sequencing works 1/11/2019
                    // Bob asked not to set the camera to the saved settings 7/24/2018
                    camera.BlackLevelHigh = ToolSettings.BlackLevelHigh;
                    camera.BlackLevelLow = ToolSettings.BlackLevelLow;
                    camera.BlackSunHigh = ToolSettings.BlackSunHigh;
                    camera.BlackSunLow = ToolSettings.BlackSunLow;

                    camera.HBin = ToolSettings.HBin;
                    SensorGain GainValue = camera.GetGainValueFromIndex(camera.HighGainTable, ToolSettings.HighGainIndex);
                    if ((GainValue.Index == 0) && (GainValue.Gain == 1.0) && (camera.HighGainTable != null))
                    {
                        ToolSettings.HighGainIndex = camera.HighGainTable[0].Index;
                    }
                    camera.HighGainIndex = ToolSettings.HighGainIndex;
                    camera.HorizontalScanDirInverted = ToolSettings.HorizontalScanDirInverted;

                    GainValue = camera.GetGainValueFromIndex(camera.LowGainTable, ToolSettings.LowGainIndex);
                    if ((GainValue.Index == 0) && (GainValue.Gain == 1.0) && (camera.LowGainTable != null))
                    {
                        ToolSettings.LowGainIndex = camera.LowGainTable[0].Index;
                    }
                    camera.LowGainIndex = ToolSettings.LowGainIndex;
                    camera.PixelDepth = ToolSettings.PixelDepth;
                    camera.UseSoftwareBinning = ToolSettings.UseSoftwareBinning;
                    camera.VBin = ToolSettings.VBin;
                    camera.VerticalScanDirInverted = ToolSettings.VerticalScanDirInverted;
                    camera.BinningSquared = ToolSettings.BinningSquared;  // Set this after everything else has been set so that it is correct

                    OnNewActiveCameraAction(new ActiveCameraEventArgs(camera));

                }

                // Window Location Properties
                MainWindowHeight = ToolSettings.MainWindowHeight;
                MainWindowLeft = ToolSettings.MainWindowLeft;
                MainWindowState = (WindowState)ToolSettings.MainWindowState;
                MainWindowTop = ToolSettings.MainWindowTop;
                MainWindowWidth = ToolSettings.MainWindowWidth;
            }
            catch { }   // Just swallow any errors

            OnSettingsAction(new EventArgs());

            camera.SetForcePropertyChange(false);
        }

        public void PublishStatus(string text, bool isAppend)
        {
            if (ActiveCamera != null)
            {
                ActiveCamera.PublishStatus(text, isAppend);
            }
        }

        public void RemoveCamera(FliProCameraListEntry entry)
        {
            if ((entry.IsGen2Camera || entry.IsLegacyCamera) && entry.Camera.IsConnected)
            {
                MessageBox.Show("Attached Cameras cannot be deleted from the list.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                return;
            }

            if (AvailableCameraCount <= 1)
            {
                MessageBox.Show("Deleting the last Camera is not permitted.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                return;
            }

            for (int i = 0; i < AvailableCameraCount; i++)
            {
                if (entry == AvailableCameras[i])
                {
                    int savedActiveIndex = ActiveIndex;
                    AvailableCameras.RemoveAt(i);
                    if (savedActiveIndex == i)
                    {
                        ActiveIndex = 0;
                    }
                    else if (savedActiveIndex > i)
                    {
                        ActiveIndex = i - 1;
                    }
                }
            }
        }

        public void RescanForCameras()
        {
            // First disconnect from all the cameras
            foreach(FliProCameraListEntry cameraEntry in AvailableCameras)
            {
                cameraEntry.Camera.Disconnect();
            }
            monitor.Rescan();
        }

        public void SaveToolSequence(string fileName, ObservableCollection<string> sequenceList)
        {
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_SEQUENCES_FOLDER);

            // Create a new XmlSerializer instance with the type of the test class
            XmlSerializer SerializerObj = new XmlSerializer(typeof(ObservableCollection<string>));

            try
            {
                if (!Directory.Exists(pathName))
                {
                    // This is the first time this app has been run for this user so create their sequence files
                    Directory.CreateDirectory(pathName);

                    // Change the security level so that everyone has access to this directory
                    DirectoryInfo dInfo = new DirectoryInfo(pathName);
                    DirectorySecurity dSecurity = dInfo.GetAccessControl();
                    dSecurity.AddAccessRule(new FileSystemAccessRule(new SecurityIdentifier(WellKnownSidType.WorldSid, null),
                        FileSystemRights.FullControl, InheritanceFlags.ObjectInherit | InheritanceFlags.ContainerInherit,
                        PropagationFlags.NoPropagateInherit, AccessControlType.Allow));
                    dInfo.SetAccessControl(dSecurity);
                }

                // Delete the file if it already exists
                if (File.Exists(pathName + @"\" + fileName))
                {
                    if (MessageBox.Show(Path.GetFileNameWithoutExtension(fileName) + " already exists. Do you want to overwrite it?", "Name Already Exists",
                        MessageBoxButton.YesNo, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly) == MessageBoxResult.No)
                    {
                        return;
                    }
                    File.Delete(pathName + @"\" + fileName);
                }

                // Create the new Settings file stream to write the serialized object to a file
                TextWriter WriteFileStream = new StreamWriter(pathName + @"\" + fileName);
                SerializerObj.Serialize(WriteFileStream, sequenceList);
                WriteFileStream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error encountered saving file:\n" + pathName + @"\" + fileName + "\n\nError: " + ex.Message, "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
        }

        public void SaveToolSequenceSettings(string fileName, bool raiseMessage = true)
        {
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_TOOL_SETTINGS_FOLDER);

            // Create a new XmlSerializer instance with the type of the test class
            XmlSerializer SerializerObj = new XmlSerializer(typeof(FliProToolSettings));

            try
            {
                // Delete the file if it already exists
                if (File.Exists(pathName + @"\" + fileName))
                {
                    if (!String.Equals(fileName, TOOL_SETTINGS_DEFAULT_FILE) && !String.Equals(fileName, TOOL_SETTINGS_PREVIOUS_FILE))
                    {
                        if (raiseMessage)
                        {
                            if (MessageBox.Show(Path.GetFileNameWithoutExtension(fileName) + " already exists. Do you want to overwrite it?", "Name Already Exists",
                                MessageBoxButton.YesNo, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly) == MessageBoxResult.No)
                            {
                                return;
                            }
                        }
                    }
                    File.Delete(pathName + @"\" + fileName);
                }

                // Create the new Settings file stream to write the serialized object to a file
                TextWriter WriteFileStream = new StreamWriter(pathName + @"\" + fileName);
                SerializerObj.Serialize(WriteFileStream, ToolSettings);
                WriteFileStream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error encountered saving file:\n" + pathName + @"\" + fileName + "\n\nError: " + ex.Message, "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
        }

        public void SaveToolSettings(string fileName, bool isSequence, bool raiseMessage = true)
        {
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData),
                (isSequence ? TOOL_TOOL_SETTINGS_FOLDER : TOOL_SETTINGS_FOLDER));

            // Create a new XmlSerializer instance with the type of the test class
            XmlSerializer SerializerObj = new XmlSerializer(typeof(FliProToolSettings));

            try
            {
                // Delete the file if it already exists
                if (File.Exists(pathName + @"\" + fileName))
                {
                    if (!String.Equals(fileName, TOOL_SETTINGS_DEFAULT_FILE) && !String.Equals(fileName, TOOL_SETTINGS_PREVIOUS_FILE))
                    {
                        if (raiseMessage)
                        {
                            if (MessageBox.Show(Path.GetFileNameWithoutExtension(fileName) + " already exists. Do you want to overwrite it?", "Name Already Exists",
                                MessageBoxButton.YesNo, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly) == MessageBoxResult.No)
                            {
                                return;
                            }
                        }
                    }
                    File.Delete(pathName + @"\" + fileName);
                }

                // Create the new Settings file stream to write the serialized object to a file
                TextWriter WriteFileStream = new StreamWriter(pathName + @"\" + fileName);
                SerializerObj.Serialize(WriteFileStream, ToolSettings);
                WriteFileStream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error encountered saving file:\n" + pathName + @"\" + fileName + "\n\nError: " + ex.Message, "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
        }

        //-=-= FIXME: This will break if  a file name matches a camera serial number or a test pattern name, etc.
        public void SetActiveCamera(string CameraID)
        {
            try
            {
                int oldActiveIndex = ActiveIndex;

                // Camera is not the ActiveCamera so find it
                for (int i = 0; i < AvailableCameras.Count; i++)
                {
                    if (CameraID.Equals(AvailableCameras[i].Camera.SerialNumber))
                    {
                        if (!ActiveCamera.SerialNumber.Equals(CameraID))
                        {
                            DisplayCamera = AvailableCameras[i].Camera;
                            ActiveCamera = AvailableCameras[i].Camera;
                            ActiveIndex = i;
                        }
                        else if (DisplayCamera != ActiveCamera)
                        {
                            DisplayCamera = ActiveCamera;
                            ActiveIndex = i;
                        }
                        break;
                    }
                }

                //MvvmUtils.Invoke(() => OnActiveCameraAction(new ActiveCameraEventArgs(ActiveCamera)));
                if (oldActiveIndex == ActiveIndex)
                {
                    // Only call OnActiveCameraAction when the camera is the same.
                    // If it was changed OnNewActiveCameraAction was called in the setter for ActiveIndex.
                    OnActiveCameraAction(new ActiveCameraEventArgs(DisplayCamera));
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error in SetActiveCamera: " + ex.Message, "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
        }

        /// <summary>
        /// Finds the last connected camera and sets it as the active camera.
        /// If none are available, it returns without changing the ActiveCamera.
        /// </summary>
        public void SetActiveCameraToForefront()
        {
            if ((ActiveCamera.Type == CameraType.FILE) || (ActiveCamera.Type == CameraType.GENERATED))
            {
                for (int i = AvailableCameras.Count - 1; i >= 0 ; i--)
                {
                    if ((AvailableCameras[i].Type != CameraType.FILE) || (ActiveCamera.Type != CameraType.GENERATED))
                    {
                        if (AvailableCameras[i].Camera.IsConnected)
                        {
                            ActiveCamera = AvailableCameras[i].Camera;
                            ActiveIndex = i;
                            break;
                        }
                    }
                }
            }
            else
            {
                // Make sure the Active camera is th active index
                if (AvailableCameras[ActiveIndex].Camera != ActiveCamera)
                {
                    for (int i = AvailableCameras.Count - 1; i >= 0; i--)
                    {
                        if (AvailableCameras[i].Camera == ActiveCamera)
                        {
                            DisplayCamera = AvailableCameras[i].Camera;
                            ActiveIndex = i;
                            break;
                        }
                    }
                }
            }
        }

        public void SetStatusProgress(int count)
        {
            StatusProgressValue = 0;
            StatusProgressMaximum = count;
            StatusProgressVisible = true;
        }

        /// <summary>
        /// Called to initialize the ActualStatusProgressValue to 0 because in Continuous mode we loop from 0 to 100 repeatedly
        /// </summary>
        public void StatusProgressInitialize()
        {
            FramesCaptured = 0;
            ActualStatusProgressValue = 0;
            ActualStatusDisplayProgressValue = 0;
            StatusProgressValue = 0;
            StatusProgressVisible = true;
            StatusStartTime = DateTime.Now;
        }

        public void StatusProgressFinished()
        {
            // We don't want to stop the progress bar if we are only paused
            if (!GrabPaused)
            {
                StatusProgressVisible = false;
                ActualStatusProgressValue = 0;
                ActualStatusDisplayProgressValue = 0;
                StatusProgressTimer.Enabled = false;
                ActualStatusFrameRate = String.Empty;
                ProgressTimerClock = String.Empty;
                ExposureStatusText = String.Empty;
            }
        }

        public void StatusProgressIncrement(bool processed)
        {
            if (StatusProgressValue < StatusProgressMaximum)
            {
                try
                {
                    ActualStatusProgressValue++;
                    if (processed)
                    {
                        ActualStatusDisplayProgressValue++;
                    }
                    TimeSpan duration = DateTime.Now - StatusStartTime;
                    if (duration.TotalSeconds > 0)
                    {
                        ActualStatusFrameRate = (((double)ActualStatusProgressValue) / duration.TotalSeconds).ToString("0.00") + "/" +
                            (((double)ActualStatusDisplayProgressValue) / duration.TotalSeconds).ToString("0.00") +" fps";
                    }
                }
                catch
                {
                    // In case the user lets continuous run for a long time and ActualStatusProgressValue gets bigger than an int
                    ActualStatusProgressValue = 0;
                    ActualStatusDisplayProgressValue = 0;
                }
                StatusProgressValue++;
            }
            if (StatusProgressValue >= StatusProgressMaximum)
            {
                StatusProgressVisible = false;
            }
        }

        /// <summary>
        /// This override will display the amount of time remaining for the current frame in tenths of a second
        /// </summary>
        /// <param name="value"></param>
        /// <param name="unit"></param>
        public void SetStatusProgressTimer(bool useStatusBar)
        {
            long absoluteTime = ActiveCamera.GetAbsoluteExposureTime(ActiveCamera.ExposureTime, ActiveCamera.ExposureUnits);
            long shutterTime = ActiveCamera.GetAbsoluteExposureTime(ActiveCamera.ShutterOpenDelay,  ExposureUnits_Enum.units_mSecs);

            if (!ActiveCamera.ManualShutterControlEnabled)
            {
                absoluteTime += shutterTime;
            }

            StatusProgressTimer.Enabled = false;  // Make sure any previously running timer is stopped

            ProgressTimeStartTime = DateTime.Now;
            if (ActiveCamera.Type == CameraType.GEN1)
            {
                ProgressTimeMaximum = (int)(absoluteTime);  // Absolute ExposureTime is in milliseconds.
            }
            else if (ActiveCamera.Type == CameraType.GEN2)
            {
                ProgressTimeMaximum = (int)(absoluteTime / 100000000);  // Absolute ExposureTime is in 10us.
            }
            ProgressTimeValue = 0;

            UseStatusBar = useStatusBar;
            if (useStatusBar)
            {
                // This is the case where we are using the progress bar
                StatusProgressValue = 0;
                StatusProgressMaximum = ProgressTimeMaximum;
                StatusProgressVisible = true;
            }

            // On GEN1 cameras the progress of a frame is posted by the GrabFrameTimerEvent function so no need to have the timer running
            if ((ActiveCamera.Type != CameraType.GEN1) ||
                ((ActiveCamera.FrameCountType == FliFrameCount.CONTINUOUS) || (ActiveCamera.FrameCountType == FliFrameCount.MULTIPLE)))
            {
                StatusProgressTimer.Enabled = true;
            }
        }

        private void StatusProgressTimerEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            TimeSpan ts = DateTime.Now - ProgressTimeStartTime;
            ProgressTimeValue = (int)(ts.TotalMilliseconds / (ActiveCamera.IsLegacyCamera ? 1 : 100));

            if (ProgressTimeValue > ProgressTimeMaximum)
            {
                // We have reached the end so turn off the progress display
                StatusProgressTimer.Enabled = false;
            }
            else
            {
                long remainingTime = (ProgressTimeMaximum - ProgressTimeValue) * (ActiveCamera.IsLegacyCamera ? 1 : 100);  // Get the remaining time in milliseconds

                // Always show the remaining time in the bottom status bar
                ExposureStatusText = GetClockFromTime(remainingTime, ExposureUnits_Enum.units_mSecs);
                ProgressTimerClock = ExposureStatusText;

                // Only show it at the top if the capture type is a single frame
                if (UseStatusBar)
                {
                    StatusProgressValue = ProgressTimeValue;  // Move the progress bar
                }
            }
        }

        public bool ToolSettingsFileExists(string fileName)
        {
            string pathName = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), TOOL_SETTINGS_FOLDER);
            return File.Exists(pathName + @"\" + fileName);
        }

        public int UpdateExposureTimeSlider(int newTimeValue)
        {
            // Indecies go from 0 to 298
            // 0 to 10 are used as is
            // 11 to 28 are 5 values for each index (15, 20, ... 100)
            // 29 to 118 are 10 values for each index (110, 120, ... 1000)
            // 119 to 208 are 100 values for each index (1100, 1200, ... 10000)
            // 209 to 298 are 1000 values for each index (11000, 12000, ... 100000)
            // Any larger than this the user can change the scale or type in their value
            int newSliderValue = 0;

            if (newTimeValue < 11)
            {
                newSliderValue = (int)newTimeValue;
            }
            else if (newTimeValue < 100)
            {
                newSliderValue = 10 + Convert.ToInt32((newTimeValue - 10) / 5);
            }
            else if (newTimeValue < 1000)
            {
                newSliderValue = 29 + Convert.ToInt32((newTimeValue - 100) / 10);
            }
            else if (newTimeValue < 10000)
            {
                newSliderValue = 119 + Convert.ToInt32((newTimeValue - 1000) / 100);
            }
            else
            {
                newSliderValue = Math.Min(209 + Convert.ToInt32((newTimeValue - 10000) / 1000), 299);
            }
            return newSliderValue;
        }

        public void UpdateToolSettings()
        {
            // Camera Control Panel Properties
            ToolSettings.CoolerDutyCycleIsEnabledOn = ActiveCamera.CoolerDutyCycleIsEnabledOn;
            ToolSettings.CoolerDutyCyclePower = ActiveCamera.CoolerDutyCyclePower;
            ToolSettings.CoolerTemperatureOn = ActiveCamera.CoolerTemperatureOn;
            ToolSettings.DeviceType = ActiveCamera.DeviceType;
            ToolSettings.FanIsEnabledOn = ActiveCamera.FanIsEnabledOn;
            ToolSettings.FanPower = ActiveCamera.FanPower;
            ToolSettings.HeaterEnabledOn = ActiveCamera.HeaterEnabledOn;
            ToolSettings.HeaterPower = ActiveCamera.HeaterPower;
            ToolSettings.IlluminationOffDelay = ActiveCamera.IlluminationOffDelay;
            ToolSettings.IlluminationOn = ActiveCamera.IlluminationOn;
            ToolSettings.IlluminationOnDelay = ActiveCamera.IlluminationOnDelay;
            ToolSettings.LEDOn = ActiveCamera.LEDOn;
            ToolSettings.MaximumFrameStorage = ActiveCamera.MaximumFrameStorage;
            ToolSettings.ReadDuringExposureOn = ActiveCamera.ReadDuringExposureOn;
            ToolSettings.ShutterClose = ActiveCamera.ShutterClose;
            ToolSettings.ManualShutterControlEnabled = ActiveCamera.ManualShutterControlEnabled;
            ToolSettings.ShutterOpen = ActiveCamera.ShutterOpen;
            ToolSettings.TemperatureReadInterval = ActiveCamera.TemperatureReadInterval;
            ToolSettings.TargetTemperature = ActiveCamera.TargetTemperature;

            // Exposure Panel Properties
            ToolSettings.ExposureTime = ActiveCamera.ExposureTime;
            ToolSettings.ExposureUnits = ActiveCamera.ExposureUnits.GetHashCode();
            ToolSettings.ExposureDelayTime = ActiveCamera.ExposureDelayTime;
            ToolSettings.ExposureDelayUnits = ActiveCamera.ExposureDelayUnits.GetHashCode();
            ToolSettings.ExposureFrameIntervalTime = ActiveCamera.ExposureFrameIntervalTime;
            ToolSettings.ExposureFrameIntervalUnits = ActiveCamera.ExposureFrameIntervalUnits.GetHashCode();
            ToolSettings.Immediate = ActiveCamera.Immediate;
            ToolSettings.UseExposureDelay = ActiveCamera.UseExposureDelay;

            // File Name Setup Properties
            ToolSettings.BinningOrder = BinningOrder;
            ToolSettings.CameraModelOrder = CameraModelOrder;
            ToolSettings.DateOrder = DateOrder;
            ToolSettings.DiagRoutineOrder = DiagRoutineOrder;
            ToolSettings.ExposureTimeOrder = ExposureTimeOrder;
            ToolSettings.FrameNumberOrder = FrameNumberOrder;
            ToolSettings.GainOrder = GainOrder;
            ToolSettings.GrabTypeOrder = FrameTypeOrder;
            ToolSettings.IsCheckedBinning = IsCheckedBinning;
            ToolSettings.IsCheckedCameraModel = IsCheckedCameraModel;
            ToolSettings.IsCheckedDate = IsCheckedDate;
            ToolSettings.IsCheckedExposureTime = IsCheckedExposureTime;
            ToolSettings.IsCheckedFrameNumber = IsCheckedFrameNumber;
            ToolSettings.IsCheckedGain = IsCheckedGain;
            ToolSettings.IsCheckedMean = IsCheckedMean;
            ToolSettings.IsCheckedMode = IsCheckedMode;
            ToolSettings.IsCheckedPreamble = IsCheckedPreamble;
            ToolSettings.IsCheckedSensorTemperature = IsCheckedSensorTemperature;
            ToolSettings.IsCheckedSerialNumber = IsCheckedSerialNumber;
            ToolSettings.ModeOrder = ModeOrder;
            ToolSettings.Preamble = Preamble;
            ToolSettings.RawFileName = RawFileName;
            ToolSettings.SensorTemperatureOrder = SensorTemperatureOrder;
            ToolSettings.SerialNumberOrder = SerialNumberOrder;

            // File Panel Properties
            ToolSettings.RawFilePath = RawFilePath;

            // Force Training Panel Properties
            ToolSettings.IsCheckedAutoForceTraining = IsCheckedAutoForceTraining;
            ToolSettings.AutoForceRangeTemperature = AutoForceRangeTemperature;

            // Frame Panel Properties
            ToolSettings.ImageDataEnabled = ActiveCamera.ImageDataEnabled;
            ToolSettings.ImageHeight = ActiveCamera.ImageHeight;
            ToolSettings.ImageUpperLeftX = ActiveCamera.ImageUpperLeftX;
            ToolSettings.ImageUpperLeftY = ActiveCamera.ImageUpperLeftY;
            ToolSettings.ImageWidth = ActiveCamera.ImageWidth;

            // Grab Panel Properties
            ToolSettings.FrameCountType = ActiveCamera.FrameCountType.GetHashCode();
            ToolSettings.FrameType = ActiveCamera.FrameType.GetHashCode();

            // Merge Panel Properties
            ToolSettings.ADUsPerElectron = ActiveCamera.ADUsPerElectron;
            ToolSettings.GainRatio = ActiveCamera.GainRatio;
            ToolSettings.HighLowThreshold = ActiveCamera.HighLowThreshold;
            ToolSettings.LineOffset = ActiveCamera.LineOffset;

            // Sensor Panel Properties
            ToolSettings.BinningSquared = ActiveCamera.BinningSquared;
            ToolSettings.BlackLevelHigh = ActiveCamera.BlackLevelHigh;
            ToolSettings.BlackLevelLow = ActiveCamera.BlackLevelLow;
            ToolSettings.BlackSunHigh = ActiveCamera.BlackSunHigh;
            ToolSettings.BlackSunLow = ActiveCamera.BlackSunLow;
            ToolSettings.HBin = ActiveCamera.HBin;
            ToolSettings.HighGainIndex = ActiveCamera.HighGainIndex;
            ToolSettings.HorizontalScanDirInverted = ActiveCamera.HorizontalScanDirInverted;
            ToolSettings.LowGainIndex = ActiveCamera.LowGainIndex;
            ToolSettings.Mode = ActiveCamera.Mode;
            ToolSettings.PixelDepth = ActiveCamera.PixelDepth;
            ToolSettings.UseSoftwareBinning = ActiveCamera.UseSoftwareBinning;
            ToolSettings.VBin = ActiveCamera.VBin;
            ToolSettings.VerticalScanDirInverted = ActiveCamera.VerticalScanDirInverted;

            // Window Location Properties
            ToolSettings.MainWindowHeight = MainWindowHeight;
            ToolSettings.MainWindowLeft = MainWindowLeft;
            ToolSettings.MainWindowState = MainWindowState.GetHashCode();
            ToolSettings.MainWindowTop = MainWindowTop;
            ToolSettings.MainWindowWidth = MainWindowWidth;

            SaveToolSettings(TOOL_SETTINGS_PREVIOUS_FILE, false);
            SaveToolSettings(ActiveCamera.SerialNumber + ".xml", false, false);
        }

        public bool ValidCameraActive(bool raiseError = true)
        {
            bool isValid = false;

            if (ActiveIndex < 0)
            {
                if (raiseError)
                {
                    MessageBox.Show("There are no Cameras connected.", "Warning", MessageBoxButton.OK,
                        MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                }
                return isValid;
            }
            if (!AvailableCameras[ActiveIndex].IsLegacyCamera &&
                !AvailableCameras[ActiveIndex].IsGen2Camera)
            {
                // Search for a connected camera
                bool found = false;
                for (int i = 0; i < AvailableCameras.Count; i++)
                {
                    if (AvailableCameras[i].IsGen2Camera)
                    {
                        found = true;
                        ActiveIndex = i;
                        break;
                    }
                }
                if (!found)
                {
                    for (int i = 0; i < AvailableCameras.Count; i++)
                    {
                        if (AvailableCameras[i].IsLegacyCamera)
                        {
                            found = true;
                            ActiveIndex = i;
                            break;
                        }
                    }
                }
            }
            if (AvailableCameras[ActiveIndex].IsFromFile)
            {
                if (raiseError)
                {
                    MessageBox.Show("This function is not valid for Files. Select a Camera from the drop list.", "Warning", MessageBoxButton.OK,
                       MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                }
            }
            else if (AvailableCameras[ActiveIndex].IsGeneratedBitmap)
            {
                if (raiseError)
                {
                    MessageBox.Show("This function  is not valid for Generated Bitmaps. Select a Camera from the drop list.", "Warning", MessageBoxButton.OK,
                       MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                }
            }
            else if (!ActiveCamera.IsConnected)
            {
                if (raiseError)
                {
                    MessageBox.Show("Camera " + ActiveCamera.SerialNumber + " is no longer connected.\nSelect a Camera from the drop list that is connected.",
                       "Warning", MessageBoxButton.OK, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                }
            }
            else
            {
                isValid = true;
            }

            return isValid;
        }

        #endregion Public Methods

        //**************************************************************
        #region Private Methods

        /// <summary>
        /// Event Notifier that signals that a target USB device has been connected
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Monitor_Connected(object sender, USBMonitorEventArgs e)
        {
            try
            {
                var app = Application.Current;
                switch (e.Info.Key)
                {
                    case FLI_LEGACY_CAMERA_KEY:
                        MvvmUtils.Invoke(() => ExamineLegacyCamera(e.Info));
                        break;

                    case FLI_PRO_CAMERA_KEY:
                        MvvmUtils.Invoke(() => ExamineFliProCamera(e.Info));
                        break;

                    default:
                        // Not a VID we were looking for so just return
                        return;
                }
            }
            catch (FliException ex)
            {
                MessageBox.Show("Error in Monitor_Connected: " + ex.ToString(), "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error in Monitor_Connected: " + ex.Message, "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
        }

        private void Monitor_Disconnected(object sender, USBMonitorEventArgs e)
        {
            switch (e.Info.Key)
            {
                default:
                    // When a camera disconnects, make sure we save the current settings so that if it reconnects we can restore the state
                    UpdateToolSettings();

                    // Find the camera in the List and disconnect from it
                    for (int i = 0; i < AvailableCameras.Count; i++)
                    {
                        if ((AvailableCameras[i].usbDeviceInfo != null) && (e.Info.IID.Equals(AvailableCameras[i].usbDeviceInfo.IID)))
                        {
                            AvailableCameras[i].Camera.Disconnect();
                            AvailableCameras[i].Refresh();
                            // For now we want to insure that any cameras that were connected to this usbDeviceInfo.IID are marked as disconnected
                            // break;
                        }
                    }
                    break;
            }
        }

        private void ExamineFliProCamera(USBDeviceInfo usbDeviceInfo)
        {
            List<FPRODEVICEINFO> connectedList = new List<FPRODEVICEINFO>();
            int connectedIndex = 0;
            int availableIndex = 0;
            int status = -1;

            // Set cursor as hourglass
            Mouse.OverrideCursor = System.Windows.Input.Cursors.Wait;

            // Ran into a problem where the driver is not processing the connection as fast as we are.
            // That means we need to loop to see if it finds it. Remember, we were notified to check by the USB driver.
            for (int tries = 0; tries < 10; tries++)
            {
                connectedList = GetConnectedCameras();
                if (connectedList.Count > 0) break;
                Thread.Sleep(1000);
            }

            // Set cursor as default arrow
            Mouse.OverrideCursor = null;

            if (connectedList.Count == 0)
            {
                MessageBox.Show("Could not find the attached camera.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                return;
            }

            // First find the camera in the connected list from flipro library
            for (connectedIndex = 0; connectedIndex < connectedList.Count; connectedIndex++)
            {
                if (usbDeviceInfo.IID.Equals(connectedList[connectedIndex].cSerialNo))
                {
                    break;
                }
            }

            if (connectedIndex >= connectedList.Count)
            {
#if SERIAL_NUMBER_ENABLED
                // The camera was not found in the list from FliPro so stop processing
                return;
#else
                connectedIndex = 0; //-=-= Take first camera if device serial numbers are not defined
#endif
            }

            FPRODEVICEINFO deviceInfo = connectedList[connectedIndex];

            // Loop through the list of Available Cameras 
            for (availableIndex = 0; availableIndex < AvailableCameras.Count; availableIndex++)
            {
                // Only want to examine Gen2 Cameras here
                if (AvailableCameras[availableIndex].IsGen2Camera)
                {
                    // See if this Fli Pro Camera has the same Serial Number
                    if (deviceInfo.cSerialNo.Equals(AvailableCameras[availableIndex].DeviceInfo.cSerialNo))
                    {
                        // The camera has previouly been attached and is being reattached.
                        break;
                    }
                }
            }

            if (availableIndex >= AvailableCameras.Count)
            {
                // This is a new camera so we need to create it's class and insert it into AvailableCameras
                FliProCamera newCamera = new FliProCameraGen2(deviceInfo);
                FliProCameraListEntry newCameraListEntry = new FliProCameraListEntry(newCamera);
                newCameraListEntry.usbDeviceInfo = usbDeviceInfo;
                status = newCamera.Connect();
                if (status < 0)
                {
                    MessageBox.Show("Error encountered trying to connect to " + deviceInfo.cFriendlyName + " - " + deviceInfo.cSerialNo,
                        "Warning", MessageBoxButton.OK, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                    return;
                }

                AvailableCameras.Add(newCameraListEntry);
                availableIndex = AvailableCameras.Count - 1;
            }
            else
            {
                // This camera exists in our list so just replace it with a new Connect
                AvailableCameras[availableIndex].insertDate = DateTime.Now;
                AvailableCameras[availableIndex].usbDeviceInfo = usbDeviceInfo;
                AvailableCameras[availableIndex].DeviceInfo = deviceInfo;
                AvailableCameras[availableIndex].Camera.IsInitializedFromSettings = false;
                status = AvailableCameras[availableIndex].Camera.Connect();
                if (status < 0)
                {
                    MessageBox.Show("Error encountered trying to connect to " + deviceInfo.cFriendlyName + " - " + deviceInfo.cSerialNo,
                        "Warning", MessageBoxButton.OK, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                    return;
                }

                AvailableCameras[availableIndex].Camera.SetDeviceInfo(deviceInfo);
                AvailableCameras[availableIndex].Refresh();
            }

            // Set this camera as the Active Camera
            ActiveCamera = AvailableCameras[availableIndex].Camera;
            ActiveIndex = availableIndex;
            OnActiveCameraAction(new ActiveCameraEventArgs(AvailableCameras[availableIndex].Camera));
            //OnNewActiveCameraAction(new ActiveCameraEventArgs(ActiveCamera));
            if (_instance != null)
            {
                // Only call this after the FliProCameraManager has been initialized.
                // Otherwise we get a recurse error because a downstream call needs an instance of 
                ActiveCamera.SetForcePropertyChange(true);
                LoadToolSettings(ActiveCamera.SerialNumber + ".xml", false, false);
                PopulateCameraWithToolSettings(ActiveCamera);
                ActiveCamera.SetForcePropertyChange(false);
            }
        }

        private void ExamineLegacyCamera(USBDeviceInfo usbDeviceInfo)
        {
            int availableIndex = 0;
            int foundIndex = -1;
            List<int> openLegacyIndicesList = new List<int>();
            int status = -1;

            // Search to see if this camera has been attached in the past
            for (availableIndex = 0; availableIndex < AvailableCameras.Count; availableIndex++)
            {
                // Only want to examine Legacy Cameras here
                if (AvailableCameras[availableIndex].IsLegacyCamera)
                {
                    if (usbDeviceInfo.IID.Equals(AvailableCameras[availableIndex].usbDeviceInfo.IID))
                    {
                        // We found the camera in the list so save the index
                        foundIndex = availableIndex;
                    }

                    // Keep track of all the open legacy cameras so we don't try and reopen them below
                    if (AvailableCameras[availableIndex].Camera.IsConnected)
                    {
                        openLegacyIndicesList.Add(availableIndex);
                    }
                }
            }

            if (foundIndex < 0)
            {
                // This is a new camera so we need to check the list and get the deviceId
                List<FliDeviceId> connectedList = GetDeviceList();
                int connectedIndex = 0;
                FliProCameraLegacy fliProCameraLegacy = null;

                // First find the camera in the connected list from fli library
                for (connectedIndex = 0; connectedIndex < connectedList.Count; connectedIndex++)
                {
                    int openIndex = 0;
                    for (openIndex = 0; openIndex < openLegacyIndicesList.Count; openIndex++)
                    {
                        int index = openLegacyIndicesList[openIndex];
                        if (connectedList[connectedIndex].Name.Equals(((FliProCameraLegacy)AvailableCameras[index].Camera).Name))
                        {
                            // Then we found this connected camera in our list of open cameras so we don't need to open it
                            break;
                        }
                    }
                    if (openIndex >= openLegacyIndicesList.Count)
                    {
                        // We didn't find this camera in the open list so create it and connect
                        fliProCameraLegacy = new FliProCameraLegacy(connectedList[connectedIndex]);
                        status = fliProCameraLegacy.Connect();
                        if (status < 0)
                        {
                            MessageBox.Show("Error encountered trying to connect to " + usbDeviceInfo.Name + " - " + usbDeviceInfo.IID,
                                "Warning", MessageBoxButton.OK, MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                            return;
                        }
                    }
                    else
                    {
                        fliProCameraLegacy = (FliProCameraLegacy)AvailableCameras[openLegacyIndicesList[openIndex]].Camera;
                    }

                    if (usbDeviceInfo.IID.Equals(fliProCameraLegacy.SerialNumber, StringComparison.OrdinalIgnoreCase))
                    {
                        // Then this connected camera is our target camera
                        break;
                    }
                    else
                    {
                        // This is not our camera so if we opened it we need to close it
                        if (openIndex >= openLegacyIndicesList.Count)
                        {
                            fliProCameraLegacy.Disconnect();
                        }
                    }
                }

                if (connectedIndex >= connectedList.Count)
                {
                    // The camera was not found in the list of connected Legacy cameras so stop processing
                    return;
                }

                // Getting here means this is the new camera so we need to create it's class and insert it into AvailableCameras
                FliProCameraListEntry newCameraListEntry = new FliProCameraListEntry(fliProCameraLegacy);
                newCameraListEntry.LegacyDeviceId = connectedList[connectedIndex];
                newCameraListEntry.usbDeviceInfo = usbDeviceInfo;
                AvailableCameras.Add(newCameraListEntry);
                foundIndex = AvailableCameras.Count - 1;
            }
            else
            {
                // This camera exists in our list so just replace it with a new Connect
                AvailableCameras[foundIndex].insertDate = DateTime.Now;
                AvailableCameras[foundIndex].usbDeviceInfo = usbDeviceInfo;  // In case the port is different
                status = ((FliProCameraLegacy)AvailableCameras[foundIndex].Camera).Connect();
                if (status < 0)
                {
                    MessageBox.Show("Error encountered trying to connect to " + usbDeviceInfo.Name + " - " + usbDeviceInfo.IID, "Warning", MessageBoxButton.OK,
                                MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                    return;
                }
                AvailableCameras[foundIndex].Refresh();
            }

            // Set this camera as the Active Camera
            ActiveCamera = AvailableCameras[foundIndex].Camera;
            ActiveIndex = foundIndex;
            OnActiveCameraAction(new ActiveCameraEventArgs(AvailableCameras[foundIndex].Camera));
            if (_instance != null)
            {
                // Only call this after the FliProCameraManager has been initialized.
                // Otherwise we get a recurse error because a downstream call needs an instance of 
                ActiveCamera.SetForcePropertyChange(true);
                LoadToolSettings(ActiveCamera.SerialNumber + ".xml", false, false);
                ActiveCamera.SetForcePropertyChange(false);
            }
        }

        #endregion Private Methods

        //**************************************************************
        #region Public Events

        public event EventHandler<ActiveCameraEventArgs> ActiveCameraAction;

        public virtual void OnActiveCameraAction(ActiveCameraEventArgs e)
        {
            EventHandler<ActiveCameraEventArgs> handler = ActiveCameraAction;
            if (handler != null)
            {
                MvvmUtils.Invoke(() => handler(this, e));
            }
        }

        public event EventHandler<ActiveJobCompleteEventArgs> ActiveJobCompleteAction;

        public virtual void OnActiveJobCompleteAction(ActiveJobCompleteEventArgs e)
        {
            EventHandler<ActiveJobCompleteEventArgs> handler = ActiveJobCompleteAction;
            if (handler != null)
            {
                MvvmUtils.Invoke(() => handler(this, e));
            }
        }

        public event EventHandler<ActiveCameraEventArgs> NewActiveCameraAction;

        protected virtual void OnNewActiveCameraAction(ActiveCameraEventArgs e)
        {
            EventHandler<ActiveCameraEventArgs> handler = NewActiveCameraAction;
            if (handler != null)
            {
                MvvmUtils.Invoke(() => handler(this, e));
            }
        }

        public event EventHandler<EventArgs> SettingsAction;

        protected virtual void OnSettingsAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = SettingsAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        #endregion Public Events

    }
}
