//**************************************************************
// Finger Lakes Instrumentation Camera Interface library
//
// The NativeMethods class contains static methods which use interp services
// to properly invoke and marshal functions exported by libflipro.dll (and libflieng.dll).
//
// These method definitions MUST be consistent with the 
// versions of libflipro.dll (and libflieng.dll) referenced by the application.
// Specifically, the method definitions must match the function prototype
// declarations in libflipro.h
//
// NOTE: These static methods are NOT guaranteed to be MTSafe. 
// They are intended to be invoked by a higher abstraction level layer 
// providing the appropriate critical region protection.
//
// Refer to FLI Software Development Library documentation and libflipro.h
// Doxygen comments for detailed descriptions of each function.

using System;
using System.Runtime.InteropServices;
using System.Text;
using static FliProCameraLib.FliPro;

namespace FliProCameraLib
{
    //using static FliPro;  Not ready for C#6 yet

    internal static partial class NativeMethods
    {
        //**************************************************************
        #region Library Functions

        //-------------------------------------------
        /// <summary>
        /// Get the current API dll version string
        /// Prior to call, version.Capacity must be >= len
        /// </summary>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_GetAPIVersion(
                    [Out, MarshalAs(UnmanagedType.LPWStr)]
                    StringBuilder version,
                    uint uiLength);

        #endregion Library Functions

        //**************************************************************
        #region Camera Enumeration, Open, Close

        //-------------------------------------------
        /// <summary>
        /// Retrieve a list of all connected cameras
        /// </summary>
        /// <param name="DeviceInfo">FPRODEVICEINFO array to receive info on all connected cameras</param>
        /// <param name="NumDevices">On input, max capacity of DeviceInfo[]. On output, actual camera count </param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_GetCameraList(
                    [Out] FPRODEVICEINFO[] DeviceInfo,
                    ref uint NumDevices);

        //-------------------------------------------
        /// <summary>
        /// Open a connection to the specified camera
        /// </summary>
        /// <param name="DevInfo"></param>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_Open(
                    ref FPRODEVICEINFO DevInfo,
                    out int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Close the connection to a camera
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_Close(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Get the current camera firmware and fpga version strings
        /// </summary>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_GetDeviceVersion(
            int iHandle,
            out FPRODEVICEVERSIONS Version);

        #endregion Device Enumeration, Open, Close

        //**************************************************************
        #region Frame Data Functions

        //-------------------------------------------
        /// <summary>
        /// Abort the current frame capture sequence
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureAbort(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Terminates the current frame capture sequence and makes available the data accumulated so far
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureEnd(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Start a frame capture sequence
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiFrameCount">Number of frames to capture. 0 = no limit</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureStart(
                    int iHandle,
                    uint uiFrameCount);

        //-------------------------------------------
        /// <summary>
        /// Stop a frame capture sequence
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureStop(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Capture one thumbnail frame image
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureThumbnail(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Enable/disable image data in frame. When disabled, only dummy pixels (if enabled) are produced
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnableImageData"></param>
        /// <param name="bEnableDummyPixels"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetImageDataEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bEnable);

        //-------------------------------------------
        /// <summary>
        /// Enable/disable image data in frame. When disabled, only dummy pixels (if enabled) are produced
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnableImageData"></param>
        /// <param name="bEnableDummyPixels"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetImageDataEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable);

        //-------------------------------------------
        /// <summary>
        /// Get Enable test image generation settings
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable"></param>
        /// <param name="eFormat"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetTestImageEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    out TESTIMAGEMODE eFormat);

        //-------------------------------------------
        /// <summary>
        /// Get the dummy pixel configuration to be appended to row data.
        /// If enabled, eight dummy pixels will are appended to every other row
        /// starting with the second row of data.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Enable"> true if dummy pixels will be appended</param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetDummyPixelEnable(
                    int iHandle,
                    out bool Enable);

        //-------------------------------------------
        /// <summary>
        /// Get the dummy row configuration to be appended to frame data.
        /// If enabled, the specified number of dummy rows are appended after the last row
        /// of image frame data.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pNumRowss">number of rows to append when Enable = true</param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetDummyRowPostFrameEnable(
                    int iHandle,
                    out int pNumRows);

        //-------------------------------------------
        /// <summary>
        /// Get the dummy row configuration to be prepended to frame data.
        /// If enabled, the specified number of dummy rows are prepended before the first row
        /// of image frame data.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pNumRowss">number of rows to prepend when Enable = true</param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetDummyRowPreFrameEnable(
                    int iHandle,
                    out int pNumRows);


        //-------------------------------------------
        /// <summary>
        /// Get the current pixel depth and LSB position
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="PixelDepth">Current pixel depth in bits</param>
        /// <param name="PixedlLSB">The Pixel LSB offset</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetPixelConfig(
                    int iHandle,
                    out uint PixelDepth,
                    out uint PixelLSB);

        //-------------------------------------------
        /// <summary>
        /// Get the next available Video frame
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="FrameData">Buffer to receive frame data</param>
        /// <param name="Size">On input: desired frame size. On return: actual frame size</param>
        /// <param name="uiTimeoutMS">Timeout in milliSec</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetVideoFrame(
                    int iHandle,
                    IntPtr pFrameData,
                    ref uint Size,
                    uint uiTimeoutMS);

        //-------------------------------------------
        /// <summary>
        /// Get the next available Video frame after an external trigger has fired
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="FrameData">Buffer to receive frame data</param>
        /// <param name="Size">On input: desired frame size. On return: actual frame size</param>
        /// <param name="uiTimeoutMS">Timeout in milliSec</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetVideoFrameExt(
                    int iHandle,
                    IntPtr pFrameData,
                    ref uint Size);

        //-------------------------------------------
        /// <summary>
        /// Get the next available Video frame
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="FrameData">Buffer to receive frame data</param>
        /// <param name="Size">On input: desired frame size. On return: actual frame size</param>
        /// <param name="uiTimeoutMS">Timeout in milliSec</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetVideoFrame(
                    int iHandle,
                    byte[] FrameData,
                    ref uint Size,
                    uint uiTimeoutMS,
                    [MarshalAs(UnmanagedType.I1)]
                    bool Trigger = false);

        //-------------------------------------------
        /// <summary>
        /// Check if a frame is available to transfer
        /// Note: This may not be useable due to hw time constraints.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bAvailable">true if a frame is available</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_IsAvailable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.I1)]
                    out bool bAvailable);

        //-------------------------------------------
        /// <summary>
        /// Enable/Disable dummy pixels at the end of image row data.
        /// When enabled, eight dummy pixels are appended to every other row of image data
        /// starting withthe second row of data.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetDummyPixelEnable(
                    int iHandle,
                    bool bEnable);

        //-------------------------------------------
        /// <summary>
        /// Sets the configuratrion of dummy rows to be appended afte the last row
        /// of frame data
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiNumRows"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetReferenceRowPostFrameCount(
                    int iHandle,
                    uint uiNumRows);

        //-------------------------------------------
        /// <summary>
        /// Sets the configuratrion of dummy rows to be prepended before the first row
        /// of frame data
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiNumRows"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetReferenceRowPreFrameCount(
                    int iHandle,
                    uint uiNumRows);

        //-------------------------------------------
        /// <summary>
        /// Enable test image generation
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable"></param>
        /// <param name="eFormat"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetTestImageEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    TESTIMAGEMODE eFormat);

        //-------------------------------------------
        /// <summary>
        /// Define the Tracking frame area. During tracking the entire row is captured so we don't need a width parameter.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiStartRow"></param>
        /// <param name="uiEndRow"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetTrackingArea(
                    int iHandle,
                    uint uiStartRow,
                    uint uiEndRow);

        //-------------------------------------------
        /// <summary>
        /// Enable/Disable Tracking frame generation
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiNumTrackingFrames"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetTrackingAreaEnable(
                    int iHandle,
                    uint uiNumTrackingFrames);

        //-------------------------------------------
        /// <summary>
        /// Set the current pixel configuration
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiPixelDepth">pixel depth in bits</param>
        /// <param name="uiPixelLSB">pixel LSB offset</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetPixelConfig(
                    int iHandle,
                    uint uiPixelDepth,
                    uint uiPixelLSB);

        //-------------------------------------------
        /// <summary>
        /// Define the image size and location
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiColOffset">start column of the image frame</param>
        /// <param name="uiRowOffset">start row of the image frame</param>
        /// <param name="uiWidth">Width of image frame in pixels</param>
        /// <param name="uiHeight">Height of image frame in pixels</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetImageArea(
                    int iHandle,
                    uint uiColOffset,
                    uint uiRowOffset,
                    uint uiWidth,
                    uint uiHeight);

        #endregion Frame Data Functions

        //**************************************************************
        #region Camera Control

        //-------------------------------------------
        /// <summary>
        /// Read the base temperature of the camera
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Temp">temperature in deg C</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetBaseTemperature(
                    int iHandle,
                    out int Temp);

        //-------------------------------------------
        /// <summary>
        /// Read the current exposure time for the image sensor (in nanoseconds)
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiExposureTime">Exposure Time (in 10uSec increments)</param>
        /// <param name="uiDelayTime">The time in nanoseconds between consecutive exposures</param>
        /// <param name="bImmediate">Flag to not finish flushing the sensor before exposing</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetExposure(
                    int iHandle,
                    out long ExposureTime,
                    out long DelayTime,
                    out bool Immediate);

        //-------------------------------------------
        /// <summary>
        /// Get the current external trigger settings
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">External trigger enabled if true</param>
        /// <param name="eTrigType">The type of external trigger as well as the edge to trigger on</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetExternalTriggerEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    out FliTriggerMode eTrigType);

        //-------------------------------------------
        /// <summary>
        /// Get the current fan enable status
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">Fan on if true</param>
        /// <param name="uiPwrPercentage">Fan power level (0-100)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetFanEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bOn);

        //-------------------------------------------
        /// <summary>
        /// Get the Frame interval in nanoseconds
        /// </summary>
        /// <param name="iHandle"></param>
        /// <param name="FrameInterval"></param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetFrameInterval(
                    int iHandle,
                    out long FrameInterval);

        //-------------------------------------------
        /// <summary>
        /// Get the current heater configuration
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">Heater on if true</param>
        /// <param name="uiPwrPercentage">Heater power level (0-100)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetHeaterConfig(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bOn,
                    out uint uiPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Get the current heater configuration
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiPwrPercentage">Heater power level (0-100)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetHeaterPower(
                    int iHandle,
                    out uint uiPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Get the current illumination delays
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOnDelay">Delay for turning on illumination (in tbd mSec increments)</param>
        /// <param name="uiOffDelay">Delay for turning off illumination (in tbd mSec increments)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetIlluminationDelay(
                    int iHandle,
                    out uint uiOnDelay,
                    out uint uiOffDelay);

        //-------------------------------------------
        /// <summary>
        /// Get the current illumination on/off status
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">Illumination on if true</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetIlluminationOn(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bOn);

        //-------------------------------------------
        /// <summary>
        /// Reads the internal sensor temperature of the camera.  
        /// Note that if this is called during an exposure and 'Read During Exposure' is not
        /// enabled, the last value successfully read will be returned. 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Temp">temperature in deg C</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetSensorTemperature(
                    int iHandle,
                    out int Temp);

        //-------------------------------------------
        /// <summary>
        /// Read the "sensor temperature during exposure" enable flag
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">Read during exposure enabled if true</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetSensorTemperatureReadEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bEnable);

        //-------------------------------------------
        /// <summary>
        /// Read the sensor temperatures
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pAmbientTemp">The current Ambient temperature on the camera</param>
        /// <param name="pBaseTemp">The current Base temperature on the camera</param>
        /// <param name="pAmbientTemp">The current temperature of the Cooler unit on the camera</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetTemperatures(
                    int iHandle,
                    ref double pAmbientTemp,
                    ref double pBaseTemp,
                    ref double pCoolerTemp);

        //-------------------------------------------
        /// <summary>
        /// Read the Cooler Duty Cycle
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pDutyCycle">The percentage at which the Cooler is running </param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetCoolerDutyCycle(
                    int iHandle,
                    ref int pDutyCycle);
        
        //-------------------------------------------
        /// <summary>
        /// Get the max time full power is applied to shutter 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiTime">max full power duration (in 10 mSec increments)/param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetShutterFullPowerTime(
                    int iHandle,
                    out uint uiTime);

        //-------------------------------------------
        /// <summary>
        /// Get the base temperature setpoint
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="SetPoint">Temperature setpoint (in 0.1 deg C increments)/param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetTemperatureSetPoint(
                    int iHandle,
                    out int SetPoint);

        //-------------------------------------------
        /// <summary>
        /// Set exposure time
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiExposureTime">Exposure time in nanoseconds</param>
        /// <param name="uiDelayTime">The time in nanoseconds between consecutive exposures</param>
        /// <param name="bImmediate">Flag to not finish flushing the sensor before exposing</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetExposure(
                    int iHandle,
                    long uiExposureTime,
                    long uiDelayTime,
                    bool bImmediate);

        //-------------------------------------------
        /// <summary>
        /// Set external trigger enable and active edge
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">external trigger enabled if true</param>
        /// <param name="eTrigType">The type of external trigger as well as the edge to trigger on</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetExternalTriggerEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    FliTriggerMode eTrigType);


        //-------------------------------------------
        /// <summary>
        /// Turn fan on/off 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn"></param>
        /// <param name="uiPwrPercentage"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetFanEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOn);

        //-------------------------------------------
        /// <summary>
        /// Set the frame interval in nanoseconds
        /// </summary>
        /// <param name="iHandle"></param>
        /// <param name="FrameInterval"></param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetFrameInterval(
                    int iHandle,
                    long FrameInterval);

        //-------------------------------------------
        /// <summary>
        /// Turn Heater on/off at specified power level
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn"></param>
        /// <param name="uiPwrPercentage"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetHeaterConfig(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOn,
                    byte uiPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Turn Heater on/off at specified power level
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiPwrPercentage"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetHeaterPower(
                    int iHandle,
                    int uiPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Specify the illumination on and off delays
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOnDelay">Delay for turning on illumination (in tbd mSec increments)</param>
        /// <param name="uiOffDelay">Delay for turning off illumination (in tbd mSec increments)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetIlluminationDelay(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U2)]
                    ushort uiOnDelay,
                    [MarshalAs(UnmanagedType.U2)]
                    ushort uiOffDelay);

        //-------------------------------------------
        /// <summary>
        /// Turn Illumination on/off
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">turn illumination on if true, off if false</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetIlluminationOn(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOn);

        //-------------------------------------------
        /// <summary>
        /// Turn LED on/off
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">Turn LED on if true, off if false</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetLED(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOn);

        //-------------------------------------------
        /// <summary>
        /// Set the "sensor temperature during exposure" enable flag
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">Read during exposure enabled if true</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetSensorTemperatureReadEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable);

        //-------------------------------------------
        /// <summary>
        /// Specify the shutter full power time
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiTime">max full power duration (in 10mSec increments)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetShutterFullPowerTime(
                    int iHandle,
                    byte uiTime);

        //-------------------------------------------
        /// <summary>
        /// Open/Close the shutter
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOpen"Open shutter if true, close if false</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetShutterOpen(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOpen);


        //-------------------------------------------
        /// <summary>
        /// Overrides the normal shutter control
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOverride"Overrides shutter control if true, normal if false</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetShutterOverride(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOverride);

        //-------------------------------------------
        /// <summary>
        /// Set the base temperature setpoint
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="SetPoint">Temperature setpoint (in 0.1 deg C increments)/param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetTemperatureSetPoint(
                    int iHandle,
                    double SetPoint);

        #endregion Camera Control

        //**************************************************************
        #region Sensor Functions

        //-------------------------------------------
        /// <summary>
        /// Get the horizontal and vertical binning values
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiVertical">Vertical binning setting</param>
        /// <param name="uiHorizontal">Horizontal binning setting</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetBinning(
                    int iHandle,
                    out uint uiVertical,
                    out uint uiHorizontal);

        //-------------------------------------------
        /// <summary>
        /// Get the current black level adjust value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pAdjustValue"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetBlackLevelAdjust(
                    int iHandle,
                    out uint pAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Get the current Black Sun adjust value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pAdjustValue"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetBlackSunAdjust(
                    int iHandle,
                    out uint pAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Get the collection of camera capabilities
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Cap">Capabilities struct</param>
        /// <param name="CauiCapLength">Length of capabilities struct in bytes</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetCapabilities(
                    int iHandle,
                    out CAMERACAP Cap,
                    ref uint uiCapLength);

        //-------------------------------------------
        /// <summary>
        /// Get the current setting for the specified gain index
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eTable">Table to return, LDR, HDR, Global</param>
        /// <param name="uiGainIndex">returned gain index value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetGainIndex(
                    int iHandle,
                    GainTable eTable,
                    out uint uiGainIndex);

        //-------------------------------------------
        /// <summary>
        /// Get the specified Gain Table
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eTable">Table to return, LDR, HDR, Global</param>
        /// <param name="uiGainValues">Array of returned gain values</param>
        /// <param name="uiNumEntries">Size of GainValues array on entry, number returned on exit</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetGainTable(
                    int iHandle,
                    GainTable eTable,
                    uint[] uiGainValues,
                    ref uint uiNumEntries);

        internal static int FPROSensor_GetGainTableEx(
                     int iHandle,
                    GainTable eTable,
                    uint[] uiGainValues,
                    ref uint uiNumEntries)
        {
            if (uiNumEntries > 0)
            {
                return FPROSensor_GetGainTable(iHandle, eTable, uiGainValues, ref uiNumEntries);
            }
            else
            {
                return 0;
            }
        }


        //-------------------------------------------
        /// <summary>
        /// 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bIsHDR"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetHDREnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bIsHDR);


        //-------------------------------------------
        /// <summary>
        /// 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiModeIndex"></param>
        /// <param name="Mode"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetMode(
                    int iHandle,
                    uint uiModeIndex,
                    out SENSEMODE Mode);

        //-------------------------------------------
        /// <summary>
        /// Get the number of modes supported by the camera
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Count"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetModeCount(
                    int iHandle,
                    out uint Count,
                    out uint CurrentMode);

        //-------------------------------------------
        /// <summary>
        /// Get the camera horizontal and vertical scan direction
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pVInverted"></param>
        /// <param name="pHInverted"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetScanDirection(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool HInverted,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool VInverted);

        //-------------------------------------------
        /// <summary>
        /// Set the horizontal and vertical binning value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiVertical"></param>
        /// <param name="uiHorizontal"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetBinning(
                    int iHandle,
                    uint uiVertical,
                    uint uiHorizontal);

        //-------------------------------------------
        /// <summary>
        /// Set the black level adjust value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiAdjustValue"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetBlackLevelAdjust(
                    int iHandle,
                    uint uiAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Set the black sun adjust value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiAdjustValue"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetBlackSunAdjust(
                    int iHandle,
                    uint uiAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Set the current index for the specified gain table
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eTable">Table to return, LDR, HDR, Global</param>
        /// <param name="uiGainIndex">index to set</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetGainIndex(
                    int iHandle,
                    GainTable eTable,
                    uint uiGainIndex);

        //-------------------------------------------
        /// <summary>
        /// Set the camera mode
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiModeIndex"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetMode(
                    int iHandle,
                    uint uiModeIndex);

        //-------------------------------------------
        /// <summary>
        /// Set the camera horizontal and vertical scan direction
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pVInverted"></param>
        /// <param name="pHInverted"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetScanDirection(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool pHInverted,
                    [MarshalAs(UnmanagedType.U1)]
                    bool pVInverted);

        #endregion Sensor Functions

        //**************************************************************
        #region Low Level Interface Functions

        //-------------------------------------------
        /// <summary>
        /// Send a Raw USB Message
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="TxData"></param>
        /// <param name="uiTxLength"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCmd_SendRaw(
                    int iHandle,
                    byte[] TxData,
                    uint uiTxLength);

        //-------------------------------------------
        /// <summary>
        /// Send a Raw USB Message and Receive a Response
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="TxData">Byte array containg the data to send</param>
        /// <param name="uiTxLength">The length of the byte array containg the data to send</param>
        /// <param name="RxData">The returned data byte array</param>
        /// <param name="uiRxLength">the length of the returned data byte array</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCmd_SendRecvRaw(
                    int iHandle,
                    byte[] TxData,
                    uint uiTxLength,
                    [Out] byte[] RxData,
                    ref uint uiRxLength);

        //-------------------------------------------
        /// <summary>
        /// Read an FPGA Register
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiReg"></param>
        /// <param name="pValue"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCmd_ReadReg(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    FPROFPGAREG uiReg,
                    out uint Value);

        //-------------------------------------------
        /// <summary>
        /// Write a value to an FPGA Register
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiReg"></param>
        /// <param name="uiValue"></param>
        /// <param name="uiMask"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCmd_WriteReg(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    FPROFPGAREG uiReg,
                    uint uiValue,
                    uint uiMask);

        #endregion Low Level Interface Functions

        //**************************************************************
        #region Debug Functions

        //-------------------------------------------
        /// <summary>
        /// Set the library debug level
        /// </summary>
        /// <param name="bEnable"> enable debug logging if true</param>
        /// <param name="level">Maximum debug level to be logged</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPRODebug_EnableLevel(
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    DBGLEVEL level);

        //-------------------------------------------
        /// <summary>
        /// Write a message to the debug log
        /// </summary>
        /// <param name="level">Debug log level</param>
        /// <param name="message">message content</param>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        internal static extern void FPRODebug_Write(
                    [MarshalAs(UnmanagedType.I4)]
                    DBGLEVEL level,
                    [MarshalAs(UnmanagedType.LPStr)]
                    string message);

        #endregion Debug Functions

        //**************************************************************
        #region Engineering Functions

        #region Engineering Private Static Extern
        ///
        /// For all of these Private Static External functions, the Managed CS code must call the
        /// managed wrapper function below.  They are of the same name, but with a CS prefix.  The
        /// reason for this is due to the more complex marshalling code required to get the data
        /// back and forth from the dll and managed CS space.
        /// 

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_GetConfig(
                    int iHandle,
                    ref MFENG_CFGCAM cfgCam,
                    IntPtr pProgressCallback);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_GetConfigFromFile(
                    [MarshalAs(UnmanagedType.LPWStr)]
                    StringBuilder version, 
                    ref MFENG_CFGCAM cfgCam);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_FreeConfig(
                    ref MFENG_CFGCAM cfgCam);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_SaveConfigToBin(
                   [MarshalAs(UnmanagedType.LPWStr)]
                    StringBuilder version,
                   ref MFENG_CFGCAM cfgCam);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_SaveConfigToCSV(
                    [MarshalAs(UnmanagedType.LPWStr)]
                    StringBuilder version,
                    ref MFENG_CFGCAM cfgCam);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_SetConfig(
                     int iHandle,
                     ref MFENG_CFGCAM cfgCam,
                     IntPtr pProgressCallback);


        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCfg_SetDeviceRegs(
                      int iHandle,
                      uint uiDeviceId,
                      ref IntPtr pRegs,
                      uint uiNumRegs);

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCfg_SetHSMTimings(
                      int iHandle,
                      ref IntPtr pIdleTimings,
                      uint uiNumIdleTimings,
                      ref IntPtr pBinningTimings,
                      uint uiNumBinningTimings);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCfg_SetVSMTimings(
                      int iHandle,
                      ref IntPtr pIdleTimings,
                      uint uiNumIdleTimings,
                      ref IntPtr pLongExposureTimings,
                      uint uiNumLongExposureTimings);

        #endregion Engineering Private Static Extern

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENGCam_SetSerialNumber(
                      int iHandle,
                      [MarshalAs(UnmanagedType.LPWStr)]
                      StringBuilder serialNumber,
                      uint uiNumChars);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_GetFwVersion(
                      int iHandle,
                      out uint pVersion);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_GetCoolerDAC(
                      int iHandle,
                      out uint pValue);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_GetCoolerDutyOverride(
                      int iHandle,
                      [MarshalAs(UnmanagedType.U1)]
                      out bool pEnable,
                      out uint pPercent);

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG_GetFanSpeed(
                      int iHandle,
                      out uint pPwrPercentage);


        //-------------------------------------------
        /// <summary>
        /// Get the time between setting Shutter Open and the shutter actually opening or closing
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOpenDelay">Shutter open delay (in tbd mSec increments)/param>
        /// <param name="uiCloseDelay">Shutter close delay (in tbd mSec increments)/param>
        /// <returns>negative on error</returns>
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FENG_GetShutterDelay(
                    int iHandle,
                    out uint uiOpenDelay,
                    out uint uiCloseDelay);

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_SetCoolerDAC(
                      int iHandle,
                      [MarshalAs(UnmanagedType.U4)]
                      uint pValue);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_SetCoolerDutyOverride(
                      int iHandle,
                      [MarshalAs(UnmanagedType.U1)]
                      bool pEnable,
                      [MarshalAs(UnmanagedType.U4)]
                      uint pPercent);

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG_SetFanSpeed(
                      int iHandle,
                      [MarshalAs(UnmanagedType.U4)]
                      uint pPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Specify the shutter open and close delays
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOpenDelay">Shutter open delay (in tbd mSec increments)</param>
        /// <param name="uiCloseDelay">Shutter close delay (in tbd mSec increments)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FENG_SetShutterDelay(
                    int iHandle,
                    uint uiOpenDelay,
                    int uiCloseDelay);


        #region Engineering Marshalling Wrapper Functions
        public static int CSFENGCam_GetConfig(int iHandle, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;

            iResult = FENGCam_GetConfig(iHandle, ref mcfgCam, (IntPtr)0);
            if (iResult >= 0)
            {
                // Marshal the configuration to managed space
                MarshalConfigToManaged(ref mcfgCam, ref cscfgCam);

                // Free the unmanaged space so the caller does not have to
                FENGCam_FreeConfig(ref mcfgCam);
            }

            return (iResult);
        }

        public static int CSFENGCam_GetConfigFromFile(StringBuilder strFileName, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;
            //MFENG_CFGATODSETTING mcfgSetting;

            // needed for sizeof
            //mcfgSetting = new MFENG_CFGATODSETTING();

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;
            mcfgCam.cfgStartup.pCfgVSMIdleTimingRegs = (IntPtr)null; //(IntPtr)Marshal.AllocCoTaskMem(Marshal.SizeOf(mcfgSetting) * 26);
            mcfgCam.cfgStartup.pCfgVSMLongExpTimingRegs = (IntPtr)null; //(IntPtr)Marshal.AllocCoTaskMem(Marshal.SizeOf(mcfgSetting) * 26);
            mcfgCam.cfgStartup.uiNumVSMIdleTimingRegs = 0;
            mcfgCam.cfgStartup.uiNumVSMLongExpTimingRegs = 0;
            iResult = FENGCam_GetConfigFromFile(strFileName, ref mcfgCam);
            if (iResult >= 0)
            {
                // Marshal the configuration to managed space
                MarshalConfigToManaged(ref mcfgCam, ref cscfgCam);

                // Free the unmanaged space so the caller does not have to
                FENGCam_FreeConfig(ref mcfgCam);
            }

            return (iResult);
        }

        public static int CSFENGCam_SaveConfigToBin(StringBuilder strFileName, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;

            MarshalConfigToUnManaged(ref cscfgCam, ref mcfgCam);
            iResult = FENGCam_SaveConfigToBin(strFileName, ref mcfgCam);

            return (iResult);
        }
        public static int CSFENGCam_SaveConfigToCSV(StringBuilder strFileName, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;

            MarshalConfigToUnManaged(ref cscfgCam, ref mcfgCam);
            iResult = FENGCam_SaveConfigToCSV(strFileName, ref mcfgCam);

            return (iResult);
        }

        public static int CSFENGCam_SetConfig(int iHandle, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;

            MarshalConfigToUnManaged(ref cscfgCam, ref mcfgCam);
            iResult = FENGCam_SetConfig(iHandle, ref mcfgCam,(IntPtr)0);

            return (iResult);
        }


        public static int CSFENGCam_SetDeviceRegs(int iHandle,uint uiDeviceId, ref MFENG_CFGREG[] csRegs, uint uiNumRegs)
        {
            int iResult;
            IntPtr pRegs;

            // Marshal the data
            pRegs = MarshalCfgRegsToUnmanaged(uiNumRegs, ref csRegs);

            // call the FENG API ....
            iResult = FENGCfg_SetDeviceRegs(iHandle, uiDeviceId, ref pRegs, uiNumRegs);

            return (iResult);
        }

        public static int CSFENGCam_SetHSMTimings(int iHandle, ref MFENG_CFGREG[] csIdleTimings,uint uiNumIdleTimings,ref MFENG_CFGREG[] csBinningTimings, uint uiNumBinningTimings)
        {
            int iResult;
            IntPtr pIdleTimings;
            IntPtr pBinningTimings;

            // Marshal the data
            pIdleTimings = MarshalCfgRegsToUnmanaged(uiNumIdleTimings, ref csIdleTimings);
            pBinningTimings = MarshalCfgRegsToUnmanaged(uiNumBinningTimings, ref csBinningTimings);

            // call the FENG API ....
            iResult= FENGCfg_SetHSMTimings(iHandle, ref pIdleTimings, uiNumIdleTimings, ref pBinningTimings, uiNumBinningTimings);

            return (iResult);
        }

        public static int CSFENGCam_SetVSMTimings(int iHandle, ref MFENG_CFGREG[] csIdleTimings, uint uiNumIdleTimings, ref MFENG_CFGREG[] csLongExposureTimings, uint uiNumLongExposureTimings)
        {
            int iResult;
            IntPtr pIdleTimings;
            IntPtr pLongExposureTimings;

            // Marshal the data
            pIdleTimings = MarshalCfgRegsToUnmanaged(uiNumIdleTimings, ref csIdleTimings);
            pLongExposureTimings = MarshalCfgRegsToUnmanaged(uiNumLongExposureTimings, ref csLongExposureTimings);

            // call the FENG API ....
            iResult = FENGCfg_SetVSMTimings(iHandle, ref pIdleTimings, uiNumIdleTimings, ref pLongExposureTimings, uiNumLongExposureTimings);

            return (iResult);
        }


        private static void MarshalConfigToManaged(ref MFENG_CFGCAM mcfgCam, ref CSFENG_CFGCAM cscfgCam)
        {
            IntPtr ptrModes;
            CSFENG_CFGMODES cscfgMode;
            MFENG_CFGMODES  mcfgMode;

            // just needed for marshalling size
            cscfgMode = new CSFENG_CFGMODES();

            // Marshal the modes
            cscfgCam.uiNumModes = mcfgCam.uiNumModes;
            if (cscfgCam.uiNumModes > 0)
            {
                cscfgCam.modes = new CSFENG_CFGMODES[cscfgCam.uiNumModes];
                ptrModes = mcfgCam.pModes;
                for (uint i = 0; (i < cscfgCam.uiNumModes) && (ptrModes != null); ++i)
                {
                    mcfgMode= (MFENG_CFGMODES)Marshal.PtrToStructure(ptrModes, typeof(MFENG_CFGMODES));

                    // now we can marshal to the actual structure
                    MarshalModeToManaged(ref mcfgMode, ref cscfgCam.modes[i]);
                    //cscfgCam.modes[i] = (CSFENG_CFGMODES)Marshal.PtrToStructure(ptrModes, typeof(CSFENG_CFGMODES));
                    ptrModes += Marshal.SizeOf(cscfgMode);
                }
            }

            // Marshal the Startup
            MarshalStartupToManaged(ref mcfgCam.cfgStartup, ref cscfgCam.cfgStartup);

            // Marshal the Device Info
            MarshalDeviceInfoToManaged(ref mcfgCam.cfgCamInfo, ref cscfgCam.cfgCamInfo);

            //cscfgCam.cfgStartup = mcfgCam.cfgStartup;
        }

        private static void MarshalModeToManaged(ref MFENG_CFGMODES fromMode, ref CSFENG_CFGMODES toMode)
        {
            // Do the discrete settings first
            //HDR and LDR flags
            if (fromMode.bHDREnabled == 0)
                toMode.bHDREnabled = false;
            else
                toMode.bHDREnabled = true;
            if (fromMode.bLDCEnabled == 0)
                toMode.bLDCEnabled = false;
            else
                toMode.bLDCEnabled = true;

            // Modeinfo and gains
            toMode.modeInfo = fromMode.modeInfo;
            toMode.iLDRGainIndex = fromMode.iLDRGainIndex;
            toMode.iHDRGainIndex = fromMode.iHDRGainIndex;
            toMode.iGlobalGainIndex = fromMode.iGlobalGainIndex;
            toMode.uiBlackLevel = fromMode.uiBlackLevel;
            toMode.uiBlackSun = fromMode.uiBlackSun;
            toMode.dacImage1 = fromMode.dacImage1;
            toMode.dacImage2 = fromMode.dacImage2;

            // now the HSM Timing Settings
            toMode.uiNumHSMIdleTimingRegs = fromMode.uiNumHSMIdleTimingRegs;
            MarshalCfgRegsToManaged(fromMode.pCfgHSMIdleTimingRegs, fromMode.uiNumHSMIdleTimingRegs, out toMode.pCfgHSMIdleTimingRegs);
            toMode.uiNumHSMBinningTimingRegs = fromMode.uiNumHSMBinningTimingRegs;
            MarshalCfgRegsToManaged(fromMode.pCfgHSMBinningTimingRegs, fromMode.uiNumHSMBinningTimingRegs, out toMode.pCfgHSMBinningTimingRegs);

            // A to D Regs
            toMode.uiNumAtoD1Regs = fromMode.uiNumAtoD1Regs;
            MarshalCfgRegsToManaged(fromMode.pCfgSensorAtoD1, fromMode.uiNumAtoD1Regs, out toMode.pCfgSensorAtoD1);
            toMode.uiNumAtoD1Regs = fromMode.uiNumAtoD2Regs;
            MarshalCfgRegsToManaged(fromMode.pCfgSensorAtoD2, fromMode.uiNumAtoD2Regs, out toMode.pCfgSensorAtoD2);
        }

        private static void MarshalCfgRegsToManaged(IntPtr ptrSetting, UInt32 uiNumRegs, out MFENG_CFGREG[] cfgRegs)
        {
            MFENG_CFGREG cscfgSetting;

            cscfgSetting = new MFENG_CFGREG();

            cfgRegs = null;
            if (uiNumRegs > 0)
            {
                cfgRegs = new MFENG_CFGREG[uiNumRegs];
                for (uint i = 0; (i < uiNumRegs) && (ptrSetting != null); ++i)
                {
                    cfgRegs[i] = (MFENG_CFGREG)Marshal.PtrToStructure(ptrSetting, typeof(MFENG_CFGREG));
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
        }

        private static void MarshalDeviceInfoToManaged(ref MFENG_CAMINFO mcfgCamInfo, ref CSFENG_CAMINFO cscfgCamInfo)
        {
            cscfgCamInfo.uiDeviceType = mcfgCamInfo.uiDeviceType;
            cscfgCamInfo.uiSerialNumLength = mcfgCamInfo.uiSerialNumLength;
            cscfgCamInfo.wcSerialNumber = mcfgCamInfo.wcSerialNumber;
        }

        private static void MarshalStartupToManaged(ref MFENG_CAMSTARTUP mcfgStartup, ref CSFENG_CAMSTARTUP cscfgStartup)
        {
            IntPtr ptrSetting;
            MFENG_CFGREG cscfgSetting;

            cscfgSetting = new MFENG_CFGREG();

            // First the discrete settings
            cscfgStartup.dblTemperatureSetPoint = mcfgStartup.dblTemperatureSetPoint;
            cscfgStartup.uiCoolerDAC = mcfgStartup.uiCoolerDAC;
            cscfgStartup.uiFanPercent = mcfgStartup.uiFanPercent;
            cscfgStartup.bFanEnabled = mcfgStartup.bFanEnabled;
            cscfgStartup.uiHeaterPercent = mcfgStartup.uiHeaterPercent;
            cscfgStartup.uiNumVSMIdleTimingRegs = mcfgStartup.uiNumVSMIdleTimingRegs;
            cscfgStartup.uiNumVSMLongExpTimingRegs = mcfgStartup.uiNumVSMLongExpTimingRegs;
            cscfgStartup.uiNumOtherRegs = mcfgStartup.uiNumOtherRegs;

            // Marshal the A to D settings
            if (cscfgStartup.uiNumVSMIdleTimingRegs > 0)
            {
                cscfgStartup.pCfgVSMIdleTimingRegs = new MFENG_CFGREG[cscfgStartup.uiNumVSMIdleTimingRegs];
                ptrSetting = mcfgStartup.pCfgVSMIdleTimingRegs;
                for (uint i = 0; (i < cscfgStartup.uiNumVSMIdleTimingRegs) && (ptrSetting != (IntPtr)0); ++i)
                {
                    cscfgStartup.pCfgVSMIdleTimingRegs[i] = (MFENG_CFGREG)Marshal.PtrToStructure(ptrSetting, typeof(MFENG_CFGREG));
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
            if (cscfgStartup.uiNumVSMLongExpTimingRegs > 0)
            {
                cscfgStartup.pCfgVSMLongExpTimingRegs = new MFENG_CFGREG[cscfgStartup.uiNumVSMLongExpTimingRegs];
                ptrSetting = mcfgStartup.pCfgVSMLongExpTimingRegs;
                for (uint i = 0; (i < cscfgStartup.uiNumVSMLongExpTimingRegs) && (ptrSetting != (IntPtr)0); ++i)
                {
                    cscfgStartup.pCfgVSMLongExpTimingRegs[i] = (MFENG_CFGREG)Marshal.PtrToStructure(ptrSetting, typeof(MFENG_CFGREG));
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }

            // 'Other' Registers
            if (cscfgStartup.uiNumOtherRegs > 0)
            {
                cscfgStartup.pOtherRegs = new MFENG_CFGREG[cscfgStartup.uiNumOtherRegs];
                ptrSetting = mcfgStartup.pOtherRegs;
                for (uint i = 0; (i < cscfgStartup.uiNumOtherRegs) && (ptrSetting != (IntPtr)0); ++i)
                {
                    cscfgStartup.pOtherRegs[i] = (MFENG_CFGREG)Marshal.PtrToStructure(ptrSetting, typeof(MFENG_CFGREG));
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
        }

        private static void MarshalConfigToUnManaged(ref CSFENG_CFGCAM cscfgCam, ref MFENG_CFGCAM mcfgCam)
        {
            IntPtr ptrModes;
            CSFENG_CFGMODES cscfgMode;
            MFENG_CFGMODES  mcfgMode;

            // just needed for marshalling size
            cscfgMode = new CSFENG_CFGMODES();
            mcfgMode = new MFENG_CFGMODES();

            // Marshal the modes
            mcfgCam.uiNumModes = cscfgCam.uiNumModes;
            if (mcfgCam.uiNumModes > 0)
            {
                ptrModes = Marshal.AllocCoTaskMem(Marshal.SizeOf(cscfgMode) * (int)mcfgCam.uiNumModes);
                mcfgCam.pModes = ptrModes;
                for (uint i = 0; (i < cscfgCam.uiNumModes) && (ptrModes != null); ++i)
                {
                    // Fill up a mode sructure
                    MarshalModeToUnmanaged(ref cscfgCam.modes[i], ref mcfgMode);

                    // now marshal to our pointer
                    Marshal.StructureToPtr(mcfgMode, ptrModes,false);

                    // fill in the next mode
                    ptrModes += Marshal.SizeOf(cscfgMode);
                }
            }

            // Marshal the Startup
            MarshalStartupToUnManaged(ref cscfgCam.cfgStartup, ref mcfgCam.cfgStartup);

            // Marshal the Device Info
            MarshalDeviceInfoToUnManaged(ref cscfgCam.cfgCamInfo, ref mcfgCam.cfgCamInfo);
        }

        private static void MarshalModeToUnmanaged(ref CSFENG_CFGMODES cscfgMode, ref MFENG_CFGMODES mcfgMode)
        {
            // HDR and LDC Flags
            if (cscfgMode.bHDREnabled)
                mcfgMode.bHDREnabled = 1;
            else
                mcfgMode.bHDREnabled = 0;
            if (cscfgMode.bLDCEnabled)
                mcfgMode.bLDCEnabled = 1;
            else
                mcfgMode.bLDCEnabled = 0;

            // First do all the discrete settings
            mcfgMode.modeInfo = cscfgMode.modeInfo;
            mcfgMode.iLDRGainIndex = cscfgMode.iLDRGainIndex;
            mcfgMode.iHDRGainIndex = cscfgMode.iHDRGainIndex;
            mcfgMode.iGlobalGainIndex = cscfgMode.iGlobalGainIndex;
            mcfgMode.uiBlackLevel = cscfgMode.uiBlackLevel;
            mcfgMode.uiBlackSun = cscfgMode.uiBlackSun;
            mcfgMode.dacImage1 = cscfgMode.dacImage1;
            mcfgMode.dacImage2 = cscfgMode.dacImage2;

            // now the HSM Timing Settings
            mcfgMode.uiNumHSMIdleTimingRegs = cscfgMode.uiNumHSMIdleTimingRegs;
            mcfgMode.pCfgHSMIdleTimingRegs = MarshalCfgRegsToUnmanaged(mcfgMode.uiNumHSMIdleTimingRegs, ref cscfgMode.pCfgHSMIdleTimingRegs);
            mcfgMode.uiNumHSMBinningTimingRegs = cscfgMode.uiNumHSMBinningTimingRegs;
            mcfgMode.pCfgHSMBinningTimingRegs= MarshalCfgRegsToUnmanaged(mcfgMode.uiNumHSMBinningTimingRegs,ref cscfgMode.pCfgHSMBinningTimingRegs);

            // A to D Regs
            mcfgMode.uiNumAtoD1Regs = cscfgMode.uiNumAtoD1Regs;
            mcfgMode.pCfgSensorAtoD1 = MarshalCfgRegsToUnmanaged(mcfgMode.uiNumAtoD1Regs, ref cscfgMode.pCfgSensorAtoD1);
            mcfgMode.uiNumAtoD2Regs = cscfgMode.uiNumAtoD2Regs;
            mcfgMode.pCfgSensorAtoD2 = MarshalCfgRegsToUnmanaged(mcfgMode.uiNumAtoD2Regs, ref cscfgMode.pCfgSensorAtoD2);
        }

        private static IntPtr MarshalCfgRegsToUnmanaged(UInt32 uiNumRegs, ref MFENG_CFGREG[] cfgRegs)
        {
            MFENG_CFGREG cscfgSetting;
            IntPtr ptrSetting;
            IntPtr ptrSettingStart;

            // just needed for marshalling size
            cscfgSetting = new MFENG_CFGREG();

            ptrSetting = (IntPtr)0;
            ptrSettingStart = (IntPtr)0;
            if (uiNumRegs > 0)
            {
                ptrSettingStart = Marshal.AllocCoTaskMem(Marshal.SizeOf(cscfgSetting) * (int)uiNumRegs);
                if (ptrSettingStart != (IntPtr)0)
                {
                    ptrSetting = ptrSettingStart;
                    for (uint i = 0; (i < uiNumRegs) && (ptrSetting != (IntPtr)0); ++i)
                    {
                        Marshal.StructureToPtr(cfgRegs[i], ptrSetting, false);
                        ptrSetting += Marshal.SizeOf(cscfgSetting);
                    }
                }
            }

            // return our result
            return (ptrSettingStart);
        }

        private static void MarshalDeviceInfoToUnManaged(ref CSFENG_CAMINFO cscfgCamInfo, ref MFENG_CAMINFO mcfgCamInfo)
        {
            mcfgCamInfo.uiDeviceType = cscfgCamInfo.uiDeviceType;
            mcfgCamInfo.uiSerialNumLength = cscfgCamInfo.uiSerialNumLength;
            mcfgCamInfo.wcSerialNumber = cscfgCamInfo.wcSerialNumber;
        }


        private static void MarshalStartupToUnManaged(ref CSFENG_CAMSTARTUP cscfgStartup, ref MFENG_CAMSTARTUP mcfgStartup)
        {
            IntPtr ptrSetting;
            MFENG_CFGREG cscfgSetting;

            // just needed for marshalling size
            cscfgSetting = new MFENG_CFGREG();

            // First the discrete settings
            mcfgStartup.dblTemperatureSetPoint = cscfgStartup.dblTemperatureSetPoint;
            mcfgStartup.uiCoolerDAC = cscfgStartup.uiCoolerDAC;
            mcfgStartup.uiFanPercent = cscfgStartup.uiFanPercent;
            mcfgStartup.bFanEnabled = cscfgStartup.bFanEnabled;
            mcfgStartup.uiHeaterPercent = cscfgStartup.uiHeaterPercent;
            mcfgStartup.uiNumVSMIdleTimingRegs = cscfgStartup.uiNumVSMIdleTimingRegs;
            mcfgStartup.uiNumVSMLongExpTimingRegs = cscfgStartup.uiNumVSMLongExpTimingRegs;
            mcfgStartup.uiNumOtherRegs = cscfgStartup.uiNumOtherRegs;

            // Marshal the VSM Settings
            if (mcfgStartup.uiNumVSMIdleTimingRegs > 0)
            {
                ptrSetting = Marshal.AllocCoTaskMem(Marshal.SizeOf(cscfgSetting) * (int)mcfgStartup.uiNumVSMIdleTimingRegs);
                mcfgStartup.pCfgVSMIdleTimingRegs = ptrSetting;
                for (uint i = 0; (i < mcfgStartup.uiNumVSMIdleTimingRegs) && (ptrSetting != null); ++i)
                {
                    Marshal.StructureToPtr(cscfgStartup.pCfgVSMIdleTimingRegs[i], ptrSetting, false);
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
            if (mcfgStartup.uiNumVSMLongExpTimingRegs > 0)
            {
                ptrSetting = Marshal.AllocCoTaskMem(Marshal.SizeOf(cscfgSetting) * (int)mcfgStartup.uiNumVSMLongExpTimingRegs);
                mcfgStartup.pCfgVSMLongExpTimingRegs = ptrSetting;
                for (uint i = 0; (i < mcfgStartup.uiNumVSMLongExpTimingRegs) && (ptrSetting != null); ++i)
                {
                    Marshal.StructureToPtr(cscfgStartup.pCfgVSMLongExpTimingRegs[i], ptrSetting, false);
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }

            // 'Other' Registers
            if (mcfgStartup.uiNumOtherRegs > 0)
            {
                ptrSetting = Marshal.AllocCoTaskMem(Marshal.SizeOf(cscfgSetting) * (int)mcfgStartup.uiNumOtherRegs);
                mcfgStartup.pOtherRegs = ptrSetting;
                for (uint i = 0; (i < mcfgStartup.uiNumOtherRegs) && (ptrSetting != null); ++i)
                {
                    Marshal.StructureToPtr(cscfgStartup.pOtherRegs[i], ptrSetting, false);
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
        }

        #endregion Engineering Marshalling Wrapper Functions

        #endregion Engineering Functions
    }
}
