﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// Configuration file management class
//
// Copyright(c) 2014-2016 AppliedLogix, LLC
// Portions of this library are Copyright (c) 2014 VariPath Technologies, Inc.
// Derived from IniFileCS.cs Copyright(c) 2010, Ludvik Jerabek,
// which is released under the Code Project Open License (CPOL)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;

namespace FliUtils.Config
{
    // IniFile class used to read and write ini files by loading the file into memory
    public class IniFile
    {
        // List of IniSection objects keeps track of all the sections in the INI file
        private Hashtable m_sections;

        // Public constructor
        public IniFile()
        {
            m_sections = new Hashtable(StringComparer.InvariantCultureIgnoreCase);
        }

        // Loads the Reads the data in the ini file into the IniFile object
        public Boolean LoadFile(String sFileName, Boolean bMerge = false)
        {
            if (!bMerge)
            {
                RemoveAllSections();
            }

            //  Clear the object... 
            IniSection tempsection = null;

            StreamReader oReader;
            try
            {
                oReader = new StreamReader(sFileName);
            }
            catch(Exception)
            {
                // Cannot open or read file
                return false;
            }

            Regex regexcomment = new Regex("^([\\s]*#.*)", (RegexOptions.Singleline | RegexOptions.IgnoreCase));

            Regex regexsection = new Regex("^[\\s]*\\[[\\s]*([^\\[\\s].*[^\\s\\]])[\\s]*\\][\\s]*$", (RegexOptions.Singleline | RegexOptions.IgnoreCase));

            Regex regexkey = new Regex("^\\s*([^=\\s]*)[^=]*=(.*)", (RegexOptions.Singleline | RegexOptions.IgnoreCase));

            while (!oReader.EndOfStream)
            {
                String line = oReader.ReadLine();
                if (line != String.Empty)
                {
                    Match m = null;
                    if (regexcomment.Match(line).Success)
                    {
                        m = regexcomment.Match(line);
                    }
                    else if (regexsection.Match(line).Success)
                    {
                        m = regexsection.Match(line);
                        tempsection = AddSection(m.Groups[1].Value);
                    }
                    else if (regexkey.Match(line).Success && tempsection != null)
                    {
                        m = regexkey.Match(line);
                        tempsection.AddKey(m.Groups[1].Value).Value = m.Groups[2].Value;
                    }
                    else if (tempsection != null)
                    {
                        //  Handle Key without value
                        tempsection.AddKey(line);
                    }
                    else
                    {
                        //  This should not occur unless the tempsection is not created yet...
                    }
                }
            }
            oReader.Close();
            return true;
        }

        // Used to save the data back to the file or your choice
        public void SaveFile(String sFileName)
        {
            StreamWriter oWriter = new StreamWriter(sFileName, false);
            foreach (IniSection s in Sections)
            {
                oWriter.WriteLine(String.Format("[{0}]", s.Name));
                foreach (IniSection.IniKey k in s.Keys)
                {
                    if (k.Value != String.Empty)
                    {
                        oWriter.WriteLine(String.Format("{0}={1}", k.Name, k.Value));
                    }
                    else
                    {
                        oWriter.WriteLine(String.Format("{0}", k.Name));
                    }
                }
            }
            oWriter.Close();
        }

        // Number of sections
        public UInt32 NumSections 
        {
            get {return (UInt32)m_sections.Count;}
        }

        // Total number of keys
        public UInt32 NumKeys 
        {
            get
            {
                UInt32 numkeys = 0;

                foreach (IniSection section in m_sections.Values) 
                {
                    numkeys += section.NumKeys;
                }
                return numkeys;
            }
        }

        // Gets all the sections names
        public System.Collections.ICollection Sections
        {
            get { return m_sections.Values; }
        }

        public IEnumerable<IniSection> FindSections(string prefix)
        {
            var selected = from string s in m_sections.Keys
                           where s.StartsWith(prefix,StringComparison.CurrentCultureIgnoreCase)
                           select m_sections[s];

            return selected.Cast<IniSection>().OrderBy(s => s.Name);
        }

        // Adds a section to the IniFile object, returns a IniSection object to the new or existing object
        public IniSection AddSection(String sSection)
        {
            IniSection s = null;
            sSection = sSection.Trim();
            // Trim spaces
            if (m_sections.ContainsKey(sSection))
            {
                s = (IniSection)m_sections[sSection];
            }
            else
            {
                s = new IniSection(this, sSection);
                m_sections[sSection] = s;
            }
            return s;
        }

        // Removes a section by its name sSection, returns trus on success
        public Boolean RemoveSection(String sSection)
        {
            sSection = sSection.Trim();
            return RemoveSection(GetSection(sSection));
        }

        // Removes section by object, returns trus on success
        public Boolean RemoveSection(IniSection Section)
        {
            if (Section != null)
            {
                try
                {
                    m_sections.Remove(Section.Name);
                    return true;
                }
                catch (Exception)
                {
                    // ok
                }
            }
            return false;
        }

        //  Removes all existing sections, returns trus on success
        public Boolean RemoveAllSections()
        {
            m_sections.Clear();
            return (m_sections.Count == 0);
        }

        // Returns an IniSection to the section by name, NULL if it was not found
        public IniSection GetSection(String sSection)
        {
            if(String.IsNullOrEmpty(sSection))
            {
                return null;
            }

            sSection = sSection.Trim();
            // Trim spaces
            if (m_sections.ContainsKey(sSection))
            {
                return (IniSection)m_sections[sSection];
            }
            return null;
        }

        //  Returns a KeyValue in specified section as a String
        public String GetKeyValue(String sSection, String sKey,
                                  String defvalue = "")
        {
            IniSection s = GetSection(sSection);
            if (s != null)
            {
                IniSection.IniKey k = s.GetKey(sKey);
                if (k != null && !String.IsNullOrEmpty(k.Value))
                {
                    return k.Value;
                }
            }

            return defvalue;
        }

        // Returns a KeyValue in specified section as an unsigned int
        public UInt32 GetKeyValue(String sSection, String sKey,
                                 UInt32 defvalue,
                                 UInt32 minvalue = UInt32.MinValue,
                                 UInt32 maxvalue = UInt32.MaxValue,
                                 Int32 fromBase = 10)
        {
            String strval = GetKeyValue(sSection, sKey);
            UInt32 val = defvalue;
            if (!String.IsNullOrEmpty(strval))
            {
                try
                {
                    val = Convert.ToUInt32(strval, fromBase);
                }
                catch (Exception)
                {
                    // ok;
                }
            }
            if (val < minvalue)
            {
                val = minvalue;
            }
            else if (val > maxvalue)
            {
                val = maxvalue;
            }

            return val;
        }

        // Returns a KeyValue in specified section as a signed int
        public Int32 GetKeyValue(String sSection, String sKey,
                                 Int32 defvalue,
                                 Int32 minvalue = Int32.MinValue,
                                 Int32 maxvalue = Int32.MaxValue)
        {
            String strval = GetKeyValue(sSection, sKey);
            Int32 val = defvalue;
            if (!String.IsNullOrEmpty(strval))
            {
                try
                {
                    val = Convert.ToInt32(strval);
                }
                catch (Exception)
                {
                    // ok
                }
            }
            if (val < minvalue)
            {
                val = minvalue;
            }
            else if (val > maxvalue)
            {
                val = maxvalue;
            }

            return val;
        }

        // Returns a KeyValue in specified section as an Int16
        public Int16 GetKeyValue(String sSection, String sKey,
                                 Int16 defvalue,
                                 Int16 minvalue = Int16.MinValue,
                                 Int16 maxvalue = Int16.MaxValue)
        {
            String strval = GetKeyValue(sSection, sKey);
            Int16 val = defvalue;
            if (!String.IsNullOrEmpty(strval))
            {
                try
                {
                    val = Convert.ToInt16(strval);
                }
                catch(Exception)
                {
                    // ok
                }
            }
            if (val < minvalue)
            {
                val = minvalue;
            }
            else if (val > maxvalue)
            {
                val = maxvalue;
            }

            return val;
        }

        // Returns a KeyValue in specified section as a double
        public Double GetKeyValue(String sSection, String sKey,
                                  Double defvalue,
                                  Double minvalue = Double.MinValue,
                                  Double maxvalue = Double.MaxValue)
        {
            String strval = GetKeyValue(sSection, sKey);
            Double val = defvalue;
            if (!String.IsNullOrEmpty(strval))
            {
                try
                {
                    val = Convert.ToDouble(strval);
                }
                catch(Exception)
                {
                    // ok;
                }
            }

            if (val < minvalue)
            {
                val = minvalue;
            }
            else if (val > maxvalue)
            {
                val = maxvalue;
            }

            return val;
        }

        // Returns a KeyValue in specified section as a boolean
        public Boolean GetKeyValue(String sSection, String sKey,
                                  Boolean defvalue)
        {
            String strval = GetKeyValue(sSection, sKey, "").ToLower();

            if (strval == "1" || strval == "t" || strval == "true" ||
               strval == "on" || strval == "y" || strval == "yes" ||
               strval == "enable")
            {
                return true;
            }
            else if (strval == "0" || strval == "f" || strval == "false" ||
                     strval == "off" || strval == "n" || strval == "no" ||
                     strval == "disable")
            {
                return false;
            }
            else
            {
                return defvalue;
            }
        }

        // Sets a KeyValuePair in a certain section
        public Boolean SetKeyValue(String sSection, String sKey, String sValue)
        {
            IniSection s = AddSection(sSection);
            if (s != null)
            {
                IniSection.IniKey k = s.AddKey(sKey);
                if (k != null)
                {
                    k.Value = sValue;
                    return true;
                }
            }
            return false;
        }

        // Renames an existing section returns true on success, false if the section didn't exist or there was another section with the same sNewSection
        public Boolean RenameSection(String sSection, String sNewSection)
        {
            //  Note String trims are done in lower calls.
            Boolean bRval = false;
            IniSection s = GetSection(sSection);
            if (s != null)
            {
                bRval = s.SetName(sNewSection);
            }
            return bRval;
        }

        // Renames an existing key returns true on success, false if the key didn't exist or there was another section with the same sNewKey
        public Boolean RenameKey(String sSection, String sKey, String sNewKey)
        {
            //  Note String trims are done in lower calls.
            IniSection s = GetSection(sSection);
            if (s != null)
            {
                IniSection.IniKey k = s.GetKey(sKey);
                if (k != null)
                {
                    return k.SetName(sNewKey);
                }
            }
            return false;
        }

        // IniSection class 
        public class IniSection
        {
            //  IniFile IniFile object instance
            private IniFile m_pIniFile;
            //  Name of the section
            private String m_sSection;
            //  List of IniKeys in the section
            private Hashtable m_keys;

            // Constructor so objects are internally managed
            protected internal IniSection(IniFile parent, String sSection)
            {
                m_pIniFile = parent;
                m_sSection = sSection;
                m_keys = new Hashtable(StringComparer.InvariantCultureIgnoreCase);
            }

            // Number of keys in the section
            public UInt32 NumKeys 
            {
                get { return (UInt32) m_keys.Count; }
            }

            // Returns and hashtable of keys associated with the section
            public System.Collections.ICollection Keys
            {
                get { return m_keys.Values; }
            }

            // Returns the section name
            public String Name
            {
                get { return m_sSection; }
            }

            // Adds a key to the IniSection object, returns a IniKey object to the new or existing object
            public IniKey AddKey(String sKey)
            {
                sKey = sKey.Trim();
                IniSection.IniKey k = null;
                if (sKey.Length != 0)
                {
                    if (m_keys.ContainsKey(sKey))
                    {
                        k = (IniKey)m_keys[sKey];
                    }
                    else
                    {
                        k = new IniSection.IniKey(this, sKey);
                        m_keys[sKey] = k;
                    }
                }
                return k;
            }

            // Removes a single key by String
            public Boolean RemoveKey(String sKey)
            {
                return RemoveKey(GetKey(sKey));
            }

            // Removes a single key by IniKey object
            public Boolean RemoveKey(IniKey Key)
            {
                if (Key != null)
                {
                    try
                    {
                        m_keys.Remove(Key.Name);
                        return true;
                    }
                    catch (Exception)
                    {
                        // ok
                    }
                }
                return false;
            }

            // Removes all the keys in the section
            public Boolean RemoveAllKeys()
            {
                m_keys.Clear();
                return (m_keys.Count == 0);
            }

            // Returns a IniKey object to the key by name, NULL if it was not found
            public IniKey GetKey(String sKey)
            {
                sKey = sKey.Trim();
                if (m_keys.ContainsKey(sKey))
                {
                    return (IniKey)m_keys[sKey];
                }
                return null;
            }

            // Sets the section name, returns true on success, fails if the section
            // name sSection already exists
            public Boolean SetName(String sSection)
            {
                sSection = sSection.Trim();
                if (sSection.Length != 0)
                {
                    // Get existing section if it even exists...
                    IniSection s = m_pIniFile.GetSection(sSection);
                    if (s != this && s != null) return false;
                    try
                    {
                        // Remove the current section
                        m_pIniFile.m_sections.Remove(m_sSection);
                        // Set the new section name to this object
                        m_pIniFile.m_sections[sSection] = this;
                        // Set the new section name
                        m_sSection = sSection;
                        return true;
                    }
                    catch (Exception)
                    {
                        // ignore
                    }
                }
                return false;
            }

            // Returns the section name
            public String GetName()
            {
                return m_sSection;
            }

            // IniKey class
            public class IniKey
            {
                //  Name of the Key
                private String m_sKey;
                //  Value associated
                private String m_sValue;
                //  Pointer to the parent CIniSection
                private IniSection m_section;

                // Constuctor so objects are internally managed
                protected internal IniKey(IniSection parent, String sKey)
                {
                    m_section = parent;
                    m_sKey = sKey;
                }

                // Returns the name of the Key
                public String Name
                {
                    get { return m_sKey; }
                }

                // Sets or Gets the value of the key
                public String Value
                {
                    get { return m_sValue; }
                    set { m_sValue = value; }
                }

                // Sets the value of the key
                public void SetValue(String sValue)
                {
                    m_sValue = sValue;
                }
                // Returns the value of the Key
                public String GetValue()
                {
                    return m_sValue;
                }

                // Sets the key name
                // Returns true on success, fails if the section name sKey already exists
                public Boolean SetName(String sKey)
                {
                    sKey = sKey.Trim();
                    if (sKey.Length != 0)
                    {
                        IniKey k = m_section.GetKey(sKey);
                        if (k != this && k != null) return false;
                        try
                        {
                            // Remove the current key
                            m_section.m_keys.Remove(m_sKey);
                            // Set the new key name to this object
                            m_section.m_keys[sKey] = this;
                            // Set the new key name
                            m_sKey = sKey;
                            return true;
                        }
                        catch (Exception ex)
                        {
                            Trace.WriteLine(ex.Message);
                        }
                    }
                    return false;
                }

                // Returns the name of the Key
                public String GetName()
                {
                    return m_sKey;
                }
            } // End of IniKey class
        } // End of IniSection class
    } // End of IniFile class

}

//--------------------------------------------------------
// End of file
//--------------------------------------------------------
