﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// Periodic Background job 
//
// Copyright(c) 2016 AppliedLogix, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************

using System;
using System.Threading;

namespace FliUtils.Jobs
{
    public class PeriodicJob
    {
        //**************************************************************
        #region Constructor

        public PeriodicJob(Action action, TimeSpan period, bool waitBeforeAction = true, 
                           string name = "", ThreadPriority priority = ThreadPriority.Normal)
        {
            _stopped = true;
            _period = period;
            _action = action;
            _waitBeforeAction = waitBeforeAction;
            _name = name;
            _priority = priority;
        }

        #endregion Constructor

        //**************************************************************
        #region Private Properties

        private volatile bool _stopped;
        private bool _waitBeforeAction;
        private Action _action;
        private Thread _thread;
        private TimeSpan _period;
        private ThreadPriority _priority;
        private string _name;

        #endregion Private Properties

        //**************************************************************
        #region Public Properties

        public int ThreadId { get { return (_thread == null) ? 0 : _thread.ManagedThreadId; } }
        public string ThreadName { get { return (_thread == null) ? "null" : _thread.Name; } }

        #endregion Public Properties

        //**************************************************************
        #region Public Methods

        //-------------------------------------------
        /// <summary>
        /// Start the thread
        /// </summary>
        public void Start()
        {
            if (_stopped)
            {
                _stopped = false;
                _thread = new Thread(new ThreadStart(ThreadBody));
                _thread.IsBackground = true;
                if (!String.IsNullOrEmpty(_name))
                {
                    _thread.Name = _name;
                }
                _thread.Priority = _priority;
                _thread.Start();
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Stop the thread
        /// </summary>
        public void Stop()
        {
            if (!_stopped)
            {
                _stopped = true;
                _thread.Interrupt();
            }
        }

        #endregion Public Methods

        //**************************************************************
        #region Private Methods

        private void Wait(DateTime actionStartTime)
        {
            TimeSpan waitTime = actionStartTime.Add(_period) - DateTime.UtcNow;
            if (waitTime.Ticks > 0)
            {
                Thread.Sleep(waitTime);
            }
        }

        private void ThreadBody()
        {
            try
            {
                // If period is zero, perform the action exactly once and then exit.
                if (_period.Ticks == 0)
                {
                    _action();
                    return;
                }

                // Perform the acion periodically, until stopped or Interrupted.
                DateTime actionStartTime = DateTime.UtcNow;
                while (!_stopped)
                {
                    if (_waitBeforeAction)
                    {
                        Wait(actionStartTime);
                    }

                    try
                    {
                        actionStartTime = DateTime.UtcNow;
                        _action();
                    }
                    catch (ThreadInterruptedException)
                    {
                        throw;
                    }
                    catch (Exception)
                    {
                        throw;
                    }

                    if (!_waitBeforeAction)
                    {
                        Wait(actionStartTime);
                    }
                }
            }
            catch (ThreadInterruptedException)
            {
                if (_stopped)
                {
                    // Normal exit path when Stop is called
                }
                else
                {
                    throw;
                }
            }
            catch (Exception)
            {
                throw;
            }
        }

        #endregion Private Methods
    }
}
