﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// Start/Stop Background job 
//
// Copyright(c) 2016 AppliedLogix, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************

using System;
using System.Diagnostics;
using System.Threading;

namespace FliUtils.Jobs
{
    public class StartStopJob
    {
        //**************************************************************
        #region Constructors

        public StartStopJob(Action<string[]> startAction, string name = "",
            ThreadPriority priority = ThreadPriority.Normal) :
            this(startAction, null, TimeSpan.Zero, name, priority)
        { }

        public StartStopJob(Action<string[]> startAction, Action stopAction, TimeSpan stopTimeLimit, 
                            string name = "", ThreadPriority priority = ThreadPriority.Normal )
        {
            _stopped = true;
            _startAction = startAction;
            _stopAction = stopAction;
            _stopTimeLimit = stopTimeLimit;
            _name = name;
            _priority = priority;
        }

        #endregion Constructors

        //**************************************************************
        #region Private Properties

        private volatile bool _stopped;
        private Thread _thread;
        private Action<string[]> _startAction;
        private Action _stopAction;
        private TimeSpan _stopTimeLimit;
        private string[] _startArgs;
        private ThreadPriority _priority;
        private string _name;

        #endregion Private Properties

        //**************************************************************
        #region Public Properties

        public int ThreadId { get { return (_thread == null) ? 0 : _thread.ManagedThreadId; } }
        public string ThreadName { get { return (_thread == null) ? "null" : _thread.Name; } }

        #endregion Public Properties

        //**************************************************************
        #region Public Methods

        //-------------------------------------------
        /// <summary>
        /// Start the thread, run the Start Action
        /// </summary>
        public void Start()
        {
            Start(new string[0]);
        }

        //-------------------------------------------
        /// <summary>
        /// Start the thread, run the Start Action
        /// </summary>
        /// <param name="args"></param>
        public void Start(string[] args)
        {
            lock (this)
            {
                if (_stopped)
                {
                    _startArgs = args;
                    _stopped = false;
                    _thread = new Thread(new ThreadStart(JobThread));
                    _thread.IsBackground = true;
                    if (!String.IsNullOrEmpty(_name))
                    {
                        _thread.Name = _name;
                    }
                    _thread.Priority = _priority;
                    _thread.Start();
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Stop the thread, run the stop action, and wait for thread to complete
        /// </summary>
        public void Stop()
        {
            Thread threadToJoin = null;
            lock (this)
            {
                if (!_stopped)
                {
                    threadToJoin = _thread;
                    _stopped = true;
                    Monitor.Pulse(this);
                }
            }

            if (threadToJoin != null && !threadToJoin.Join(_stopTimeLimit))
            {
                throw new Exception("Thread join timeout");
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Wait for up to the specified time limit for the thread to complete
        /// </summary>
        /// <param name="waitTimeLimit"></param>
        /// <returns>true if thread completed, false if still running</returns>
        public bool Wait(TimeSpan waitTimeLimit)
        {
            Thread threadToJoin = null;
            bool ok = true;
            lock (this)
            {
                if (!_stopped)
                {
                    threadToJoin = _thread;
                }
            }

            if (threadToJoin != null && !threadToJoin.Join(waitTimeLimit))
            {
                ok = false;
            }
            return ok;
        }

        #endregion Public Methods

        //**************************************************************
        #region Private Methods

        //-------------------------------------------
        /// <summary>
        /// Actual thread function
        /// </summary>
        private void JobThread()
        {
            try
            {
                DateTime start = DateTime.Now;
                try
                {
                    _startAction(_startArgs);
                }
                catch (ThreadInterruptedException)
                {
                    throw;
                }
                catch (Exception)
                {
                    throw;
                }
                DateTime end = DateTime.Now;
                Debug.Print("StartStopJob _startAction() finished in {0}", end - start);

                if (_stopAction == null)
                {
                    return;
                }

                while (true)
                {
                    lock (this)
                    {
                        if (_stopped)
                        {
                            start = DateTime.Now;
                            try
                            {
                                _stopAction();
                            }
                            catch (ThreadInterruptedException)
                            {
                                throw;
                            }
                            catch (Exception)
                            {
                                throw;
                            }
                            end = DateTime.Now;
                            Debug.Print("StartStopJob _stopAction() finished in {0}", end - start);

                            break;
                        }
                        Monitor.Wait(this);
                    }
                }
            }
            catch (ThreadInterruptedException)
            {
                if (!_stopped)
                {
                    throw;
                }
            }
            catch (Exception)
            {
                throw;
            }
        }

        #endregion Private Methods
    }
}
